/***************************************************************************
 *   Copyright (C) 2005 - 2006 by                                          *
 *      Christian Muehlhaeuser, Last.fm Ltd <chris@last.fm>                *
 *      Erik Jaelevik, Last.fm Ltd <erik@last.fm>                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Steet, Fifth Floor, Boston, MA  02111-1307, USA.          *
 ***************************************************************************/

#include "container.h"
#include "containerutils.h"

#include "aboutdialog.h"
#include "configwizard.h"
#include "updatewizard.h"
#include "metadata.h"
#include "logindialog.h"
#include "deleteuserdialog.h"
#include "iconshack.h"
#include "volumeslider.h"
#include "winutils.h"
#include "confirmdialog.h"
#include "browserthread.h"

#ifdef WIN32
    #include "winstyleoverrides.h"
#endif

#ifdef Q_WS_MAC
    #include "itunesscript.h"
    #include "macstyleoverrides.h"
#endif

// Singleton instance needs to be initialised
Container* Container::mpInstance = NULL;
QMutex Container::m_mutex;

static const int k_statusBarFontSizeMac = 10;

Container::Container( QWidget *parent )
    : QMainWindow( parent )
    , m_iWebService( 0 )
    , m_eUserInfo( 0 )
    , m_sidebarEnabled( false )
    , m_settingsDialog( this )
{
    // REFACTOR! The size of this monster is scary.

    setWindowTitle( tr( "Last.fm" ) );

#ifdef Q_WS_MAC
    //setWindowAttributes( WA_MacMetalStyle );
#endif

    // This is needed so we can pass MetaData objects as signal/slot params
    // with queued connections.
    qRegisterMetaType<TrackInfo>( "TrackInfo" );
    qRegisterMetaType<MetaData>( "MetaData" );
    qRegisterMetaType<UserMetaData>( "UserMetaData" );

    ui.setupUi( this );
    #ifdef Q_WS_X11
        setWindowIcon( QIcon( dataPath( "icons/as.png" ) ) );
    #endif
    setMinimumSize( 551, 400 );
    setAcceptDrops( true );

    loadServices();
    loadExtensions();
    initListener();
    m_controlInterface = new ControlInterface( this );

    for ( int i = 1; i < qApp->arguments().size(); i++ )
    {
        if ( qApp->arguments().at( i ).contains( "lastfm://" ) )
            setPreloadStation( qApp->arguments().at( i ) );
    }

    // <ToolBar>
    QToolBar* toolbar = addToolBar( "Last.fm" );
    toolbar->setObjectName( "Lastfm_Main_Toolbar" );
    toolbar->setMovable( false );
    toolbar->setContextMenuPolicy( Qt::CustomContextMenu );
    toolbar->setAutoFillBackground( true );
    toolbar->setToolButtonStyle( Qt::ToolButtonTextUnderIcon );
    #ifdef WIN32
        // Can't use 32 for height as labels get truncated on Win classic
        toolbar->setIconSize( QSize( 50, 34 ) );
    #else
        toolbar->setIconSize( QSize( 32, 32 ) );
    #endif

    ui.actionSidebar->setShortcut( tr( "Ctrl+F" ) );
    ui.actionSidebar->setToolTip( tr( "Show/hide my profile" ) );
    #ifdef WIN32
    ui.actionSidebar->setCheckable( true );
    #endif
    connect( ui.actionSidebar, SIGNAL( triggered() ),
             this,             SLOT( enableSidebar() ) );

    // Copy the menu action and add the icon to it. Don't want the icon in the
    // menu as it will obscure the tick.
    m_sidebarActionWithIcon = new QAction( toolbar );
    m_sidebarActionWithIcon->setIcon( QIcon( dataPath( "buttons/myprofile.png" ) ) );
    //*m_sidebarActionWithIcon = *ui.actionSidebar; // grr, copying not allowed!
    m_sidebarActionWithIcon->setText( tr( "My Profile" ) );
    m_sidebarActionWithIcon->setToolTip( tr( "Show/hide my profile" ) );
    #ifdef WIN32
    m_sidebarActionWithIcon->setCheckable( true );
    #endif
    connect( m_sidebarActionWithIcon, SIGNAL( triggered() ),
             this,                    SLOT( enableSidebar() ) );

    toolbar->addAction( m_sidebarActionWithIcon );
    toolbar->addSeparator();

    //m_recommendAction  = toolbar->addAction( QIcon( dataPath( "buttons/recommend.png" ) ), tr( "Recommend" ), this, SLOT( showRecommendDialog() ) );
    ui.actionRecommend->setIcon( QIcon( dataPath( "buttons/recommend.png" ) ) );
    ui.actionRecommend->setShortcut( tr( "Ctrl+R" ) );
    ui.actionRecommend->setToolTip( tr( "Recommend what's currently playing to someone" ) );
    toolbar->addAction( ui.actionRecommend );
    connect( ui.actionRecommend, SIGNAL( triggered() ),
             this,               SLOT( showRecommendDialog() ) );

    //m_tagAction = toolbar->addAction( QIcon( dataPath( "buttons/tag.png" ) ), tr( "Tag" ), this, SLOT( showTagDialog() ) );
    ui.actionTag->setIcon( QIcon( dataPath( "buttons/tag.png" ) ) );
    ui.actionTag->setShortcut( tr( "Ctrl+T" ) );
    ui.actionTag->setToolTip( tr( "Tag what's currently playing" ) );
    toolbar->addAction( ui.actionTag );
    connect( ui.actionTag, SIGNAL( triggered() ),
             this,         SLOT( showTagDialog() ) );

    toolbar->addSeparator();

    //m_loveAction = toolbar->addAction( QIcon( dataPath( "buttons/love.png" ) ), tr( "Love" ), this, SLOT( love() ) );
    ui.actionLove->setIcon( QIcon( dataPath( "buttons/love.png" ) ) );
    ui.actionLove->setShortcut( tr( "Ctrl+L" ) );
    ui.actionLove->setToolTip( tr( "Express your love for the track that's playing" ) );
    toolbar->addAction( ui.actionLove );
    connect( ui.actionLove, SIGNAL( triggered() ),
             this,          SLOT( love() ) );

    //m_banAction  = toolbar->addAction( QIcon( dataPath( "buttons/ban.png" ) ), tr( "Ban" ), this, SLOT( ban() ) );
    ui.actionBan->setIcon( QIcon( dataPath( "buttons/ban.png" ) ) );
    ui.actionBan->setShortcut( tr( "Ctrl+B" ) );
    ui.actionBan->setToolTip( tr( "Don't ever play me this again" ) );
    toolbar->addAction( ui.actionBan );
    connect( ui.actionBan, SIGNAL( triggered() ),
             this,         SLOT( ban() ) );

    //m_skipAction = new QAction( QIcon( dataPath( "buttons/skip.png" ) ), tr( "Skip" ), this );
    ui.actionSkip->setIcon( QIcon( dataPath( "buttons/skip.png" ) ) );
    ui.actionSkip->setShortcut( tr( "Ctrl+K" ) );
    ui.actionSkip->setToolTip( tr( "Skip track" ) );

    //m_playAction  = new QAction( QIcon( dataPath( "buttons/play.png" ) ), tr( "Play" ), this );
    ui.actionPlay->setIcon( QIcon( dataPath( "buttons/play.png" ) ) );
    ui.actionPlay->setShortcut( tr( "Ctrl+P" ) );
    ui.actionPlay->setToolTip( tr( "Start radio" ) );

    //m_stopAction  = new QAction( QIcon( dataPath( "buttons/stop.png" ) ), tr( "Stop" ), this );
    ui.actionStop->setIcon( QIcon( dataPath( "buttons/stop.png" ) ) );
    ui.actionStop->setShortcut( tr( "Ctrl+S" ) );
    ui.actionStop->setToolTip( tr( "Stop radio" ) );

    toolbar->addSeparator();

    toolbar->addAction( ui.actionPlay );
    toolbar->addAction( ui.actionStop );
    toolbar->addAction( ui.actionSkip );

#ifdef Q_WS_MAC
    //ui.actionSettings->setText( tr( "Options" ) );
#endif
    ui.actionSettings->setMenuRole( QAction::PreferencesRole );
    ui.actionSettings->setShortcut( tr( "Ctrl+O" ) );
    ui.actionToggleScrobbling->setShortcut( Qt::Key_Insert );

    ui.actionAboutLastfm->setMenuRole( QAction::AboutRole );

    QIcon scrobbleIcon( dataPath( "icons/scrobble16.png" ) );
    QString scrobbleTooltip = tr( "Click to enable/disable scrobbling" );
    m_scrobbleNormal = scrobbleIcon.pixmap( 16, 16, QIcon::Normal );
    m_scrobbleDisabled = scrobbleIcon.pixmap( 16, 16, QIcon::Disabled );
    m_scrobbleToggle.setImages( m_scrobbleNormal, m_scrobbleNormal, m_scrobbleNormal, m_scrobbleDisabled );
    m_scrobbleToggle.setToolTip( scrobbleTooltip );


    statusBar()->addPermanentWidget( &m_scrobbleLabel );

    #ifdef Q_WS_MAC
        QPalette lp = m_scrobbleLabel.palette();
        lp.setColor( QPalette::Text, QColor( 0x59, 0x59, 0x59 ) );
        lp.setColor( QPalette::WindowText, QColor( 0x59, 0x59, 0x59 ) );
        m_scrobbleLabel.setPalette( lp );
        QPixmap scrobbleGraphicMac( dataPath( "icons/scrobbling_graphic.png" ) );
        m_scrobbleGraphicMac.setImages( scrobbleGraphicMac, scrobbleGraphicMac, scrobbleGraphicMac, scrobbleGraphicMac );
        m_scrobbleGraphicMac.setToolTip( scrobbleTooltip );
        statusBar()->addPermanentWidget( &m_scrobbleGraphicMac );
    #endif

    statusBar()->addPermanentWidget( &m_scrobbleToggle );

    ui.actionToggleScrobbling->setCheckable( true );
    ui.actionToggleDiscoveryMode->setCheckable( true );

    // <PlayControls>
    QWidget *playControls = new QWidget( toolbar );
    ui_playcontrols.setupUi( playControls );

    QAction *playControlsAction = toolbar->addWidget( playControls );
    playControlsAction->setVisible( true );

    // </PlayControls>
    // </ToolBar>

    #ifndef Q_WS_MAC
        QMenu* fileMenu = new QMenu( tr( "&File" ), this );
        QAction* a = fileMenu->addAction( tr( "&Send to Tray" ), this, SLOT( close() ) );
        a->setToolTip( tr( "Send application to system tray" ) );
        fileMenu->addSeparator();
        fileMenu->addAction( tr( "E&xit" ), this, SLOT( actualClose() ) );
        menuBar()->insertMenu( menuBar()->actions().at( 0 ), fileMenu );
    #endif

    ui.actionTag->setEnabled( false );
    ui.actionRecommend->setEnabled( false );
    ui.actionLove->setEnabled( false );
    ui.actionBan->setEnabled( false );
    ui.actionSkip->setEnabled( false );
    ui.actionStop->setEnabled( false );
    ui.actionStop->setVisible( false );
    ui.actionPlay->setEnabled( false );

    ui.tabWidget->setAcceptDrops( false );

    #ifdef Q_WS_MAC
		QApplication::setStyle( new MacStyleOverrides() );

        QPalette p = ui.centralwidget->palette();
        p.setColor( QPalette::Window, QColor( 0xe9, 0xe9, 0xe9 ) );
        p.setColor( QPalette::Base, QColor( 0xe9, 0xe9, 0xe9 ) );
        ui.centralwidget->setPalette( p );
		p.setColor( QPalette::Button, QColor( 0xe9, 0xe9, 0xe9 ) );
		
		// Uncomment this to get a gray toolbar
		//toolbar->setPalette( p );

        // Hacky hackety McHackerson
        QLinearGradient grad( 0.5, 0.0, 0.5, 20.0 );
        grad.setColorAt( 0.0, QColor( 0x8c, 0x8c, 0x8c ) );
        grad.setColorAt( 0.05, QColor( 0xf7, 0xf7, 0xf7 ) );
        grad.setColorAt( 1.0, QColor( 0xe4, 0xe4, 0xe4 ) );
        QBrush b( grad );
        p = statusBar()->palette();
        p.setBrush( QPalette::Window, b );
        p.setBrush( QPalette::Base, b );
        p.setColor( QPalette::Text, QColor( 0x59, 0x59, 0x59 ) );
        p.setColor( QPalette::WindowText, QColor( 0x59, 0x59, 0x59 ) );
        statusBar()->setPalette( p );

        QFont f = statusBar()->font();
        f.setPixelSize( k_statusBarFontSizeMac );
        statusBar()->setFont( f );
        m_scrobbleLabel.setFont( f );

        ui.searchTab->layout()->setMargin( 0 );
        ui.metadataTab->layout()->setMargin( 0 );

    #elif defined WIN32

        // This is in order to remove the borders around the status bar widgets
        QString styleName = QApplication::style()->objectName();
        if( styleName == "windowsxp" )
        {
            statusBar()->setStyle( new WinXPStyleOverrides() );
        }
        else
        {
            statusBar()->setStyle( new WinStyleOverrides() );

            ui.searchTab->layout()->setMargin( 1 );
            ui.metadataTab->layout()->setMargin( 1 );
        }

        // The splitter is wider on Windows so compensate for this
        ui.spacerItem->changeSize( 3, 20, QSizePolicy::Fixed, QSizePolicy::Fixed );

    #endif

    m_closeTabButton = new QToolButton( ui.tabWidget );
    m_closeTabButton->setIcon( QIcon( dataPath( "buttons/closebutton.png" ) ) );
    m_closeTabButton->setContentsMargins( 4, 4, 6, 4 );
    m_closeTabButton->setAutoRaise( true );
    m_closeTabButton->setEnabled( false );
    m_closeTabButton->setVisible( false );

    ui.tabWidget->setCornerWidget( m_closeTabButton );
    //ui.tabWidget->setTabEnabled( 1, false );

    ui.songTimeBar->setEnabled( false );
    ui.songTimeBar->setItemType( ItemTrack );
    connect( this,           SIGNAL( newSong( MetaData ) ),
             ui.songTimeBar, SLOT( setNewTrack( const MetaData& ) ) );

    ui.stationTimeBar->setEnabled( false );
    ui.stationTimeBar->setVisible( false );
    ui.stationTimeBar->setBackgroundEnabled( false );
    ui.stationTimeBar->setItemType( ItemStation );

    ui.splitter->setCollapsible( 0, false );
    ui.splitter->setCollapsible( 1, false );
    ui.splitter->setStretchFactor( 0, 0 );
    ui.splitter->setStretchFactor( 1, 1 );


/* This just sets the background to the side of the tabs
    QPalette p = ui.tabWidget->palette();
    p.setColor( QPalette::Window, QColor( 0xff, 0xff, 0xff ) );
    ui.tabWidget->setPalette( p );
*/

    if ( !m_iSettings->splitterState().isNull() &&
         ui.splitter->restoreState( m_iSettings->splitterState() ) )
    {
        // If we have splitter settings and restoreState worked,
        // cache state in m_sidebarState
        m_sidebarState = m_iSettings->splitterState();
    }
    else
    {
        // It's the first run, set splitter to some default values
        // and save the state
        QList<int> widths;
        widths << 190; // default width for sidebar
        ui.splitter->setSizes( widths );
        #ifdef Q_WS_MAC
            ui.splitter->setHandleWidth( 1 );
        #endif
        m_sidebarState = ui.splitter->saveState();
    }

    // Needed for upgraded versions to get rid of the old 5px splitter
    #ifdef Q_WS_MAC
        ui.splitter->setHandleWidth( 1 );
    #endif

    restoreState( m_iSettings->appState(), CONTAINER_VERSION );

    bool resetPos = false;
    QPoint pos = m_iSettings->position();
    if ( !pos.isNull() )
    {
        LOGL( 3, "Saved pos is not null, pos: " << pos.x() << ", " << pos.y() );

        // Only move the window to its saved position if it's within
        // our screen bounds.
        QDesktopWidget* desk = QApplication::desktop();
        int maxX = 0;
        int maxY = 0;
        for ( int i = 0; i < desk->numScreens(); ++i )
        {
            if ( desk->availableGeometry( i ).right() > maxX )
            {
                maxX = desk->availableGeometry( i ).right();
            }
            if ( desk->availableGeometry( i ).bottom() > maxY )
            {
                maxY = desk->availableGeometry( i ).bottom();
            }
        }

        LOGL( 3, "maxX: " << maxX << ", maxY: " << maxY );

        if ( pos.x() < maxX && pos.y() < maxY && pos.x() > 0 && pos.y() > 0 )
        {
            LOGL( 3, "pos within screen boundaries, moving to pos" );
            move( pos );
        }
        else
        {
            LOGL( 3, "pos outside screen boundaries, resetting pos" );
            resetPos = true;
        }
    }
    else
    {
        LOGL( 3, "Saved pos is null" );
        resetPos = true;
    }

    if ( resetPos )
    {
        LOGL( 3, "Moving pos to 50, 50" );
        move( QPoint( 50, 50 ) );
    }

    if ( m_iSettings->size().isValid() )
    {
        resize( m_iSettings->size() );
    }
    else
    {
        // default size
        resize( QSize( 580, 540 ) );
    }

    initTray();

    // GUI slots
    connect( m_closeTabButton,              SIGNAL( released() ),
             this,                            SLOT( closeCurrentTab() ) );
    connect( ui.tabWidget,                  SIGNAL( currentChanged( int ) ),
             this,                            SLOT( currentTabChanged( int ) ) );
    connect( ui.actionDashboard,            SIGNAL( triggered() ),
             this,                            SLOT( gotoProfile() ) );
    connect( ui.actionSettings,             SIGNAL( triggered() ),
             this,                            SLOT( showSettingsDialog() ) );
    connect( ui.actionGetPlugin,            SIGNAL( triggered() ),
             this,                            SLOT( getPlugin() ) );
    connect( ui.actionCheckForUpdates,      SIGNAL( triggered() ),
             this,                            SLOT( checkForUpdates() ) );
    connect( ui.actionAddUser,              SIGNAL( triggered() ),
             this,                            SLOT( addUser() ) );
    connect( ui.actionDeleteUser,           SIGNAL( triggered() ),
             this,                            SLOT( deleteUser() ) );
    connect( ui.actionToggleScrobbling,     SIGNAL( triggered() ),
             this,                            SLOT( toggleScrobbling() ) );
    connect( ui.actionToggleDiscoveryMode,  SIGNAL( triggered() ),
             this,                            SLOT( toggleDiscoveryMode() ) );
    connect( ui.actionAboutLastfm,          SIGNAL( triggered() ),
             this,                            SLOT( about() ) );
    connect( ui.splitter,                   SIGNAL( splitterMoved( int, int ) ),
             this,                            SLOT( splitterMoved() ) );
    connect( ui.menuUser,                   SIGNAL( aboutToShow() ),
             this,                            SLOT( showUserMenu() ) );
    connect( ui.menuUser,                   SIGNAL( triggered( QAction* ) ),
             this,                            SLOT( userSelected( QAction* ) ) );
    connect( ui.actionSkip,                  SIGNAL( triggered() ),
             this,                            SLOT( skip() ) );
    connect( ui.actionStop,                  SIGNAL( triggered() ),
             this,                            SLOT( stop() ) );
    connect( ui.actionPlay,                  SIGNAL( triggered() ),
             this,                            SLOT( play() ) );
    connect( ui_playcontrols.volume,        SIGNAL( valueChanged( int ) ),
             m_iPlayback,                     SLOT( setVolume( int ) ) );
    connect( &m_scrobbleToggle,              SIGNAL( clicked() ),
             this,                            SLOT( toggleScrobbling() ) );

    #ifdef Q_WS_MAC
        connect( &m_scrobbleGraphicMac,          SIGNAL( clicked() ),
                 this,                            SLOT( toggleScrobbling() ) );
    #endif

    // Services & Extensions
    connect( m_iWebService, SIGNAL( loveDone( QString, QString ) ),
             this,            SLOT( loveDone() ) );
    connect( m_iWebService, SIGNAL( unloveDone( QString, QString ) ),
             this,            SLOT( unloveDone() ) );
    connect( m_iWebService, SIGNAL( banDone( QString, QString ) ),
             this,            SLOT( banDone() ) );
    connect( m_iWebService, SIGNAL( unbanDone( QString, QString ) ),
             this,            SLOT( unbanDone() ) );
    connect( m_iWebService, SIGNAL( recommendDone( QString, QString, QString ) ),
             this,            SLOT( recommendDone() ) );
    connect( m_iWebService, SIGNAL( tagDone( QString, QString, QString, bool ) ),
             this,            SLOT( tagDone( QString, QString, QString, bool ) ) );
    connect( m_iWebService, SIGNAL( handshakeResult( int ) ),
             this,            SLOT( handshakeResult( int ) ) );
    connect( m_iWebService, SIGNAL( metaDataResultStation( QString, QString, bool ) ),
             this,            SLOT( setNewStation( QString, QString, bool ) ) );
    connect( m_iWebService, SIGNAL( radioStatusChanged( QString ) ),
             this,            SLOT( setRadioStatus( QString ) ) );
    connect( m_iWebService, SIGNAL( trackStatusChanged( QString ) ),
             this,            SLOT( setTrackStatus( QString ) ) );
    connect( m_iWebService, SIGNAL( stationChangeFailed( QString ) ),
             this,            SLOT( displayStationChangeError( QString ) ) );

    connect( m_iInput,      SIGNAL( bufferFull() ),
             this,            SLOT( bufferFull() ) );
    connect( m_iInput,      SIGNAL( bufferEmpty( int ) ),
             this,            SLOT( bufferEmpty( int ) ) );
    connect( m_iInput,      SIGNAL( requestedStreaming() ),
             this,            SLOT( requestedStreaming() ) );
    connect( m_iInput,      SIGNAL( preparedStreaming() ),
             this,            SLOT( preparedStreaming() ) );
    connect( m_iInput,      SIGNAL( startedStreaming() ),
             this,            SLOT( startedStreaming() ) );
    connect( m_iInput,      SIGNAL( stoppedStreaming() ),
             this,            SLOT( stoppedStreaming() ) );
    connect( m_iInput,      SIGNAL( error( int, QString ) ),
             this,            SLOT( streamingResult( int, QString ) ), Qt::QueuedConnection /* we show a dialog, don't deadlock! */ );

    connect( m_iSettings,   SIGNAL( currentUserChanged( UserSettingsServiceInterface* ) ),
             this,            SLOT( updateUserStuff( UserSettingsServiceInterface* ) ) );
    connect( m_iSettings,   SIGNAL( reconnect() ),
             this,            SLOT( initServices() ) );

    // Decoding pipe
    connect( m_iInput,      SIGNAL( preparedStreaming() ),    m_iPlayback,  SLOT( startPlayback() ) );
    connect( m_iInput,      SIGNAL( stoppedStreaming() ),     m_iPlayback,  SLOT( stopPlayback() ) );
    connect( m_iInput,      SIGNAL( clearBuffers() ),         m_iPlayback,  SLOT( clearBuffers() ) );
    connect( m_iInput,      SIGNAL( clearBuffers() ),         m_iTranscode, SLOT( clearBuffers() ) );

    connect( m_iInput,      SIGNAL( newData( QByteArray ) ),  m_iTranscode, SLOT( dataAvailable( QByteArray ) ) );
    connect( m_iTranscode,  SIGNAL( newData( QByteArray ) ),  m_iPlayback,  SLOT( dataAvailable( QByteArray ) ) );
    connect( m_iPlayback,   SIGNAL( needData() ),             m_iInput,     SLOT( requestData() ) );

    // Auto updater
    connect( &m_updater, SIGNAL( updateCheckDone( bool, bool, QString ) ),
             this,         SLOT( updateCheckDone( bool, bool, QString ) ) );

    // Scrobbler
    connect( &m_scrobbler, SIGNAL( statusChanged( int, QString ) ),
             this,           SLOT( setScrobblerStatus( int, QString ) ) );

    // update volume _after_ signals have been connected
    ui_playcontrols.volume->setValue( m_iSettings->volume() );

#ifdef HIDE_RADIO
    ui_playcontrols.volume->setVisible( false );

    ui.actionPlay->setVisible( false );
    ui.actionStop->setVisible( false );
    ui.actionSkip->setVisible( false );
    ui.actionBan->setVisible( false );
    ui.actionToggleDiscoveryMode->setVisible( false );
    ui.tabWidget->removeTab( 0 );
#endif

#ifndef WIN32
    ui.actionGetPlugin->setEnabled( false );
    ui.actionGetPlugin->setVisible( false );
#endif

#ifdef Q_WS_MAC
    new ITunesScript( this, m_listener );
#endif

    if ( qApp->arguments().contains( "--debug" ) )
    {
        // Add crash option to menu!
        ui.menuHelp->addAction( "Crash please", this, SLOT( crash() ) );
    }
}


void
Container::initServices()
{
    // We now have a current user
    QString username = m_iSettings->currentUser()->username();
    QString password = m_iSettings->currentUser()->password();
    QString version = m_iSettings->version();

    m_sidebarEnabled = !m_iSettings->currentUser()->sidebarEnabled();
    //ui.actionSidebar->setChecked( m_sidebarEnabled );
    //m_sidebarActionWithIcon->setChecked( m_sidebarEnabled );
    //m_sidebarEnabled = !m_sidebarEnabled;

    enableSidebar();
    updateUserStuff( m_iSettings->currentUser() );
    statusBar()->showMessage( tr( "Connecting to server..." ) );

    m_iWebService->handshake( username, password, version );
    m_scrobbler.init( username, password, version );

    if ( m_iInput->isStreaming() )
        m_iInput->stopStreaming();
}


Container::~Container()
{
    delete m_iInput;
    delete m_iTranscode;
    delete m_iPlayback;

    delete m_iWebService;
    delete m_iSettings;

    delete m_sysTray;
}


Container&
Container::instance()
{
    LOGL( 3, "" );
    QMutexLocker lock( &m_mutex );

    if ( !mpInstance )
    {
        LOGL( 3, "Calling ctor" );
        mpInstance = new Container( 0 );
    }

    LOGL( 3, "Returning from instance" );
    return *mpInstance;
}


ScrobblerSubmitter&
Container::scrobbler()
{
    return m_scrobbler;
}


void
Container::destroy()
{
    delete mpInstance;
}


void
Container::initListener()
{
    connect( m_iInput,    SIGNAL( listenerCommand( TrackInfo, EPlayerCommand ) ),
             &m_listener, SLOT  ( handleRadioCommand( TrackInfo, EPlayerCommand ) ) );

    connect( &m_listener, SIGNAL( trackChanged( TrackInfo ) ),
             this,        SLOT  ( setNewSong( TrackInfo ) ), Qt::QueuedConnection );

    connect( &m_listener, SIGNAL( exceptionThrown( QString ) ),
             this,        SLOT  ( listenerException( QString ) ) );

    // Start listener worker thread
    m_listener.start();
}


void
Container::dragMoveEvent( QDragMoveEvent* event )
{
#ifndef HIDE_RADIO

    if ( event->mimeData()->urls().count() > 0 )
    {
        if ( event->mimeData()->urls().at( 0 ).toString().startsWith( "lastfm://" ) )
        {
            event->acceptProposedAction();
            return;
        }
    }

    event->ignore();

#endif // HIDE_RADIO
}


void
Container::dragEnterEvent( QDragEnterEvent* event )
{
#ifndef HIDE_RADIO

    if ( event->mimeData()->urls().count() > 0 )
    {
        if ( event->mimeData()->urls().at( 0 ).toString().startsWith( "lastfm://" ) )
        {
            event->acceptProposedAction();
            return;
        }
    }

    event->ignore();

#endif // HIDE_RADIO
}


void
Container::dropEvent( QDropEvent* event )
{
#ifndef HIDE_RADIO

    if ( event->mimeData()->urls().count() > 0 )
    {
        if ( event->mimeData()->urls().at( 0 ).toString().startsWith( "lastfm://" ) )
        {
            m_iInput->load( event->mimeData()->urls().at( 0 ).toString() );
        }
    }

#endif // HIDE_RADIO
}


void
Container::closeEvent( QCloseEvent *event )
{
    LOG( 3, "closeEvent\n" );

    qDebug() << "closeEvent:" << event->spontaneous();
    #ifdef Q_WS_MAC
    if ( !event->spontaneous() )
    {
        actualClose();
        event->ignore();
        return;
    }
    #endif

    // Just minimise to tray
    minimiseToTray();
    event->ignore();
}


void
Container::actualClose( bool confirm )
{
    bool okToClose = true;

    qDebug() << "actualClose";
    qDebug() << "confirm:" << confirm;

    if ( confirm )
    {
        QString confirmMsg = tr( "Really quit Last.fm? Any music you listen to will not be scrobbled "
                                "to your profile." );
        ConfirmDialog dlg( ConfirmDialog::Quit, confirmMsg, this );
        okToClose = dlg.exec() == QDialog::Accepted;
    }

    qDebug() << "okToClose:" << okToClose;

    if ( okToClose )
    {
        // This allows any extensions (i.e. Skype) to clear its stuff. Bit of a hack.
        emit newSong( MetaData() );

        LOG( 3, "Saving appState\n" );
        m_iSettings->setPosition( pos() );
        m_iSettings->setSize( size() );
        m_iSettings->setSplitterState( ui.splitter->saveState() );
        m_iSettings->setAppState( saveState( CONTAINER_VERSION ) );
        m_iSettings->setVolume( ui_playcontrols.volume->value() );
        m_iSettings->currentUser()->setSidebarEnabled( m_sidebarEnabled );

        LOG( 3, "Saving config\n" );
        m_iSettings->save();

        LOG( 3, "Shutting down listener\n" );
        m_listener.Stop();

        LOG( 3, "Calling exit\n" );
        QApplication::exit( 0 );
    }
}


void
Container::minimiseToTray()
{
    hide();

#ifdef WIN32
    // Do animation and fail gracefully if not possible to find systray
    RECT rectFrame;    // animate from
    RECT rectSysTray;  // animate to

    ::GetWindowRect( (HWND)winId(), &rectFrame );

    // Get taskbar window
    HWND taskbarWnd = ::FindWindow( L"Shell_TrayWnd", NULL );
    if ( taskbarWnd == NULL )
        return;

    // Use taskbar window to get position of tray window
    HWND trayWnd = ::FindWindowEx( taskbarWnd, NULL, L"TrayNotifyWnd", NULL );
    if ( trayWnd == NULL )
        return;

    ::GetWindowRect( trayWnd, &rectSysTray );
    ::DrawAnimatedRects( (HWND)winId(), IDANI_CAPTION, &rectFrame, &rectSysTray );

    // Make it release memory as when minimised
    HANDLE h = ::GetCurrentProcess();
    SetProcessWorkingSetSize( h, -1 ,-1 );
#endif // WIN32
}


void
Container::loadServices()
{
    m_iSettings = settingsService();
    m_iWebService = webService();
    m_iInput = inputService();
    m_iTranscode = transcodeService();
    m_iPlayback = playbackService();
}


void
Container::loadExtensions()
{
    #ifdef QT_NO_DEBUG
        QDir extensionsDir = QDir( qApp->applicationDirPath() + "/extensions" );
    #else
        QDir extensionsDir = QDir( qApp->applicationDirPath() + "/extensions/debug" );
    #endif

    foreach ( QString fileName, extensionsDir.entryList( QDir::Files ) )
    {
        if ( !fileName.startsWith( EXTENSION_PREFIX ) || !QLibrary::isLibrary( extensionsDir.absoluteFilePath( fileName ) ) )
            continue;

        LOGL( 3, "Trying to load extension: " << fileName );
        QObject* plugin = QPluginLoader( extensionsDir.absoluteFilePath( fileName ) ).instance();
        if ( plugin )
        {
            LOGL( 3, "Extension loaded." );
            ExtensionInterface *iExtension = qobject_cast<ExtensionInterface *>( plugin );
            if ( iExtension )
            {
                if ( iExtension->name() == "MetaData Extension" )
                {
                    iExtension->setOwner( ui.metadataFrame );
                    ui.metadataFrame->layout()->addWidget( iExtension->gui() );

                    VERIFY( connect( this,          SIGNAL( newSong( MetaData ) ),
                                     iExtension,    SLOT( setMetaData( MetaData ) ) ) );

                    VERIFY( connect( iExtension,     SIGNAL( trackModerated( const MetaData& ) ),
                                     ui.songTimeBar, SLOT( setNewTrack( const MetaData& ) ) ) );

                    VERIFY( connect( iExtension,     SIGNAL( tagButtonClicked() ),
                                     this,           SLOT( showTagDialogMD() ) ) );

                    VERIFY( connect( iExtension,  SIGNAL( urlHovered( const QString& ) ),
                                     this,        SLOT( displayUrlInStatusBar( const QString& ) ) ) );

                    VERIFY( connect( m_iInput,    SIGNAL( requestedStreaming() ),
                                     iExtension,  SLOT( displayTuningIn() ) ) );
                    VERIFY( connect( m_iInput,    SIGNAL( stoppedStreaming() ),
                                     iExtension,  SLOT( displayNotListening() ) ) );

                    LOGL( 3, "Added metadata extension to GUI." );
                }

                else if ( iExtension->name() == "Sidebar Extension" )
                {
                    iExtension->setOwner( ui.sidebarFrame );
                    ui.sidebarFrame->layout()->addWidget( iExtension->gui() );

                    connect( iExtension,    SIGNAL( showFriendsTab() ),
                             this,            SLOT( showFriendsTab() ) );
                    connect( iExtension,    SIGNAL( stationFocused( QString ) ),
                             this,            SLOT( setLastFocusedStation( QString ) ) );
                    connect( iExtension,    SIGNAL( statusMessage( QString ) ),
                             this,            SLOT( statusMessage( QString ) ) );

                    LOGL( 3, "Added sidebar extension to GUI." );
                }

                else if ( iExtension->name() == "Search Extension" )
                {
                    iExtension->setOwner( ui.searchFrame );
                    ui.searchFrame->layout()->addWidget( iExtension->gui() );

                    connect( iExtension,    SIGNAL( stationFocused( QString ) ),
                             this,            SLOT( setLastFocusedStation( QString ) ) );
                    connect( this,          SIGNAL( playStarted() ),
                             iExtension,      SLOT( clearSelection() ) );
                    connect( m_iSettings,   SIGNAL( userSwitch( UserSettingsServiceInterface* ) ),
                             iExtension,      SLOT( clearSelection() ) );
                }

                else if ( iExtension->name() == "UserInfo Extension" )
                {
                    m_eUserInfo = iExtension;
                }

                else if ( fileName.contains( "notify" ) )
                {
                    // Add it to Options
                    LOGL( 3, "Loaded a notify extension" );
                    m_settingsDialog.addExtension( iExtension );

                    connect( this, SIGNAL( newSong( MetaData ) ), 
                             iExtension, SLOT( notify( MetaData ) ) );
                }

                ui.tabWidget->setCurrentIndex( 0 );
            }
        } // endif plugin
        else
        {
            LOGL( 1, "Failed to load " << fileName );
        }
    }
}


void
Container::addExtensionToTabWidget( ExtensionInterface* extension )
{
    // check if we're already shown
    if ( !extension->guiEnabled() )
    {
        // dynamically allocate new tab
        QFrame* frame = new QFrame();
        QBoxLayout* layout = new QBoxLayout( QBoxLayout::TopToBottom, frame );
        extension->setOwner( frame );
        frame->layout()->addWidget( extension->gui() );
    }

    ui.tabWidget->addTab( extension->owner(), extension->tabCaption() );
    ui.tabWidget->setCurrentWidget( extension->owner() );

    LOGL( 3, "Added " << extension->name() << "extension to GUI." );
}


void
Container::currentTabChanged( int index )
{
    if ( index > 1 )
    {
        m_closeTabButton->setVisible( true );
        m_closeTabButton->setEnabled( true );
    }
    else
    {
        m_closeTabButton->setEnabled( false );
        m_closeTabButton->setVisible( false );
    }
}


void
Container::handshakeResult( int result )
{
    switch ( result )
    {
        case -1:
        {
            #ifndef HIDE_RADIO
            LOGL( 1, "Radio handshake failed: Could not connect to server." );
            statusBar()->showMessage( tr( "Couldn't connect to radio streaming server." ) );
            #endif // HIDE_RADIO

            break;
        }

        case 0:
        {
            #ifndef HIDE_RADIO
            QMessageBox::information( this, tr( "Error" ), tr( "Could not connect to server: Wrong username or password." ) );
            LOGL( 1, "Radio handshake failed: Wrong username or password." );
            statusBar()->showMessage( tr( "Couldn't connect to server: Wrong username or password." ) );

            showSettingsDialog();
            #endif // HIDE_RADIO

            break;
        }

        case 1:
        {
            #ifndef HIDE_RADIO
            LOGL( 3, "Radio streamer handshake successful." );
            statusBar()->showMessage( QString( tr( "Radio streamer up and running." ) ) );

            ui.actionPlay->setEnabled( true );
            ui.actionPlay->setVisible( true );
            ui.actionStop->setVisible( false );

            // do we have to load a station since we got started with a cli-arg?
            m_controlInterface->setHandshaked();
            if ( m_preloadStation.contains( "lastfm://" ) )
                m_iInput->load( m_preloadStation );
            else
                if ( m_iSettings->currentUser()->resumePlayback() && !m_iSettings->currentUser()->resumeStation().isEmpty() )
                    m_iInput->load( m_iSettings->currentUser()->resumeStation() );
            #endif // HIDE_RADIO

            if ( !m_iSettings->firstRun() )
            {
                // Don't do this until here as we won't have received a base host
                // earlier.
                checkForUpdates( false );
            }

            break;
        }
    }
}

void
Container::displayUrlInStatusBar( const QString& url )
{
    QUrl realUrl( QUrl::fromEncoded( url.toLatin1() ) );
    QString msg = realUrl.toString();

    if ( msg != statusBar()->currentMessage() )
    {
        statusBar()->showMessage( msg );
    }
}

void
Container::streamingResult( int error, const QString& reason )
{
#ifndef HIDE_RADIO

    LOGL( 3, "Streamer returned HTTP code " << error << ": " << reason );
    QMessageBox::information( this, tr( "Radio Error" ), reason );

    stoppedStreaming();

#endif // HIDE_RADIO
}


void
Container::enableSidebar()
{
    UserSettingsServiceInterface* user = m_iSettings->currentUser();
    if ( m_sidebarEnabled )
    {
        m_sidebarEnabled = false;
        ui.actionSidebar->setChecked( false );
        m_sidebarActionWithIcon->setChecked( false );

        QPixmap pix = IconShack::instance().GetGoodUserIconExpanded( user->icon() );
        m_sidebarActionWithIcon->setIcon( QIcon( pix ) );

        m_sidebarState = ui.splitter->saveState();

        QList<int> l;
        l << 0;
        ui.splitter->setCollapsible( 0, true );
        ui.splitter->setSizes( l );
    }
    else
    {
        m_sidebarEnabled = true;
        ui.actionSidebar->setChecked( true );
        m_sidebarActionWithIcon->setChecked( true );

        QPixmap pix = IconShack::instance().GetGoodUserIconCollapsed( user->icon() );
        m_sidebarActionWithIcon->setIcon( QIcon( pix ) );

        ui.splitter->restoreState( m_sidebarState );
        ui.splitter->setCollapsible( 0, false );

        //FIXME (can't remember why this is done)
        if ( ui.splitter->sizes().at( 0 ) < 190 )
        {
            QList<int> l;
            l << 190;
            ui.splitter->setSizes( l );
        }
    }

    m_iSettings->currentUser()->setSidebarEnabled( m_sidebarEnabled );
}


void
Container::splitterMoved()
{
    if ( ui.splitter->sizes().at( 0 ) > 0 && !m_sidebarEnabled )
    {
        m_sidebarEnabled = true;
        ui.actionSidebar->setChecked( true );
        m_sidebarActionWithIcon->setChecked( true );
    }

    if ( ui.splitter->sizes().at( 0 ) <= 0 && m_sidebarEnabled )
    {
        m_sidebarEnabled = false;
        ui.actionSidebar->setChecked( false );
        m_sidebarActionWithIcon->setChecked( false );
    }
}


void
Container::getPlugin()
{
    ConfigWizard( this, ConfigWizard::Plugin ).exec();
}


void
Container::addMediaDevice( const QString& uid )
{
    ConfigWizard( NULL, ConfigWizard::MediaDevice, uid ).exec();
}


void
Container::checkForUpdates( bool invokedByUser )
{
    bool userAllowed = true;
    #ifdef WIN32
        userAllowed = !CWinUtils::IsLimitedUser();
    #endif

    m_userCheck = invokedByUser;
    if ( m_userCheck )
    {
        if ( userAllowed )
        {
            QApplication::setOverrideCursor( QCursor( Qt::WaitCursor ) );
        }
        else
        {
            QMessageBox::information(
                this,
                tr( "Update Error" ),
                tr( "An update installs file on your computer, therefore you "
                    "must be logged in as an Administrator to do it. Please "
                    "get an Administrator to run this for you." ),
                QMessageBox::Ok);
        }
    }

    if ( userAllowed )
    {
        m_updater.checkForUpdates();
    }

}


void
Container::updateCheckDone( bool updatesAvailable, bool error, QString errorMsg )
{
    QApplication::restoreOverrideCursor();

    if ( error )
    {
        // Can't connect to the internet
        LOG( 2, "Update check failed. Error: " << errorMsg << "\n" );
        if ( m_userCheck )
        {
            QMessageBox::critical( this, tr( "Connection Problem" ),
                tr( "Last.fm couldn't connect to the Internet to check "
                    "for updates.\n\nError: %1" ).arg( errorMsg ) );
        }
        return;
    }

    // No connection error, let's see if we have updates
    if ( !updatesAvailable )
    {
        LOG( 3, "Update check said no updates available.\n" );
        if ( m_userCheck )
        {
            QMessageBox::information( this, tr( "Up To Date" ),
                tr( "No updates available. All your software is up to date!\n" ) );
        }
        return;
    }

    // Go ahead and launch update wizard
    LOG( 3, "New updates available. Launching update wizard.\n" );

    UpdateWizard wiz( m_updater, this );
    wiz.exec();
}

void
Container::listenerException(
    QString msg )
{
    // Can't do much else than fire up a dialog box of doom at this stage
    QMessageBox::critical( this, tr( "Plugin Listener Error" ), msg );
}


void
Container::about()
{
    AboutDialog( this ).exec();
}


void
Container::initTray()
{
    m_sysTray = new SysTray( this );
    //m_sysTray->setUser( m_iSettings->currentUser() );
    //m_sysTray->setStatus( true );
    m_sysTray->show();

    // Connect user change to tray
    connect( m_iSettings, SIGNAL( currentUserChanged( UserSettingsServiceInterface* ) ),
             m_sysTray,   SLOT  ( setUser( UserSettingsServiceInterface* ) ) );
    connect( this,        SIGNAL( newSong( MetaData ) ),
             m_sysTray,   SLOT  ( setTrack( MetaData ) ) );
}


void
Container::showUserMenu()
{
    // Work out whether we have old items to delete
    QList<QAction*> actions = ui.menuUser->actions();
    if ( actions.size() != 3 )
    {
        // Delete old items
        for ( int i = actions.size() - 1; i >= 3; --i )
        {
            ui.menuUser->removeAction( actions.at( i ) );
        }
    }

    // For each user, add them and put a check next to the current
    QString currentUser = m_iSettings->currentUsername();
    QStringList users = m_iSettings->allUsers();
    for ( int i = 0; i < users.size(); ++i )
    {
        QString username = users.at( i );
        QAction* action = new QAction( ui.menuUser );
        action->setData( "user" );
        action->setText( username );
        action->setCheckable( true );
        if ( username == currentUser )
        {
            action->setChecked( true );
        }

        ui.menuUser->addAction( action );
    }
}


void
Container::userSelected( QAction* action )
{
    if ( action->data().toString() == "user" )
    {
        QString username = action->text();
        UserSettingsServiceInterface* user = m_iSettings->user( username );

        LOGL( 3, "Switching user to: " << user->username() );

        if ( !user->rememberPass() )
        {
            LoginDialog( this, LoginWidget::LOGIN, username ).exec();
        }
        else
        {
            m_iSettings->setCurrentUsername( username );
            m_iSettings->save( true, false );
        }
    }
}


void
Container::addUser()
{
    if ( LoginDialog( this, LoginWidget::ADD_USER ).exec() == QDialog::Accepted )
        ui.tabWidget->setCurrentIndex( 0 );
}


void
Container::deleteUser()
{
    DeleteUserDialog( this ).exec();
}


void
Container::toggleScrobbling()
{
    UserSettingsServiceInterface* user = m_iSettings->currentUser();
    user->setLogToProfile( !user->logToProfile() );
}


void
Container::updateUserStuff( UserSettingsServiceInterface* user )
{
    setWindowTitle( tr( "Last.fm %1 %2" )
        .arg( QChar(8211) ) // en dash
        .arg( user->username() ) );

    QPixmap pix = m_sidebarEnabled ?
        IconShack::instance().GetGoodUserIconCollapsed( user->icon() ) :
        IconShack::instance().GetGoodUserIconExpanded( user->icon() );
    m_sidebarActionWithIcon->setIcon( QIcon( pix ) );

    bool enabled = user->logToProfile();
    ui.songTimeBar->setEnabled( enabled );
    m_iWebService->enableScrobbling( enabled );

    //m_scrobbleAction->setChecked( enabled );
    ui.actionToggleScrobbling->setChecked( enabled );
    m_scrobbleLabel.setText( 
        QString( " " + tr( "Scrobbling %1" ) ).arg( enabled ? tr( "on" ) : tr( "off" ) ) + " " );

    if ( enabled )
    {
        m_scrobbleToggle.setImages( m_scrobbleNormal, m_scrobbleNormal, m_scrobbleNormal, m_scrobbleDisabled );
    }
    else
    {
        m_scrobbleToggle.setImages( m_scrobbleDisabled, m_scrobbleDisabled, m_scrobbleDisabled, m_scrobbleDisabled );
    }

}


void
Container::bufferFull()
{
#ifndef HIDE_RADIO
    ui.songTimeBar->resume();
    ui.stationTimeBar->resume();
    statusBar()->showMessage( tr( "Radio playback running." ) );
#endif // HIDE_RADIO
}


void
Container::bufferEmpty( int size )
{
#ifndef HIDE_RADIO
    // If this buffer wasn't big enough, we should try to increase it
    if ( size == 0 )
        m_iInput->setBufferSize( m_iInput->bufferSize() + 4096 );

    //qDebug() << size << "/" << m_iInput->bufferSize();
    ui.songTimeBar->pause();
    ui.stationTimeBar->pause();
    statusBar()->showMessage( QString( tr( "Buffering... (%1%2)" ) )
                             .arg( (100 * size) / (m_iInput->bufferSize() + 1) )
                             .arg( "%" ) );
#endif // HIDE_RADIO
}


void
Container::setNewSong( TrackInfo data )
{
    m_metaData.populate( data );

    if ( data.artist().isEmpty() && data.track().isEmpty() )
    {
        // Playback stopped
        ui.songTimeBar->clear();
        ui.songTimeBar->setEnabled( false );
        ui.songTimeBar->setTimerText( "" );

        ui.stationTimeBar->clear();
        ui.stationTimeBar->setEnabled( false );
        ui.stationTimeBar->setVisible( false );

        ui.actionTag->setEnabled( false );
        ui.actionRecommend->setEnabled( false );
        ui.actionLove->setEnabled( false );
        ui.actionBan->setEnabled( false );
        ui.actionSkip->setEnabled( false );

        statusBar()->showMessage( "" );

        m_stopped = true;
    }
    else
    {
        m_iWebService->recentTracks();

        // Playback started
        if ( data.source() == MetaData::Radio )
        {
            ui.songTimeBar->useInternalTimer();
            ui.songTimeBar->setMaximum( data.duration() );
            ui.songTimeBar->setReverse( true );
            ui.songTimeBar->start();

            ui.actionBan->setEnabled( true );
            ui.actionSkip->setEnabled( true );
        }
        else if ( data.source() == MetaData::Player )
        {
            // The reason for these extra checks is that one user had
            // crashes here (bug #56). This could happen if the player listener
            // receives a START, emits a queued signal, then immediately
            // receives a STOP. If the first queued signal doesn't get here
            // until after the STOP, GetActivePlayer will return NULL.
            CPlayerConnection* plyr = m_listener.GetActivePlayer();
            if ( plyr == NULL )
            {
                LOG( 1, "Caught a NULL player in setNewSong, not setting stopwatch" );
            }
            else
            {
                CStopWatch& watch = plyr->GetStopWatch();
                ui.songTimeBar->useStopWatch( &watch );
                ui.songTimeBar->setReverse( true );
            }

            ui.actionBan->setEnabled( false );
            ui.actionSkip->setEnabled( false );
        }
        else
            Q_ASSERT( false );

        ui.actionTag->setEnabled( true );
        ui.actionRecommend->setEnabled( true );
        ui.actionLove->setEnabled( true );
        //ui.tabWidget->setTabEnabled( 1, true );

        QHash<QString, QString> itemdata;
        itemdata.insert( "artist", data.artist() );
        itemdata.insert( "track", data.track() );
        ui.songTimeBar->setItemData( itemdata );

        // Only display scrobble timer if we're logging to profile and the track
        // is scrobblable.
        bool showTimer = m_iSettings->currentUser()->logToProfile() &&
                         m_metaData.isScrobblable();
        ui.songTimeBar->setEnabled( showTimer );

        if ( m_stopped )
        {
            m_stopped = false;
            emit playStarted();
            
            // This is annoying in the case of foobar which sends a stop and
            // then a new start on track change.
            //showMetaDataWidget();
        }

        // Clear any status messages on the start of a new track or it will look odd
        statusBar()->showMessage( "" );
    }

    ui.songTimeBar->setGreen( false );

    emit newSong( m_metaData );
}


void
Container::setNewStation( const QString& name, const QString& feed, bool newStation )
{
#ifndef HIDE_RADIO

    ui.stationTimeBar->setText( tr( "Station: %1 %2" ).arg( name ).arg( feed.isEmpty() ? "" : "(" + feed + ")" ) );
    ui.stationTimeBar->setMaximum( -1 );

    if ( newStation || !ui.stationTimeBar->isActive() )
        ui.stationTimeBar->start();

    ui.stationTimeBar->setEnabled( true );
    ui.stationTimeBar->setVisible( true );
#endif // HIDE_RADIO
}


void
Container::setTrackStatus( const QString& msg )
{
    ui.songTimeBar->setText( msg );
}


void
Container::setRadioStatus( const QString& msg )
{
#ifndef HIDE_RADIO

    if ( !msg.isEmpty() )
    {
        ui.stationTimeBar->setText( msg );
        ui.stationTimeBar->setMaximum( -1 );
        ui.stationTimeBar->setEnabled( true );
        ui.stationTimeBar->setVisible( true );
    }
    else
    {
        // This isn't very good. It will make the radio bar disappear if you're
        // listening to a station and then try to launch a tag station with
        // not enough content. Can't think of a workaround right now though.
        ui.stationTimeBar->setEnabled( false );
        ui.stationTimeBar->setVisible( false );
    }

#endif // HIDE_RADIO
}


void
Container::setScrobblerStatus( int status, QString msg )
{
    ScrobblerSubmitter::StatusCode statusCode = (ScrobblerSubmitter::StatusCode)status;

    // Hacky way of working out whether we're listening to radio
    if ( !ui.actionBan->isEnabled() )
    {
        // This stuff is horrible. Refactor!
        if ( statusCode == ScrobblerSubmitter::ScrobbledOK ||
             statusCode == ScrobblerSubmitter::ScrobbleFailed )
        {
            statusBar()->showMessage( "" );
            ui.songTimeBar->setTimerText( "" );
            if ( statusCode == ScrobblerSubmitter::ScrobbledOK )
            {
                ui.songTimeBar->setTimerText( tr( "track scrobbled" ) );
                ui.songTimeBar->setGreen( true );
            }
            else if ( statusCode == ScrobblerSubmitter::ScrobbleFailed )
            {
                ui.songTimeBar->setTimerText( tr( "error" ) );
            }
        }
        else
        {
            statusBar()->showMessage( msg );
        }
    }

    if ( statusCode == ScrobblerSubmitter::BadAuth )
    {
        // Change user icon to excl mark
        m_sysTray->setStatus( false );
        m_sysTray->setUser( m_iSettings->currentUser() );
    }
    else
    {
        m_sysTray->setStatus( true );
        m_sysTray->setUser( m_iSettings->currentUser() );
    }
}

void
Container::displayStationChangeError( const QString& error )
{
    QMessageBox::information( this, tr( "Could not tune in" ),  error );
}

void
Container::showTagDialog( int defaultTagType )
{
    TagDialog d( defaultTagType );
    d.setSong( m_metaData );
    d.exec();
}

void
Container::showTagDialogMD()
{
    showTagDialog( 0 );
}

void
Container::tagDone( const QString& artist, const QString& album, const QString& track, bool error )
{
    QString token = artist;
    if ( !album.isEmpty() )
        token += " - " + album;
    else
    if ( !track.isEmpty() )
        token += " - " + track;

    if ( error )
        statusBar()->showMessage( tr( "Tagging %1 failed" ).arg( token ) );
    else
        statusBar()->showMessage( tr( "You tagged %1 successfully" ).arg( token ) );
}


void
Container::play()
{
    if ( !m_lastFocusedStation.isEmpty() )
        m_iInput->load( m_lastFocusedStation );
    else
    {
        m_iInput->startStreaming();

        // Don't get requestedStreaming in this case so must do this
        ui.actionPlay->setVisible( false );
        ui.actionStop->setVisible( true );
    }

}


void
Container::stop()
{
    m_iInput->stopStreaming();
}


void
Container::love()
{
    QString confirmMsg = QString( tr( "Are you sure you want to add %1 to your loved tracks?" ) )
        .arg( m_metaData.track() );
    ConfirmDialog dlg( ConfirmDialog::Love, confirmMsg, this );
    if ( dlg.exec() == QDialog::Accepted )
    {
        ui.actionLove->setEnabled( false );
        m_iWebService->love( m_metaData.artist(), m_metaData.track() );
    }
}


void
Container::ban()
{
    QString confirmMsg = QString( tr( "Are you sure you want to ban %1?" ) )
        .arg( m_metaData.track() );
    ConfirmDialog dlg( ConfirmDialog::Ban, confirmMsg, this );
    if ( dlg.exec() == QDialog::Accepted )
    {
        ui.actionSkip->setEnabled( false );
        ui.actionLove->setEnabled( false );
        ui.actionBan->setEnabled( false );

        m_iWebService->ban( m_metaData.artist(), m_metaData.track() );
        m_iWebService->skip();
    }
}


void
Container::skip()
{
    ui.actionSkip->setEnabled( false );
    ui.actionLove->setEnabled( false );
    ui.actionBan->setEnabled( false );

    ui.songTimeBar->stop();
    ui.songTimeBar->clear();
    ui.songTimeBar->setEnabled( false );
    ui.songTimeBar->setTimerText( "" );

    m_iInput->nextSong();
}


void
Container::requestedStreaming()
{
    showMetaDataWidget();
    
    ui.actionPlay->setVisible( false );
    ui.actionStop->setVisible( true );
}


void
Container::preparedStreaming()
{
#ifndef HIDE_RADIO
    ui.actionPlay->setVisible( false );
    ui.actionStop->setVisible( true );
    ui.actionStop->setEnabled( true );
#endif // HIDE_RADIO
}


void
Container::startedStreaming()
{
#ifndef HIDE_RADIO

    // We must do this here as the radio streamer doesn't pay any attention
    // to commands when it's not running, so any setting applied straight after
    // the handshake would have been sent into a void.
    bool enabled = m_iSettings->currentUser()->logToProfile();
    m_iWebService->enableScrobbling( enabled, true );

    enabled = ui.actionToggleDiscoveryMode->isChecked();
    m_iWebService->enableDiscoveryMode( enabled );

#endif // HIDE_RADIO
}


void
Container::stoppedStreaming()
{
#ifndef HIDE_RADIO
    ui.stationTimeBar->setEnabled( false );
    ui.stationTimeBar->setVisible( false );

    ui.actionStop->setVisible( false );
    ui.actionPlay->setVisible( true );
    
    m_stopped = true;
#endif // HIDE_RADIO
}


void
Container::gotoProfile()
{
    QByteArray user = QUrl::toPercentEncoding( m_iSettings->currentUsername() );
    #ifdef HIDE_RADIO
        QString dashUrl = "http://www.lastfm.jp/user/" + user;
    #else
        QString dashUrl = "http://www.last.fm/user/" + user;
    #endif

    new BrowserThread( dashUrl );
}

void
Container::crash()
{
    int n = doCrash();
    LOGL( 4, n );
}

// This will get optimised out unless we put it in a function that returns
// something.
int
Container::doCrash()
{
    QString* s = NULL;
    return s->length();
}


