/***************************************************************************
 *   Copyright (C) 2005 - 2006 by                                          *
 *      Christian Muehlhaeuser, Last.fm Ltd <chris@last.fm>                *
 *      Erik Jaelevik, Last.fm Ltd <erik@last.fm>                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Steet, Fifth Floor, Boston, MA  02111-1307, USA.          *
 ***************************************************************************/

#include "playerlistener.h"

#include "ui_container.h"
#include "ui_playcontrols.h"

#include "controlinterface.h"
#include "extensioninterface.h"
#include "inputinterface.h"
#include "playbackinterface.h"
#include "transcodeinterface.h"
#include "settingsserviceinterface.h"
#include "webserviceinterface.h"

#include "settingsdialog.h"
#include "recommenddialog.h"
#include "tagdialog.h"
#include "imagebutton.h"

#include "autoupdater.h"
#include "scrobbler.h"
#include "settingsdialog.h"
#include "systray.h"

class MetaData;

class Container : public QMainWindow
{
    Q_OBJECT

    public:
        static Container& instance();
        void destroy();

        ScrobblerSubmitter& scrobbler();

    public slots:
        void initServices();

        void checkForUpdates( bool invokedByUser = true );
        void showSettingsDialog() { m_settingsDialog.exec(); }
        void showRecommendDialog() { RecommendDialog d; d.setSong( m_metaData ); d.exec(); }
        void showTagDialog( int defaultTagType = -1 );
        void showTagDialogMD();

        void showMetaDataWidget() { ui.tabWidget->setCurrentIndex( 1 ); }
        void restoreWindow() { showNormal(); activateWindow(); raise(); }

        void setPreloadStation( const QString& url ) { m_preloadStation = url; }
        void setLastFocusedStation( const QString& url ) { m_lastFocusedStation = url; }
        void actualClose( bool confirm = true );

        void requestedStreaming();

    signals:
        void newSong( MetaData song );
        void playStarted();

    protected:
        void closeEvent( QCloseEvent * event );
        void dropEvent( QDropEvent* event );
        void dragEnterEvent( QDragEnterEvent* event );
        void dragMoveEvent( QDragMoveEvent* event );

    private:
        // Singleton, hence private ctor/dtor
        Container( QWidget* parent = 0 );
        ~Container();

        void initListener();
        void initTray();

        void loadServices();
        void loadExtensions();

        void addExtensionToTabWidget( ExtensionInterface* extension );

        // Singleton instance
        static Container* mpInstance;

        friend class SysTray;
        SysTray* m_sysTray;
        friend class ControlInterface;
        ControlInterface* m_controlInterface;

        Ui::ContainerWindow ui;
        Ui::PlayControls ui_playcontrols;

        QByteArray m_sidebarState;
        QAction* m_sidebarActionWithIcon;
        QAction* m_skipAction;
        QAction* m_playAction;
        QAction* m_stopAction;

        QLabel m_scrobbleLabel;
        ImageButton m_scrobbleToggle;
        #ifdef Q_WS_MAC
        ImageButton m_scrobbleGraphicMac;
        #endif
        QPixmap m_scrobbleNormal;
        QPixmap m_scrobbleDisabled;
        QToolButton* m_closeTabButton;

        SettingsServiceInterface* m_iSettings;
        WebServiceInterface* m_iWebService;
        InputInterface* m_iInput;
        TranscodeInterface* m_iTranscode;
        PlaybackInterface* m_iPlayback;

        SettingsDialog m_settingsDialog;
        ExtensionInterface* m_eUserInfo;

        CPlayerListener m_listener;
        CAutoUpdater m_updater;
        bool m_userCheck;

        MetaData m_metaData;
        ScrobblerSubmitter m_scrobbler;

        QString m_preloadStation;
        QString m_lastFocusedStation;
        bool m_sidebarEnabled;
        bool m_stopped;

        static QMutex m_mutex;

    private slots:
        void enableSidebar();
        void splitterMoved();

        void about();
        void addUser();
        void deleteUser();
        void getPlugin();
        void addMediaDevice( const QString& uid );
        void showUserMenu();

        void handshakeResult( int result );
        void toggleScrobbling();
        void toggleDiscoveryMode() { m_iWebService->enableDiscoveryMode( ui.actionToggleDiscoveryMode->isChecked() ); }

        void preparedStreaming();
        void startedStreaming();
        void stoppedStreaming();
        void streamingResult( int error, const QString& reason );

        void bufferFull();
        void bufferEmpty( int size );

        void setNewSong( TrackInfo data );
        void setNewStation( const QString& name, const QString& feed, bool newStation );
        void setTrackStatus( const QString& message );
        void setRadioStatus( const QString& message );
        void displayStationChangeError( const QString& message );

        void displayUrlInStatusBar( const QString& url );
        void statusMessage( const QString& message ) { statusBar()->showMessage( message ); }

        void love();
        void ban();
        void skip();

        void play();
        void stop();

        void loveDone() { statusBar()->showMessage( tr( "You added this track to your Loved tracks." ) ); }
        void unloveDone() { statusBar()->showMessage( tr( "You removed this track from your Loved tracks." ) ); }
        void banDone() { statusBar()->showMessage( tr( "You banned this track." ) ); }
        void unbanDone() { statusBar()->showMessage( tr( "You unbanned this track." ) ); }
        void tagDone( const QString& artist, const QString& album, const QString& track, bool error );
        void recommendDone() { statusBar()->showMessage( tr( "Recommendation sent." ) ); }

        void userSelected( QAction* action );
        void updateCheckDone( bool updatesAvailable, bool error, QString errorMsg );
        void updateUserStuff( UserSettingsServiceInterface* user );

        void closeCurrentTab() { if ( ui.tabWidget->currentIndex() > 1 ) ui.tabWidget->removeTab( ui.tabWidget->currentIndex() ); }
        void currentTabChanged( int index );
        void showFriendsTab() { if ( false && m_eUserInfo ) addExtensionToTabWidget( m_eUserInfo ); }

        void listenerException( QString msg );
        void setScrobblerStatus( int, QString );

        void minimiseToTray();
        void gotoProfile();
        void crash();
        int doCrash();
};
