/***************************************************************************
 *   Copyright (C) 2005 - 2006 by                                          *
 *      Christian Muehlhaeuser, Last.fm Ltd <chris@last.fm>                *
 *      Erik Jaelevik, Last.fm Ltd <erik@last.fm>                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Steet, Fifth Floor, Boston, MA  02111-1307, USA.          *
 ***************************************************************************/

// Got to have this first so Winsock doesn't moan about multiple defines
#include "container.h"

#include "lastfmapplication.h"
#include "configwizard.h"
#include "containerutils.h"
#include "logindialog.h"

#include "Loqqer.h"
#include <QDebug>
#include <QtGui>

#ifdef WIN32
    #include <windows.h>
#endif

LastFmApplication::LastFmApplication( int& argc, char** argv ) :
    QApplication( argc, argv ),
    m_endSessionEmitted( false )
{
    initLogger();
	initTranslator();

    setApplicationName( tr( "Last.fm" ) );
    setQuitOnLastWindowClosed( false );

    #ifndef WIN32
        setWindowIcon( QIcon( dataPath( "icons/as.ico" ) ) );
    #endif

    #ifdef Q_WS_X11
        QApplication::setStyle( "plastique" );
    #endif

    // This is needed so that relative paths will work on Windows regardless
    // of where the app is launched from.
    QDir::setCurrent( applicationDirPath() );

}


void
LastFmApplication::init()
{
    SettingsServiceInterface* settings = settingsService();
    Q_ASSERT( settings != NULL );

    LOGL( 1, "App version: " << settings->version() );

    Container& window = Container::instance();
    bool firstRun = settings->firstRun();
    if ( firstRun )
    {
        LOG( 3, "First run, launching config wizard\n" );
        QFile( savePath( "mediadevice.db" ) ).remove();

        ConfigWizard wiz( NULL, ConfigWizard::Login );
        if ( wiz.exec() == QDialog::Rejected )
        {
            // If user cancels config wizard, we need to exit
            QApplication::exit( 0 );
            return;
        }
    }

    for ( int i = 1; i < qApp->arguments().size(); i++ )
    {
        if ( qApp->arguments().at( i ).contains( "addMediaDevice" ) )
        {
            ConfigWizard wiz( NULL, ConfigWizard::MediaDevice, qApp->arguments().at( i ).split( "addMediaDevice/" ).at( 1 ) );
            wiz.exec();
        }
    }

    // Do we have a current user?
    QString currentUser = settings->currentUsername();
    bool doLogin = false;
    if ( currentUser.isEmpty() )
    {
        LOG( 3, "No current user\n" );
        doLogin = true;
    }
    else
    {
        UserSettingsServiceInterface* user = settings->user( currentUser );
        doLogin = !user->rememberPass();
    }

    if ( doLogin && !firstRun )
    {
        LOG( 3, "Ask for login\n" );

        LoginDialog login( NULL, LoginWidget::LOGIN, currentUser );
        if ( login.exec() != QDialog::Accepted )
        {
            QApplication::exit( 0 );
            return;
        }
    }

    // Refresh user as it might now have changed
    UserSettingsServiceInterface* user = settings->currentUser();
    LOG( 3, "Starting app for user: " << user->username() << "\n" );

    window.initServices();

    // This is needed for the app to shut down properly when asked by Windows
    // to shut down.
    connect( this,    SIGNAL( endSession() ),
             &window, SLOT  ( actualClose() ) );

    if ( !arguments().contains( "-tray" ) )
    {
        window.show();
    }
}


void
LastFmApplication::initLogger()
{
    gLogger.Init( savePath( "container.log" ), false );
    gLogger.SetLevel( 4 );
}

void
LastFmApplication::initTranslator()
{
    QString langCode;

    #ifdef HIDE_RADIO
		langCode = "jp";
	#else

        // First check settings
        langCode = settingsService()->appLanguage();

        if ( langCode.isEmpty() )
        {
            // If none found, use system locale
            QLocale::Language qtLang = QLocale::system().language();
            langCode = qtLanguageToLangCode( qtLang );
            settingsService()->setAppLanguage( langCode );

            LOGL( 3, "System locale: " << langCode );
        }
    
	#endif

    setLanguage( langCode );
    installTranslator( &m_translatorApp );
    installTranslator( &m_translatorQt );
}

void
LastFmApplication::setLanguage( QString langCode )
{
    LOGL( 3, "Setting language to: " << langCode );

    m_lang = langCode;

    // Discards previously loaded translations
    m_translatorApp.load( dataPath( "i18n/lastfm_%1" ).arg( langCode ) );
    m_translatorQt.load( dataPath( "i18n/qt_%1" ).arg( langCode ) );
}

#ifdef WIN32
bool
LastFmApplication::winEventFilter( MSG * msg, long * result )
{
    /*
    typedef struct MSG {
        HWND        hwnd;
        UINT        message;
        WPARAM      wParam;
        LPARAM      lParam;
        DWORD       time;
        POINT       pt;
    }*/

    // This message is sent by Windows when we're being shut down as part
    // of a Windows shutdown. Don't want to just minimise to tray so we
    // emit a special endSession signal to Container. It can get sent
    // several times though so we must guard against emitting the signal
    // more than once.
    if ( msg->message == WM_QUERYENDSESSION )
    {
        if ( !m_endSessionEmitted )
        {
            m_endSessionEmitted = true;
            emit endSession();
        }
        *result = 1; // tell Windows it's OK to shut down
        return true; // consume message
    }

    return false; // let Qt handle it
}
#endif // WIN32


void
LastFmApplication::customEvent( QEvent* event )
{
    if ( event->type() == EventLoopStarted )
    {
        init();
    }
}

