/***************************************************************************
 *   Copyright (C) 2005 - 2006 by                                          *
 *      Christian Muehlhaeuser, Last.fm Ltd <chris@last.fm>                *
 *      Erik Jaelevik, Last.fm Ltd <erik@last.fm>                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Steet, Fifth Floor, Boston, MA  02111-1307, USA.          *
 ***************************************************************************/

#include "containerutils.h"

#include <QtGui>
#include <QHttp>
#include <QObject>
#include <QProcess>

#include "inputinterface.h"
#include "playbackinterface.h"
#include "settingsserviceinterface.h"
#include "transcodeinterface.h"
#include "webserviceinterface.h"

//#include "Loqqer.h"
#include "libLastFMTools/md5/md5.h"

#ifdef WIN32
    #define _WINSOCKAPI_   /* Prevent inclusion of winsock.h in windows.h */
    #include "windows.h"
    #include "shfolder.h"
#endif

static SettingsServiceInterface* g_settingsService = NULL;
static WebServiceInterface*      g_webService = NULL;
static InputInterface*           g_inputService = NULL;
static TranscodeInterface*       g_transcodeService = NULL;
static PlaybackInterface*        g_playbackService = NULL;

QString MD5Digest( const char *token )
{
    md5_state_t md5state;
    unsigned char md5pword[16];

    md5_init( &md5state );
    md5_append( &md5state, (unsigned const char *)token, (int)strlen( token ) );
    md5_finish( &md5state, md5pword );

    char tmp[33];
    strncpy( tmp, "\0", sizeof(tmp) );
    for ( int j = 0; j < 16; j++ )
    {
        char a[3];
        sprintf( a, "%02x", md5pword[j] );
        tmp[2*j] = a[0];
        tmp[2*j+1] = a[1];
    }

    return QString::fromAscii( tmp );
}


QString dataPath( QString file )
{
    return QString( "%1/data/%2" ).arg( QApplication::applicationDirPath() ).arg( file );
}


QString savePath( QString file )
{
    QString prefix;
    QString midfix;

    prefix = QApplication::applicationDirPath();
    #ifdef WIN32
        if ( ( QSysInfo::WindowsVersion & QSysInfo::WV_DOS_based ) == 0 )
        {
            // Use this for non-DOS-based Windowses
            char acPath[MAX_PATH];
            HRESULT h = SHGetFolderPathA( NULL, CSIDL_LOCAL_APPDATA | CSIDL_FLAG_CREATE,
                                          NULL, 0, acPath );
            if ( h == S_OK )
            {
                prefix = QString::fromLocal8Bit( acPath );
                prefix.append( "/Last.fm/Client" );
            }
            else
            {
                //LOG( 1, "Couldn't get Local Settings save path\n" );
            }
        }
    #else
        prefix = QApplication::applicationDirPath();
    #endif

    QDir saveDir( prefix );
    if (!saveDir.exists())
    {
        saveDir.mkpath( prefix );
    }

    return QString( "%1/%2" ).arg( prefix ).arg( file );
}


QString QHttpStateToString(int state)
{
    QString msg;

    switch (state)
    {
        case QHttp::Unconnected:
        {
            msg = QCoreApplication::translate( "WebService", "No connection." );
        }
        break;

        case QHttp::HostLookup:
        {
            msg = QCoreApplication::translate( "WebService", "Looking up host..." );
        }
        break;

        case QHttp::Connecting:
        {
            msg = QCoreApplication::translate( "WebService", "Connecting..." );
        }
        break;

        case QHttp::Sending:
        {
            msg = QCoreApplication::translate( "WebService", "Sending request..." );
        }
        break;

        case QHttp::Reading:
        {
            msg = QCoreApplication::translate( "WebService", "Downloading." );
        }
        break;

        case QHttp::Connected:
        {
            msg = QCoreApplication::translate( "WebService", "Connected." );
        }
        break;

        case QHttp::Closing:
        {
            msg = QCoreApplication::translate( "WebService", "Closing connection..." );
        }
        break;
    }

    return msg;
}

QString qtLanguageToLangCode( QLocale::Language qtLang )
{
    QString langCode;

    switch ( qtLang )
    {
        case QLocale::English:    langCode = "en"; break;
        case QLocale::French:     langCode = "fr"; break;
        case QLocale::Italian:    langCode = "it"; break;
        case QLocale::German:     langCode = "de"; break;
        case QLocale::Spanish:    langCode = "es"; break;
        case QLocale::Portuguese: langCode = "pt"; break;
        case QLocale::Polish:     langCode = "pl"; break;
        case QLocale::Russian:    langCode = "ru"; break;
        case QLocale::Japanese:   langCode = "jp"; break;
        case QLocale::Chinese:    langCode = "cn"; break;
        case QLocale::Korean:     langCode = "kr"; break;
        default:                  langCode = "en"; break;
    }

    return langCode;
}

static void loadServiceError( QString name )
{
    Q_ASSERT( !"good" );
    QMessageBox::critical( 0,
        QCoreApplication::translate( "Container", "Error" ),
        QCoreApplication::translate( "Container", "Couldn't load service: %1! The application won't be able to startup!" ).arg( name ) );

    QCoreApplication::exit( 1 );
}


static QObject* loadService( QString name )
{
    #ifndef QT_NO_DEBUG
        QDir servicesDir = QDir( qApp->applicationDirPath() + "/services/debug" );
    #else
        QDir servicesDir = QDir( qApp->applicationDirPath() + "/services" );
    #endif

    foreach ( QString fileName, servicesDir.entryList( QDir::Files ) )
    {
        if ( fileName.startsWith( EXTENSION_PREFIX ) || !QLibrary::isLibrary( servicesDir.absoluteFilePath( fileName ) ) )
            continue;

        if ( fileName.contains( name ) )
        {
            QObject* plugin = QPluginLoader( servicesDir.absoluteFilePath( fileName ) ).instance();

            if ( plugin == NULL )
            {
                loadServiceError( name );
            }

            return plugin;
        }
    }

    return NULL;
}

/*
    We're taking a slight risk by not guarding these calls with mutexes because
    they can in theory be accessed by different threads. However, since they
    only get loaded once at startup, this is not a problem in practice. If
    we decide to synchronise access in the future we'd need a separate mutex
    for each getter function, as otherwise we'd end up with same thread
    deadlocks.
*/
SettingsServiceInterface* settingsService()
{
    if ( g_settingsService == NULL )
    {
        QObject* obj = loadService( "service_settings" );
        g_settingsService = qobject_cast<SettingsServiceInterface*>( obj );
    }
    
    return g_settingsService;
}


WebServiceInterface* webService()
{
    if ( g_webService == NULL )
    {
        QObject* obj = loadService( "service_web" );
        g_webService = qobject_cast<WebServiceInterface*>( obj );
    }
    
    return g_webService;
}


InputInterface* inputService()
{
    if ( g_inputService == NULL )
    {
        QObject* obj = loadService( "input_http" );
        g_inputService = qobject_cast<InputInterface*>( obj );
    }
    
    return g_inputService;
}


TranscodeInterface* transcodeService()
{
    if ( g_transcodeService == NULL )
    {
        QObject* obj = loadService( "transcode_mp3" );
        g_transcodeService = qobject_cast<TranscodeInterface*>( obj );
    }
    
    return g_transcodeService;
}


PlaybackInterface* playbackService()
{
    if ( g_playbackService == NULL )
    {
        QObject* obj = loadService( "playback_rtaudio" );
        g_playbackService = qobject_cast<PlaybackInterface*>( obj );
    }
    
    return g_playbackService;
}

