/***************************************************************************
 *   Copyright (C) 2005 - 2006 by                                          *
 *      Christian Muehlhaeuser, Last.fm Ltd <chris@last.fm>                *
 *      Erik Jaelevik, Last.fm Ltd <erik@last.fm>                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Steet, Fifth Floor, Boston, MA  02111-1307, USA.          *
 ***************************************************************************/

#include "http.h"
#include "containerutils.h"

#include <QDir>

#ifdef WIN32
    #include <windows.h>
#endif

#undef LOG
#undef LOGL
#define LOG(x, y)
#define LOGL(x, y)

Http::Http( QObject* parent ) :
    QHttp( parent ),
    m_dataID( -1 ),
    m_nextId( 0 )
{
    init();
}


Http::Http( const QString& hostName, int port, QObject* parent ) :
    QHttp( parent ),
    m_dataID( -1 ),
    m_hostname( hostName ),
    m_nextId( 0 )
{
    init();
    setHost( hostName, port );
}


void
Http::init()
{
    m_cachePath = savePath( "cache/" );
	QDir dir( m_cachePath );
	if ( !dir.exists() )
	{
		dir.mkdir( m_cachePath );
	}
	
    applyProxy();
    connect( this, SIGNAL( requestFinished( int, bool ) ), this, SLOT( dataFinished( int, bool ) ) );
}


void
Http::applyProxy()
{
    SettingsServiceInterface* settings = settingsService();
    if ( settings->useProxy() )
    {
        setProxy( settings->proxyHost(), settings->proxyPort(), settings->proxyUsername(), settings->proxyPassword() );
    }
    else
    {
        //setProxy( "", 0 );
    }
}


int
Http::get( const QString& path, bool useCache )
{
    applyProxy();
    m_buffer.clear();
    QString url = m_hostname + path;

    if ( useCache && haveCachedCopy( url ) )
    {
        // Using a singleshot so that we can return an ID and have
        // the operation proceed asynchronously.
        m_cacheStack.push( CachedRequestData( ++m_nextId, url ) );
        QTimer::singleShot( 0, this, SLOT( getFromCache() ) );
        return m_nextId;
    }

    m_dataID = QHttp::get( path );
    if ( useCache )
        m_requestStack.insert( m_dataID, CachedRequestData( ++m_nextId, url ) );

    return m_dataID;
}


int
Http::post( const QString& path, QIODevice* data )
{
    applyProxy();
    m_buffer.clear();

    m_dataID = QHttp::post( path, data );
    return m_dataID;
}


int
Http::post( const QString& path, const QByteArray& data )
{
    applyProxy();
    m_buffer.clear();

    m_dataID = QHttp::post( path, data );
    return m_dataID;
}


int
Http::request( const QHttpRequestHeader& header, QIODevice* data, QIODevice* to )
{
    applyProxy();
    m_buffer.clear();

    m_dataID = QHttp::request( header, data, to );
    return m_dataID;
}


int
Http::request( const QHttpRequestHeader& header, const QByteArray& data, QIODevice* to, bool useCache )
{
    applyProxy();
    m_buffer.clear();
    QString key( data );

    if ( useCache && haveCachedCopy( key ) )
    {
        // Using a singleshot so that we can return an ID and have
        // the operation proceed asynchronously.
        m_cacheStack.push( CachedRequestData( ++m_nextId, key ) );
        QTimer::singleShot( 0, this, SLOT( getFromCache() ) );
        return m_nextId;
    }

    m_dataID = QHttp::request( header, data, to );
    if ( useCache )
        m_requestStack.insert( m_dataID, CachedRequestData( ++m_nextId, key ) );

    return m_dataID;
}


void
Http::dataFinished( int id, bool error )
{
    if ( error )
        emit errorOccured( QHttp::error(), QHttp::errorString() );
    else
        if ( id == m_dataID )
        {
            checkBuffer();

            CachedRequestData req = m_requestStack.take( id );
            if ( !req.m_cacheKey.isEmpty() )
            {
                putCachedCopy( req.m_cacheKey, m_buffer );
            }

            emit dataAvailable( m_buffer );
        }
}


bool
Http::haveCachedCopy( QString url ) const
{
    return QFile::exists( pathToCachedCopy( url ) ) && QFileInfo( pathToCachedCopy( url ) ).isReadable();
}


QString
Http::pathToCachedCopy( QString cacheKey ) const
{
    QString keyMd5 = MD5Digest( qPrintable( cacheKey ) );

    return m_cachePath + keyMd5;
}


void
Http::putCachedCopy( QString url, const QByteArray& data )
{
    #ifdef WIN32
        if ( url.size() > MAX_PATH )
        {
            url.chop( url.size() - MAX_PATH );
        }
    #endif

    QFile f( pathToCachedCopy( url ) );

    if ( !f.open( QIODevice::WriteOnly ) )
    {
        //LOGL( 1, "Failed to open file " << url << " for writing to cache" );
        return;
    }

    f.write( data );
}


void Http::getFromCache()
{
    Q_ASSERT( !m_cacheStack.isEmpty() );

    // Pick next request off cache stack
    CachedRequestData req = m_cacheStack.pop();

    QFile f( pathToCachedCopy( req.m_cacheKey ) );
    if ( !f.open( QIODevice::ReadOnly ) )
    {
        //LOGL( 1, "Failed to open cached file, returning error" );

        // TODO: emit error
        return;
    }

    // Keeping it as UTF-8, conversion will the done in parse function
    QByteArray result = f.readAll();
    emit dataAvailable( result );
}
