/***************************************************************************
 *   Copyright (C) 2005 - 2006 by                                          *
 *      Christian Muehlhaeuser, Last.fm Ltd <chris@last.fm>                *
 *      Erik Jaelevik, Last.fm Ltd <erik@last.fm>                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Steet, Fifth Floor, Boston, MA  02111-1307, USA.          *
 ***************************************************************************/

#include "metadata.h"

#include "settingsserviceinterface.h"
#include "containerutils.h"

//#include "Loqqer.h"
#undef LOG
#undef LOGL
#define LOG(x, y)
#define LOGL(x, y)

static const int SUBMISSION_THRESHOLD = 240;
static const int SUBMISSION_MIN_LENGTH = 31;

QStringList MetaData::mInvalidArtists = initInvalid();

QStringList
MetaData::initInvalid()
{
    QStringList list;
    list << "unknown artist"
         << "unknown"
         << "[unknown]"
         << "[unknown artist]";
    return list;
}


MetaData::MetaData()
    : TrackInfo()
{
}


MetaData::MetaData( const MetaData& that )
    : TrackInfo()
{
    populate( that );
}


MetaData::MetaData( const QDomElement& element )
    : TrackInfo( element )
{
}


MetaData&
MetaData::operator=( const MetaData& that )
{
    // Check for self-assignment:
    if ( &that != this )
    {
        populate( that );
    }

    return *this;
}


MetaData&
MetaData::operator=( const TrackInfo& that )
{
    // Check for self-assignment:
    if ( &that != this )
    {
        TrackInfo::populate( that );
    }

    return *this;
}


MetaData::~MetaData()
{
}


void
MetaData::populate( const TrackInfo &trackinfo )
{
    TrackInfo::populate( trackinfo );
}


void
MetaData::populate( const MetaData &metaData )
{
    TrackInfo::populate( metaData );

    setTrackTags( metaData.trackTags() );
    setArtistTags( metaData.artistTags() );
    setSimilarArtists( metaData.similarArtists() );
    setWiki( metaData.wiki() );
    setAlbumPicUrl( metaData.albumPicUrl() );
    setArtistPicUrl( metaData.artistPicUrl() );
    setArtistPageUrl( metaData.artistPageUrl() );
    setAlbumPageUrl( metaData.albumPageUrl() );
    setTrackPageUrl( metaData.trackPageUrl() );
    setWikiPageUrl( metaData.wikiPageUrl() );
    setBuyTrackString( metaData.buyTrackString() );
    setBuyTrackUrl( metaData.buyTrackUrl() );
    setBuyAlbumString( metaData.buyAlbumString() );
    setBuyAlbumUrl( metaData.buyAlbumUrl() );
    setTopFans( metaData.topFans() );
    setNumListeners( metaData.numListeners() );
    setNumPlays( metaData.numPlays() );
    setLabel( metaData.label() );
    setNumTracks( metaData.numTracks() );
    setReleaseDate( metaData.releaseDate() );
}


void
MetaData::clear()
{
    populate( MetaData() );
}


void
MetaData::timeStampMe()
{
    QDateTime now = QDateTime::currentDateTime();
    QDateTime utc = now.toUTC();
    setTimeStamp( utc.toString( "yyyy-MM-dd hh:mm:ss" ) );
}


bool
MetaData::sameAs( const MetaData &metaData )
{
    if ( artist() != metaData.artist() )
        return false;

    if ( track() != metaData.track() )
        return false;

    if ( album() != metaData.album() )
        return false;

    return true;
}


bool
MetaData::isScrobblable()
{
    // Radio tracks always scrobblable
    if ( source() == Radio )
    {
        return true;
    }

    // Check timestamp
    if ( timeStamp().isEmpty() )
    {
        return false;
    }

    // Check duration
    if ( duration() < SUBMISSION_MIN_LENGTH )
    {
        LOG( 3, "Track length is " << duration() << " s which is too short, will not submit.\n" );
        return false;
    }

    // Check if any required fields are empty
    if ( artist().isEmpty() || track().isEmpty() )
    {
        LOG( 3, "Artist, track or duration was missing, will not submit.\n" );
        return false;
    }

    // Check if dir excluded
    if ( dirExcluded( path() ) )
    {
        LOG( 3, "Track is in excluded directory '" << path() << "', " << "will not submit.\n" );
        return false;
    }

    // Check if artist name is an invalid one like "unknown"
    foreach( QString invalid, mInvalidArtists )
    {
        if ( artist().contains( invalid, Qt::CaseInsensitive ) )
        {
            LOG( 3, "Artist '" << artist() << "' is an invalid artist name, will not submit.\n" );
            return false;
        }
    }

    // All tests passed!
    return true;
}

bool
MetaData::isTrackBuyable()
{
    return !buyTrackUrl().isEmpty();
}

bool
MetaData::isAlbumBuyable()
{
    return !buyAlbumUrl().isEmpty();
}

bool
MetaData::dirExcluded(
    const QString& path )
{
    if ( path.isEmpty() )
    {
        return false;
    }

    SettingsServiceInterface* settings = settingsService();
    QStringList excluded = settings->currentUser()->excludedDirs();

    QDir dirToTest( path );
    QString pathToTest = dirToTest.canonicalPath();
    #ifdef WIN32
        pathToTest = pathToTest.toLower();
    #endif

    foreach( QString bannedPath, excluded )
    {
        QDir bannedDir( bannedPath );
        bannedPath = bannedDir.canonicalPath();
        #ifdef WIN32
            bannedPath = bannedPath.toLower();
        #endif

        // Try and match start of given path with banned dir
        if ( pathToTest.startsWith( bannedPath ) )
        {
            // Found, this path is from a banned dir
            return true;
        }
    }

    // The path wasn't found in exclusions list
    return false;
}


int
MetaData::calcGoal()
{
    // If we don't have a length or it's less than the minimum, return the
    // threshold
    if ( duration() <= 0 || duration() < SUBMISSION_MIN_LENGTH )
    {
        return SUBMISSION_THRESHOLD;
    }

    int nGoal = qMin( SUBMISSION_THRESHOLD, duration() / 2 );
    return nGoal;
}
