/***************************************************************************
 *   Copyright (C) 2005 - 2006 by                                          *
 *      Christian Muehlhaeuser, Last.fm Ltd <chris@last.fm>                *
 *      Erik Jaelevik, Last.fm Ltd <erik@last.fm>                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Steet, Fifth Floor, Boston, MA  02111-1307, USA.          *
 ***************************************************************************/

#include <QtGui>
#include <QSqlQuery>
#include <QSqlResult>
#include <QSqlError>

#include "mediadevice.h"
#include <containerutils.h>
#include <Loqqer.h>
#include <trackinfo.h>
#include <container.h>

#define ZLIB_WINAPI
#include "zip.h"

#define ESC( token ) ( QString( token ).replace( "\'", "''" ) )
QString XML_VERSION = "1.0";

static const QString k_host = "moses.last.fm";


MediaDevices::MediaDevices()
    : m_http( 0 )
{
    m_savePath = savePath( "bootstrap.xml" );

    m_gpod = iTunesDevice();

    connect( m_gpod, SIGNAL( progress( int ) ),
             this,   SIGNAL( bootstrapProgress( int ) ) );

    m_db = QSqlDatabase::database( "mediadevice" );
    if ( !m_db.isValid() )
    {
        m_db = QSqlDatabase::addDatabase( "QSQLITE", "mediadevice" );
        m_db.setDatabaseName( savePath( "mediadevice.db" ) );
    }
    m_db.open();

    qDebug() << "Opening DB" << ( m_db.isValid() ? "worked" : "failed" );

    if ( !m_db.tables().contains( "mediadevice" ) )
        initDatabase();
}


MediaDevices::~MediaDevices()
{
    m_db.close();

    delete m_http;
}


void
MediaDevices::bootStrap( bool scrobble )
{
    QFile file( m_savePath );
    file.open( QIODevice::WriteOnly | QIODevice::Text );

    QDomDocument newdoc;
    QDomElement submitQueue = newdoc.createElement( "bootstrap" );
    submitQueue.setAttribute( "product", "iTunes" );
    submitQueue.setAttribute( "version", XML_VERSION );

    TrackInfo track;
    if ( m_gpod )
    {
        qDebug() << "Reading iPod's tracks...";
        track = m_gpod->nextTrack();
        qDebug() << "Reading iPod's first track finished...";
    }

    int x = 0;
//     m_db.transaction();
    while ( !track.isEmpty() )
    {
        updateTrack( track, scrobble );
        qApp->processEvents();

        QDomElement i = track.toDomElement( newdoc );
        submitQueue.appendChild( i );

        track = m_gpod->nextTrack();
    }
//     m_db.commit();
    m_db.close(); // why?? what happens if bootstrap is called again?

    QDomNode submitNode = newdoc.importNode( submitQueue, true );
    newdoc.appendChild( submitNode );

    QTextStream stream( &file );
    stream.setCodec( "UTF-8" );
    stream << "<?xml version=\"1.0\" encoding=\"utf-8\"?>\n";
    stream << newdoc.toString();

    file.close();

    if ( scrobble )
    {
        zipFiles();
        sendZip();
    }

    settingsService()->bootstrapDone();
    emit bootstrapDone();
}


bool
MediaDevices::initDatabase()
{
    qDebug() << "Creating mediadevice database!";

    QSqlQuery query( m_db );
    query.exec( "CREATE TABLE mediadevice ( "
                    "id         INTEGER PRIMARY KEY, "
                    "artist     VARCHAR( 255 ), "
                    "album      VARCHAR( 255 ), "
                    "track      VARCHAR( 255 ), "
                    "duration   INTEGER, "
                    "timestamp  INTEGER, "
                    "playcount  INTEGER )" );

    query.exec( "CREATE INDEX id_idx ON mediadevice( artist )" );
    query.exec( "CREATE INDEX artist_idx ON mediadevice( artist )" );
    query.exec( "CREATE INDEX album_idx ON mediadevice( album )" );
    query.exec( "CREATE INDEX track_idx ON mediadevice( track )" );

    return true;
}


bool
MediaDevices::updateTrack( TrackInfo track, bool autoScrobble )
{
    QSqlQuery query( m_db );
    query.exec( QString( "SELECT id, playcount FROM mediadevice WHERE "
                         "artist = '%1' AND album = '%2' AND track = '%3'" )
                   .arg( ESC( track.artist() ) )
                   .arg( ESC( track.album() ) )
                   .arg( ESC( track.track() ) ) );

    if ( !query.first() )
    {
        qDebug() << "Adding track to database:" << track.artist() << track.album() << track.track() << track.playCount();
        query.exec( QString( "INSERT INTO mediadevice ( artist, album, track, duration, timestamp, playcount ) "
                              "VALUES ('%1', '%2', '%3', '%4', '%5', '%6') " )
                        .arg( ESC( track.artist() ) )
                        .arg( ESC( track.album() ) )
                        .arg( ESC( track.track() ) )
                        .arg( track.duration() )
                        .arg( track.timeStamp() )
                        .arg( track.playCount() ) );

//         qDebug() << query.lastError().text() << query.lastError().databaseText();
    }
    else
    {
        int id = query.value( 0 ).toInt();
        int playcount = query.value( 1 ).toInt();

        if ( track.playCount() > playcount )
        {
            qDebug() << "Updating track" << id << "in database:" << track.artist() << track.album() << track.track() << track.playCount();
            query.exec( QString( "UPDATE mediadevice SET artist = '%1', album = '%2', "
                                  "track = '%3', duration = '%4', timestamp = '%5', playcount = '%6' WHERE id = '%7'" )
                            .arg( ESC( track.artist() ) )
                            .arg( ESC( track.album() ) )
                            .arg( ESC( track.track() ) )
                            .arg( track.duration() )
                            .arg( track.timeStamp() )
                            .arg( track.playCount() )
                            .arg( id ) );

//             qDebug() << query.executedQuery();
//             qDebug() << query.lastError().text() << query.lastError().databaseText();
        }
    }

    return true;
}


bool
MediaDevices::zipFiles()
{
    QString zipFileName = "LastFMBootstrap.zip";

    QDir tempDir = QDir::temp();
    tempDir.remove( zipFileName );

    QString zipFilePathOrig = tempDir.absolutePath() + "/" + zipFileName;
    m_zipFile.setFileName( zipFilePathOrig );
    QByteArray zipFilePath = QFile::encodeName( zipFilePathOrig );

    #ifdef WIN32
        // Minizip uses DOS encoding
        char buf[MAX_PATH];
        CharToOemA( zipFilePath.data(), buf );
        zipFilePath = buf;
    #endif

    zipFile zipHandle = zipOpen( zipFilePath.data(), APPEND_STATUS_CREATE );
    if ( !zipHandle )
    {
        return false;
    }

    {
        QString filename = savePath( "bootstrap.xml" );

        // Update display
        QApplication::processEvents();

        // Zip a file
        QFile file( filename );

        bool success = file.open( QIODevice::ReadOnly );
        if ( !success )
        {
            // Couldn't open source file, move on to the next one
            return false;
        }

        QString name = QFileInfo( filename ).fileName();

        // Create entry for file in zip
        zip_fileinfo dummy;
        memset( &dummy, 0, sizeof( dummy ) );
        int res = zipOpenNewFileInZip(
            zipHandle,
            qPrintable( QFileInfo( filename ).fileName() ),
            &dummy,
            NULL, 0,
            NULL, 0,
            "",
            Z_DEFLATED,
            Z_DEFAULT_COMPRESSION );

        if ( res != ZIP_OK )
        {
            // Creation failed, move on to the next one
            zipCloseFileInZip( zipHandle );
            return false;
        }

        QByteArray fileData = file.readAll();
        res = zipWriteInFileInZip(
            zipHandle,
            reinterpret_cast<const void*>( fileData.constData() ),
            fileData.size() - 1 ); // because QByteArray adds a \0

        zipCloseFileInZip( zipHandle );
    }

    zipClose( zipHandle, "" );
    return true;
}


void
MediaDevices::sendZip()
{
    if ( m_username.isEmpty() )
    {
        SettingsServiceInterface* settings = settingsService();
        if ( settings != NULL )
        {
            UserSettingsServiceInterface* user = settings->currentUser();
            if ( user != NULL )
            {
                m_username = user->username();
                m_passMd5 = user->password();
            }
            else
            {
                QStringList allUsers = settings->allUsers();
                if ( allUsers.size() != 0 )
                {
                    m_username = allUsers.at( 0 );
                    user = settings->user( m_username );
                    if ( user != NULL )
                    {
                        m_passMd5 = user->password();
                    }
                    else
                    {
                        m_passMd5 = "unknown";
                    }
                }
                else
                {
                    m_username = "unknown";
                    m_passMd5 = "unknown";
                }
            }
        }
        else
        {
            m_username = "unknown";
            m_passMd5 = "unknown";
        }
    }

    if ( m_http )
        delete m_http;

    m_http = new Http( k_host, 80, this );
    connect( m_http, SIGNAL( requestFinished( int, bool ) ),
             this,   SLOT  ( sendFinished( int, bool ) ), Qt::QueuedConnection );
    connect( m_http, SIGNAL( dataSendProgress( int, int ) ),
             this,   SLOT  ( sendProgressMade( int, int ) ) );
    connect( m_http, SIGNAL( dataAvailable( const QByteArray& ) ),
             this,   SLOT  ( dataAvailable( const QByteArray& ) ) );
    connect( m_http, SIGNAL( responseHeaderReceived( QHttpResponseHeader ) ),
             this,   SLOT  ( headerReceived( QHttpResponseHeader ) ) );

    // Get Unix time
    time_t now;
    time( &now );
    QString time = QString::number( now );

    // Concatenate pw hash with time
    QString auth = m_passMd5 + time;
    QString authLower = m_passMd5.toLower() + time;
    // Hash the concatenated string to create auth code
    QString authMd5 = MD5Digest( auth.toUtf8() );
    QString authMd5Lower = MD5Digest( authLower.toUtf8() );

    QString path = QString( "/bootstrap/index.php?user=%1&time=%2&auth=%3&authlower=%4" )
        .arg( m_username, time, authMd5, authMd5Lower );

    QHttpRequestHeader header( "POST", path, 1, 1 );
    header.setValue( "Host", k_host );
    header.setValue( "Content-type", "multipart/form-data, boundary=AaB03x" );
    header.setValue( "Cache-Control", "no-cache" );
    header.setValue( "Accept", "*/*" );

    QByteArray bytes;
    bytes.append( "--AaB03x\r\n" );
    bytes.append( "content-disposition: " );
    bytes.append( "form-data; name=\"agency\"\r\n" );
    bytes.append( "\r\n" );
    bytes.append( "0\r\n" );
    bytes.append( "--AaB03x\r\n" );
    bytes.append( "content-disposition: " );
    bytes.append( "form-data; name=\"bootstrap\"; filename=\"" + m_zipFile.fileName() + "\"\r\n" );
    bytes.append( "Content-Transfer-Encoding: binary\r\n" );
    bytes.append( "\r\n" );

    m_zipFile.open( QIODevice::ReadOnly );
    bytes.append( m_zipFile.readAll() );
    m_zipFile.close();

    bytes.append( "\r\n" );
    bytes.append( "--AaB03x--" );
    header.setContentLength( bytes.length() );

    qDebug() << "Sending " << path;
    m_reqId = m_http->request( header, bytes );
}


void
MediaDevices::sendFinished( int id, bool error )
{
    if ( id != m_reqId )
        return;

    if ( error )
    {
    }
    else
    {
    }
}


void
MediaDevices::sendProgressMade( int done, int total )
{
    if ( done == total )
        qDebug() << "Bootstrap.zip sent to last.fm!";

    if ( total > 0 )
    {
        int kbDownloaded = done / 1024;
        int kbTotal = total / 1024;
    }
    else
    {
        // Can't find how to get the progress bar to just bounce back and forth
    }
}


void
MediaDevices::dataAvailable( QByteArray data )
{
    qDebug() << data;
}


void
MediaDevices::headerReceived( const QHttpResponseHeader &resp )
{
    qDebug() << resp.statusCode() << resp.reasonPhrase() << resp.toString();
}
