/***************************************************************************
 *   Copyright (C) 2005 - 2006 by                                          *
 *      Christian Muehlhaeuser, Last.fm Ltd <chris@last.fm>                *
 *      Erik Jaelevik, Last.fm Ltd <erik@last.fm>                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Steet, Fifth Floor, Boston, MA  02111-1307, USA.          *
 ***************************************************************************/

#include "playerconnection.h"
//#include "WebSubmitter.h"
//#include "SettingsManagerWin.h"
#include "Loqqer.h"
#include "container.h"
#include "containerutils.h"
#include "exceptions.h"

#include <QtCore>

#include <algorithm>
#include <sstream>

using namespace std;

/******************************************************************************
    CPlayerConnection
******************************************************************************/
CPlayerConnection::CPlayerConnection(
    const QString& sID) :
        QObject(),
        meState(eStopped),
        msID(sID),
        mLastCmd(PCMD_INVALID),
        mbSubmitted(false),
        mbBanned(false)
{
}

/******************************************************************************
    ~CPlayerConnection
******************************************************************************/
CPlayerConnection::~CPlayerConnection()
{
}

/******************************************************************************
    CPlayerConnection copy
******************************************************************************/
CPlayerConnection::CPlayerConnection(
    const CPlayerConnection& that) :
        QObject()
{
    Clone(that);
}

/******************************************************************************
    operator=
******************************************************************************/
CPlayerConnection&
CPlayerConnection::operator=(
    const CPlayerConnection& that)
{
    // Check for self-assignment:
    if (&that != this)
    {
        Clone(that);
    }

    return *this;
}

/******************************************************************************
    Clone
******************************************************************************/
void
CPlayerConnection::Clone(
    const CPlayerConnection& that)
{
    msID                    = that.msID;
    mLastCmd                = that.mLastCmd;
    mLastSubmission         = that.mLastSubmission;
    meState                 = that.meState;
    mbSubmitted             = that.mbSubmitted;
    mbBanned                = that.mbBanned;
    mTimer                  = that.mTimer;
}

/******************************************************************************
    OnStart
******************************************************************************/
void
CPlayerConnection::OnStart(
    const CPlayerCommand& cmd)
{
    // This will block until timer thread is finished, so that then there
    // would never be a chance of the timer thread getting into Update and
    // erroneously submitting this new track.
    mTimer.Stop();

    // Can't use [] operator on a const map so cast away the const
    CPlayerCommand& ncCmd = const_cast<CPlayerCommand&>(cmd);

    // Check for illegal commands
    string sValidCmds("catbmlp");
    map<char, string>::iterator pos;
    for (pos = ncCmd.mArgsMap.begin(); pos != ncCmd.mArgsMap.end(); ++pos)
    {
        if (sValidCmds.find(pos->first) == string::npos)
        {
            // Not found, throw
            LOG(2, "Illegal parameter in player cmd: " << pos->first <<
                " = " << pos->first << "\n");
            QString err = QT_TR_NOOP("The following argument is not valid: ") +
                pos->first;
            throw BadCommandException(qPrintable(err));
        }
    }

    // Read arguments into strings
    QString sArtist = QString::fromUtf8(ncCmd.mArgsMap['a'].c_str()).trimmed();
    QString sTrack  = QString::fromUtf8(ncCmd.mArgsMap['t'].c_str()).trimmed();
    QString sAlbum  = QString::fromUtf8(ncCmd.mArgsMap['b'].c_str()).trimmed();
    QString sMbId   = QString::fromUtf8(ncCmd.mArgsMap['m'].c_str()).trimmed();
    QString sLength = QString::fromUtf8(ncCmd.mArgsMap['l'].c_str()).trimmed();
    QString sPath   = QString::fromUtf8(ncCmd.mArgsMap['p'].c_str()).trimmed();

    mLastSubmission.clear();
    mLastSubmission.setArtist(sArtist);
    mLastSubmission.setTrack(sTrack);
    mLastSubmission.setAlbum(sAlbum);
    mLastSubmission.setDuration(sLength);
    mLastSubmission.setPath(sPath);
    mLastSubmission.setMbId(sMbId);
    mLastSubmission.setSource(MetaData::Player);
    mLastSubmission.timeStampMe();
    mLastCmd = PCMD_START;

    if (mLastSubmission.isScrobblable())
    {
        mbBanned = false;

        // Work out how long we need to listen before submission
        int nGoal = mLastSubmission.calcGoal();

        LOG(3, "Starting new track '" << sArtist << " - " << sTrack << "'. " <<
            "Goal: " << nGoal << "\n");

        mTimer.Reset();
        mTimer.SetTimeout(nGoal);
        mTimer.Start(this);
    }
    else
    {
        LOGL(3, "Track isn't scrobblable, won't start timer.");

        mbBanned = true;
    }

    meState = ePlaying;
    mbSubmitted = false;
}

/******************************************************************************
    OnStop
******************************************************************************/
void
CPlayerConnection::OnStop(
    const CPlayerCommand& cmd)
{
    // Is it OK to accept Stop at all times? Probably.
    mTimer.Stop();
    mTimer.Reset();

    meState = eStopped;
}

/******************************************************************************
    OnPause
******************************************************************************/
void
CPlayerConnection::OnPause(
    const CPlayerCommand& cmd)
{
    // Is it OK to accept pause at all times? Probably.
    mTimer.Stop();
    meState = ePaused;
}

/******************************************************************************
    OnResume
******************************************************************************/
void
CPlayerConnection::OnResume(
    const CPlayerCommand& cmd)
{
    meState = ePlaying;

    if (IsTrackBanned())
    {
        return;
    }

    if (!mLastSubmission.isScrobblable())
    {
        QString err = QT_TR_NOOP("RESUME received without a preceding START");
        LOG(2, err << "\n");
        throw BadCommandException(qPrintable(err));
    }

    // Allow resuming for tracks that are playing, just don't do anything
    if (!mbSubmitted)
    {
        // Only start timer off again if we haven't already submitted
        mTimer.Start(this);
    }

    meState = ePlaying;
}

/******************************************************************************
    GoalReached
******************************************************************************/
void
CPlayerConnection::GoalReached()
{
    if (settingsService()->currentUser()->logToProfile())
    {
        emit submissionReady(mLastSubmission);
    }
    else
    {
        LOG(3, "Scrobbling disabled, will not submit.\n");
    }

    // This bool is just to know whether to start timer again on a resume,
    // so we set it to true even if
    mbSubmitted = true;
}
