/***************************************************************************
 *   Copyright (C) 2005 - 2006 by                                          *
 *      Christian Muehlhaeuser, Last.fm Ltd <chris@last.fm>                *
 *      Erik Jaelevik, Last.fm Ltd <erik@last.fm>                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Steet, Fifth Floor, Boston, MA  02111-1307, USA.          *
 ***************************************************************************/

#include "progressframe.h"
#include <QPainter>

// progress bar colour
static const QColor k_activeTop = QColor( 0xd6, 0xde, 0xe6, 0xff );
static const QColor k_activeMiddle = QColor( 0xd0, 0xd9, 0xe2, 0xff );
static const QColor k_activeBottom = QColor( 0xca, 0xd4, 0xdc, 0xff );

// blue bg colour
static const QColor k_bkgrBlueTop = QColor( 0xeb, 0xf0, 0xf2, 0xff );
static const QColor k_bkgrBlueMiddle = QColor( 0xe5, 0xe9, 0xec, 0xff );
static const QColor k_bkgrBlueBottom = QColor( 0xdc, 0xe2, 0xe5, 0xff );

// grey bg colour
static const QColor k_bkgrGreyTop = QColor( 0xba, 0xba, 0xba, 0xff );
static const QColor k_bkgrGreyMiddle = QColor( 0xe2, 0xe2, 0xe2, 0xff );
static const QColor k_bkgrGreyBottom = QColor( 0xff, 0xff, 0xff, 0xff );

// scrobbled colour
static const QColor k_scrobbledTop = QColor( 0xba, 0xc7, 0xd7, 0xff );
static const QColor k_scrobbledMiddle = QColor( 0xb8, 0xc4, 0xd5, 0xff );
static const QColor k_scrobbledBottom = QColor( 0xb5, 0xc1, 0xd2, 0xff );


ProgressFrame::ProgressFrame( QWidget *parent ) :
    QFrame( parent ),
    m_drawBackground( true ),
    m_drawTime( false ),
    m_reverse( false ),
    m_timer( new QTimer( this ) ),
    m_timerEnabled( false ),
    m_watch( NULL ),
    m_green( false )
{
    m_timer->setInterval( 1000 );
    useInternalTimer();

    pushTimer.setSingleShot( true );
    connect( &pushTimer, SIGNAL( timeout() ), this, SLOT( popText() ) );

    m_maxValue = 0;
    m_value = 0;

    setMinimumHeight( 22 );
    setMinimumWidth( 22 );
    setFrameStyle( QFrame::StyledPanel | QFrame::Sunken );
    setLineWidth( 1 );
}


void
ProgressFrame::setText( const QString& text )
{
    m_text = text;
    m_baseText = text;
    update();
}


void
ProgressFrame::pushText( const QString& text )
{
    m_text = text;

    pushTimer.stop();
    pushTimer.start( 3000 );

    update();
}


void
ProgressFrame::popText()
{
    m_text = m_baseText;
    update();
}


void
ProgressFrame::useInternalTimer()
{
    disconnectWatch();

    if ( !m_timerEnabled )
    {
        connect( m_timer, SIGNAL( timeout() ), this, SLOT( updateTimer() ) );
        stop();
        m_drawTime = true;
        m_timerEnabled = true;
        m_timerText = "";
    }
}


void
ProgressFrame::useStopWatch( CStopWatch* watch )
{
    disableInternalTimer();
    disconnectWatch();

    m_watch = watch;
    connect( m_watch, SIGNAL( valueChanged( int ) ),
             this,    SLOT( setValue( int ) ) );

    setMaximum( m_watch->GetTimeOut() );
    setValue( m_watch->GetTime() );

    m_drawTime = true;
    m_timerText = "";
}


void
ProgressFrame::setTimerText( QString text )
{
    m_timerText = text;
    update();
}


void
ProgressFrame::setEnabled( bool enabled )
{
    m_drawTime = enabled;
    update();
}


void
ProgressFrame::disableInternalTimer()
{
    if ( m_timerEnabled )
    {
        stop();
        disconnect( m_timer, SIGNAL( timeout() ), this, SLOT( updateTimer() ) );
        m_timerEnabled = false;
    }
}


void
ProgressFrame::disconnectWatch()
{
    if ( m_watch != NULL )
    {
        disconnect( m_watch, SIGNAL( valueChanged( int ) ),
                    this,    SLOT( setValue( int ) ) );
        m_watch = NULL;
    }
}


void
ProgressFrame::paintEvent( QPaintEvent * event )
{
    QRect rect = event->rect();
    rect.adjust( 8, 0, -8, 0 );

    QColor fgTop = k_activeTop;
    QColor fgMiddle = k_activeMiddle;
    QColor fgBottom = k_activeBottom;

    QColor bgTop = k_bkgrBlueTop;
    QColor bgMiddle = k_bkgrBlueMiddle;
    QColor bgBottom = k_bkgrBlueBottom;

    if ( m_green )
    {
        fgTop = k_scrobbledTop;
        fgMiddle = k_scrobbledMiddle;
        fgBottom = k_scrobbledBottom;
    }

    QLinearGradient gradient( 0, 0, 0, height() );
    if ( !m_drawBackground && !qApp->arguments().contains( "--sanity" ) )
    {
        bgTop = k_bkgrGreyTop;
        bgMiddle = k_bkgrGreyMiddle;
        bgBottom = k_bkgrGreyBottom;

        gradient.setColorAt( 0, bgTop );
        gradient.setColorAt( 0.07, bgMiddle );
        gradient.setColorAt( 0.14, bgBottom );
        gradient.setColorAt( 1, bgBottom );

    }
    else
    {
        gradient.setColorAt( 0, bgTop );
        gradient.setColorAt( 0.5, bgMiddle );
        gradient.setColorAt( 0.51, bgBottom );
        gradient.setColorAt( 1, bgBottom );
    }

    QLinearGradient gradientActive( 0, 0, 0, height() );
    gradientActive.setColorAt( 0, fgTop );
    gradientActive.setColorAt( 0.5, fgMiddle );
    gradientActive.setColorAt( 0.51, fgBottom );
    gradientActive.setColorAt( 1, fgBottom );

    QPainter painter( this );
    drawFrame( &painter );

    int w = 0;
    if ( m_drawTime )
    {
        if ( m_maxValue > 0 )
        {
            w = (int)( ( (float)m_value / (float)m_maxValue ) * ( width() - 2 ) );
            if ( w > width() - 2 )
                w = width() - 2;
        }

        // active progressbar
        QRect r( 1, 1, w, height() - 2 );
        painter.fillRect( r, gradientActive );
    }

    // background
    QRect r( w + 1, 1, width() - 2 - w, height() - 2 );
    painter.fillRect( r, gradient );

    // icon if one is set
    if ( !icon().isNull() )
    {
        QRect r = rect;
        r.adjust( 0, ( rect.height() - icon().height() ) / 2, 0, 0 );
        r.setSize( icon().size() );
        painter.drawPixmap( r, icon() );

        rect.adjust( icon().width() + 6, 1, 0, 0 );
    }

    // texts on top
    painter.setPen( Qt::black );
    painter.setBrush( Qt::black );

    QString timeString;
    if ( m_drawTime && maximum() != 0 )
    {
        bool endReached = ( value() >= maximum() ) && ( maximum() != -1 );
        int time = m_reverse ? maximum() - value() : value();

        if ( endReached )
        {
            if ( !m_timerText.isEmpty() )
            {
                timeString = m_timerText;
                painter.drawText( rect, Qt::AlignRight | Qt::AlignVCenter, m_timerText );
            }
        }
        else
        {
            QString hrs  = QString::number( ( time / 3600 ) > 0 ? ( time / 3600 ) : 0 );
            QString mins = QString::number( ( ( time % 3600 ) / 60 ) > 0 ? ( ( time % 3600 ) / 60 ) : 0 );
            QString secs = QString::number( ( ( time % 3600 ) % 60 ) > 0 ? ( ( time % 3600 ) % 60 ) : 0 );
            timeString = QString( "%1%2%3:%4" )
                            .arg( m_reverse ? "- " : "" )
                            .arg( hrs == "0" ? "" : hrs + ":" )
                            .arg( hrs == "0" ? mins : ( mins.length() < 2 ? "0" + mins : mins ) )
                            .arg( secs.length() < 2 ? "0" + secs : secs );

            painter.drawText( rect, Qt::AlignRight | Qt::AlignVCenter, timeString );
        }
    }
    else
    {
        if ( !m_timerText.isEmpty() )
        {
            timeString = m_timerText;
            painter.drawText( rect, Qt::AlignRight | Qt::AlignVCenter, m_timerText );
        }
    }

    rect.adjust( 0, 0, -( painter.fontMetrics().boundingRect( timeString ).width() + 8 ), 0 );
    painter.drawText( rect, Qt::AlignLeft | Qt::AlignVCenter, text() );
}


void
ProgressFrame::setValue( int value )
{
    m_value = value;
    update();
}

void
ProgressFrame::setNewTrack( const MetaData& track )
{
    setTimerText( "" );
    setText( QString( "%1 %2 %3" )
        .arg( track.artist() )
        .arg( track.artist().isEmpty() ? "" : QString(" ") + QChar(8211) + QString(" ") ) // en dash
        .arg( track.track() ) );
}

void
ProgressFrame::mouseMoveEvent( QMouseEvent *event )
{
    if ( !( event->buttons() & Qt::LeftButton ) )
        return;

    QString anchor = text();
    if ( !anchor.isEmpty() )
    {
        QDrag *drag = new QDrag( this );

        QMimeData *mimeData = new QMimeData();
        mimeData->setData( "item/type", QByteArray::number( m_itemType ) );

        QHash<QString, QString> data = itemData();
        if ( data.count() )
        {
            for ( int i = 0; i < data.count(); i++ )
            {
                qDebug() << "Setting data" << data.keys().at( i ) << data.values().at( i );
                mimeData->setData( QString( "item/%1" ).arg( data.keys().at( i ) ), data.values().at( i ).toUtf8() );
            }
        }
        else
            return;

        mimeData->setText( mimeData->data( "item/artist" ) + " - " + mimeData->data( "item/track" ) );

        QPainter painter;
        QPixmap pixmap( painter.fontMetrics().width( anchor ) + 16, painter.fontMetrics().height() + 4 );
        QRect rect( 0, 0, pixmap.width() - 1, pixmap.height() - 1 );

        painter.begin( &pixmap );
        painter.setBackgroundMode( Qt::OpaqueMode );

        painter.setBrush( Qt::white );
        painter.setPen( Qt::black );
        painter.drawRect( rect );

        painter.setPen( Qt::black );
        painter.drawText( rect, Qt::AlignCenter, anchor );
        painter.end();

        drag->setMimeData( mimeData );
        drag->setPixmap( pixmap );

        Qt::DropAction dropAction = drag->start( Qt::CopyAction );
    }
}
