/***************************************************************************
 *   Copyright (C) 2005 - 2006 by                                          *
 *      Christian Muehlhaeuser, Last.fm Ltd <chris@last.fm>                *
 *      Erik Jaelevik, Last.fm Ltd <erik@last.fm>                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Steet, Fifth Floor, Boston, MA  02111-1307, USA.          *
 ***************************************************************************/

#include <QtGui>

#include "rtaudioplayback.h"
#include "containerutils.h"
#include "Loqqer.h"

float RTAudioPlayback::m_volume = 0.5;

QByteArray m_buffer;
QMutex mutex;

int
audioCallback( char *buffer, int bufferSize, void *data )
{
    if ( buffer == NULL || bufferSize != 512 )
        return 0;

    int bufs = bufferSize * 4;
    memset( buffer, 0, bufs );

    mutex.lock();
    if ( m_buffer.size() > bufs )
    {

        // Apply volume scaling
        for ( int i = 0; i < bufs / 2; i++ )
        {
            union PCMDATA
            {
                short i;
                unsigned char b[2];
            } pcmData;

            pcmData.b[0] = m_buffer[i * 2];
            pcmData.b[1] = m_buffer[i * 2 + 1];

            float pcmValue = (float)pcmData.i * RTAudioPlayback::m_volume;
            pcmData.i = (short)pcmValue;

            buffer[i * 2] = pcmData.b[0];
            buffer[i * 2 + 1] = pcmData.b[1];
        }

        m_buffer.remove( 0, bufs );
    }
    mutex.unlock();

    return 0;
}


RTAudioPlayback::RTAudioPlayback() :
    m_audio( 0 ),
    m_timer( new QTimer( this ) )
{
    gLogger.Init( savePath( "playback.log" ), false );
    gLogger.SetLevel( 4 );

    LOGL( 3, "Initialising RTAudio Playback" );
    initAudio();

    connect( m_timer, SIGNAL( timeout() ), this, SLOT( checkBuffer() ) );
}


QStringList
RTAudioPlayback::soundSystems()
{
    QStringList l;

    #ifdef WIN32
        l << "DirectSound";
    #endif

    #ifdef Q_WS_X11
        l << "Alsa";
    #endif

    #ifdef Q_WS_MAC
        l << "CoreAudio";
    #endif

    return l;
}


QStringList
RTAudioPlayback::devices()
{
	qDebug() << "Querying audio devices";
    QStringList l;

    if ( !m_audio )
        return l;

    try
    {
		qDebug() << "Device nums:" << m_audio->getDeviceCount();

        for ( int i = 1; i <= m_audio->getDeviceCount(); i++ )
        {
            RtAudioDeviceInfo info;
            info = m_audio->getDeviceInfo( i );
			qDebug() << "Device name:" << QString::fromStdString( info.name );

			if ( info.outputChannels > 0 )
            {
                l << QString::fromStdString( info.name ); // check if we can make it utf8 compatible
            }
        }
    }
    catch ( RtError &error )
    {
        LOGL( 1, "Getting device names failed. RtAudio error type: " << error.getType() <<
                 " Message: " << error.getMessage() );
    }

    return l;
}


void
RTAudioPlayback::startPlayback()
{
    if ( !m_audio )
    {
        QMessageBox::critical( qApp->activeWindow(), tr("Audio Error"), tr("No soundcard available."));
        return;
    }

    char data[2];

    try
    {
        if ( !m_timer->isActive() )
        {
            m_timer->start( 15 );
            m_audio->setStreamCallback( &audioCallback, (void *)data );
        }
        m_audio->startStream();
    }
    catch ( RtError &error )
    {
        LOGL( 1, "Starting stream failed. RtAudio error type: " << error.getType() <<
                 " Message: " << error.getMessage() );
    }
}


void
RTAudioPlayback::stopPlayback()
{
    if ( !m_audio )
    {
        return;
    }

    m_audio->stopStream();
    m_audio->cancelStreamCallback();
    m_timer->stop();

    mutex.lock();
    m_buffer.clear();
    mutex.unlock();
}


void
RTAudioPlayback::initAudio()
{
    int channels = 2;
    int nBuffers = 8;
    int bufferSize = 512;
    int sampleRate = 44100;

    try
    {
        RtAudio::RtAudioApi api = RtAudio::UNSPECIFIED;
        RtAudioFormat format = RTAUDIO_SINT16;
        m_audio = new RtAudio();

        int card = internalSoundCardID( settingsService()->soundCard() );
        LOGL( 3, "Card ID " << settingsService()->soundCard() << " - translated to internal ID " << card );

        #ifdef Q_WS_X11
        api = RtAudio::LINUX_ALSA;
        #endif

        RtAudioDeviceInfo info = m_audio->getDeviceInfo( card );
        delete m_audio;

        if ( info.nativeFormats & RTAUDIO_SINT32 )
        {
            LOGL( 3, "Using format RTAUDIO_SINT32" );
            format = RTAUDIO_SINT32;
        }
        if ( info.nativeFormats & RTAUDIO_SINT24 )
        {
            LOGL( 3, "Using format RTAUDIO_SINT24" );
            format = RTAUDIO_SINT24;
        }
        if ( info.nativeFormats & RTAUDIO_SINT16 )
        {
            LOGL( 3, "Using format RTAUDIO_SINT16" );
            format = RTAUDIO_SINT16;
        }

        m_audio = new RtAudio( card, channels, 0, 0, format, sampleRate, &bufferSize, nBuffers, api );
    }
    catch ( RtError &error )
    {
        LOGL( 1, "Initialising RtAudio failed. RtAudio error type: " << error.getType() <<
                 " Message: " << error.getMessage() );

        // Don't delete m_audio here or it will crash
        m_audio = 0;

        return;
    }

    LOGL( 1, "RtAudio successfully initialised." );
}


void
RTAudioPlayback::dataAvailable( const QByteArray &buffer )
{
    mutex.lock();
    m_buffer.append( buffer );
    mutex.unlock();
}


void
RTAudioPlayback::clearBuffers()
{
    mutex.lock();
    m_buffer.clear();
    mutex.unlock();
}


void
RTAudioPlayback::checkBuffer()
{
    // NOTE: this is wavedata! 16kb of wavedata aprox. equals 1.6kb of mp3
    if ( m_buffer.size() < 16384 * 12 )
        emit needData();
}


int
RTAudioPlayback::internalSoundCardID( int settingsID )
{
    if ( !m_audio )
        return -1;

    if ( settingsID < 0 )
        settingsID = 0;

    try
    {
        int card = 0;

        for ( int i = 1; i <= m_audio->getDeviceCount(); i++ )
        {
            RtAudioDeviceInfo info;
            info = m_audio->getDeviceInfo( i );
            if ( info.outputChannels > 0 )
            {
                if ( card++ == settingsID )
                    return i;
            }
        }
    }
    catch ( RtError &error )
    {
        LOGL( 1, "Getting internal soundcard ID failed. RtAudio error type: " << error.getType() <<
                 " Message: " << error.getMessage() );
    }

    #ifdef Q_WS_MAC
        return 3;
    #endif
    return 1;
}


Q_EXPORT_PLUGIN2( playback, RTAudioPlayback )
