/***************************************************************************
 *   Copyright (C) 2005 - 2006 by                                          *
 *      Christian Muehlhaeuser, Last.fm Ltd <chris@last.fm>                *
 *      Erik Jaelevik, Last.fm Ltd <erik@last.fm>                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Steet, Fifth Floor, Boston, MA  02111-1307, USA.          *
 ***************************************************************************/

#include <QtGui>
#include <QPixmap>
#include <QHttp>

#include "searchextension.h"
#include <containerutils.h>
#include <webserviceinterface.h>

#include <math.h>

// These are pixel sizes
#ifdef Q_WS_MAC
static const int k_fontMin = 13;
#else
static const int k_fontMin = 11;
#endif
static const int k_fontMax = 24;
static const int k_lineHeight = 29;

static const int k_maxItemsTopTags = 35;
static const int k_maxItems = 35;

static const QColor k_tagFontColour = QColor( 0x6d, 0x83, 0xa2 );

SearchExtension::SearchExtension() :
    m_widget( 0 ),
    m_type( 0 ), // set artist search as default, despite presenting the toptags
    m_currentCloudType( 1 ), // tags in cloud
    m_topTags( false )
{
    qDebug() << "Initialising Search Extension";

    m_iWebService = webService();
    m_iSettings = settingsService();
    m_iInput = inputService();

    connect( m_iWebService, SIGNAL( handshakeResult( int ) ),
             this,            SLOT( topTags() ) );
    connect( m_iWebService, SIGNAL( similarArtistsResult( const QList<QPair<QString, int> >&, QString ) ),
             this,            SLOT( searchResults( const QList<QPair<QString, int> >&, QString ) ) );
    connect( m_iWebService, SIGNAL( similarTagsResult( const QList<QPair<QString, int> >& ) ),
             this,            SLOT( searchResults( const QList<QPair<QString, int> >& ) ) );
    connect( m_iWebService, SIGNAL( topTagsResult( const QList<QPair<QString, int> >& ) ),
             this,            SLOT( searchResults( const QList<QPair<QString, int> >& ) ) );
}


QString
SearchExtension::name() const
{
    return QString( "Search Extension" );
}


QString
SearchExtension::version() const
{
    return QString( "0.0.1" );
}


/*
    On the Mac, we need to put a spacer above the scroll area to reduce the
    ugliness of the scrollbar not going all the way up to the line. On
    Windows we just set the spacer to 0.
*/
void
SearchExtension::initGui()
{
    qDebug() << "Initialising Search GUI";

    // Full widget to return
    m_widget = new QWidget( m_parent );

    // The widget inside the scroll area
    WatermarkWidget* searchExtWidget = new WatermarkWidget( m_widget );
    searchExtWidget->setWatermark( dataPath( "watermark.png" ) );
    
    ui.setupUi( searchExtWidget );
    
    QPalette palette( QTabWidget().palette() );
    palette.setColor( QPalette::Base, palette.window().color() );

    // Tag cloud
    ui.resultBrowser->setPalette( palette );
    ui.resultBrowser->setItems( QStringList() );
    ui.resultBrowser->setUniformLineHeight( k_lineHeight );
    ui.resultBrowser->setItemsSelectable( true );
    ui.resultBrowser->setAlignment( Qt::AlignLeft | Qt::AlignBottom );
    ui.resultBrowser->setJustified( true );

    // Top search widgets
    ui.statusLabel->setText( tr( "Generating popular tags..." ) );
    ui.searchType->setCurrentIndex( m_type );
    ui.searchType->setEnabled( false );
    ui.searchButton->setEnabled( false );
    ui.playButton->setEnabled( false );

    ui.searchEdit->setText( tr( "Find a station by" ) );
    ui.searchEdit->installEventFilter( this );
    QPalette defPal = ui.searchEdit->palette();
    defPal.setColor( QPalette::Text, QColor( 100, 100, 100 ) );
    ui.searchEdit->setPalette( defPal );
    m_clearSearchOnClick = true;

    m_spinnerMovie.setFileName( dataPath( "progress.mng" ) );
    m_spinnerMovie.start();
    ui.spinnerLabel->setMovie( &m_spinnerMovie );
    //ui.spinnerLabel->hide();
    
    #ifdef Q_WS_MAC
        // Hack some minsizes for the buttons as Qt is being gay on the Mac
	ui.searchButton->setFixedWidth( ui.searchButton->sizeHint().width() + 0 );
	ui.playButton->setFixedWidth( ui.playButton->sizeHint().width() + 0 );
    #endif

    // Scroll area
    m_scrollArea = new QScrollArea( m_widget );

    QPalette p = m_scrollArea->palette();
    p.setColor( QPalette::Window, QColor( 0xff, 0xff, 0xff ) );
    m_scrollArea->setPalette( p );
    m_scrollArea->setHorizontalScrollBarPolicy( Qt::ScrollBarAsNeeded );
    m_scrollArea->setVerticalScrollBarPolicy( Qt::ScrollBarAsNeeded );
    m_scrollArea->setFrameStyle( QFrame::NoFrame );
    m_scrollArea->setWidgetResizable( true );
    m_scrollArea->setWidget( searchExtWidget );

    // Create a vertical layout onto which we'll add a spacer (for Mac) and the
    // scroll area.
    QVBoxLayout* vbl = new QVBoxLayout( m_widget );
    vbl->setMargin( 0 );
    vbl->setSpacing( 0 );
    #ifdef Q_WS_MAC
        // Add space to the spacer above the scroll area and remove the same
        // amount from the one below.
        QSpacerItem* macSpace = new QSpacerItem( 1, 30, QSizePolicy::Fixed, QSizePolicy::Minimum );
        vbl->addItem( macSpace );
        ui.spacerItem->changeSize( 1, 0, QSizePolicy::Fixed, QSizePolicy::Minimum );
    #endif
    vbl->addWidget( m_scrollArea );
    
    m_widget->setLayout( vbl );
    
    connect( ui.searchEdit,    SIGNAL( returnPressed() ),
             ui.searchButton,  SLOT( animateClick() ) );
    connect( ui.searchEdit,    SIGNAL( textChanged( QString ) ),
             this,             SLOT( searchFieldChanged() ) );
    connect( ui.searchButton,  SIGNAL( clicked() ),
             this,             SLOT( search() ) );
    connect( ui.playButton,    SIGNAL( clicked() ),
             this,             SLOT( play() ) );
    connect( ui.resultBrowser, SIGNAL( clicked( int ) ),
             this,             SLOT( itemClicked( int ) ) );
}


QWidget*
SearchExtension::gui()
{
    if ( !m_widget )
        initGui();

    return m_widget;
}

bool
SearchExtension::eventFilter( QObject* watched, QEvent* event )
{
    if ( watched == ui.searchEdit )
    {
        if ( event->type() == QEvent::MouseButtonPress && m_clearSearchOnClick )
        {
            ui.searchEdit->selectAll();
            restoreSearchField();
            return true;
        }
        else
        {
            return false;
        }
    }
    else
    {
        // Pass the event on to the parent class
        return ExtensionInterface::eventFilter( watched, event );
    }
}

void
SearchExtension::topTags()
{
    m_iWebService->topTags();
    m_topTags = true;
}

void
SearchExtension::restoreSearchField()
{
    QPalette defPal = ui.searchEdit->palette();
    defPal.setColor( QPalette::Text, QColor( 0, 0, 0 ) );
    ui.searchEdit->setPalette( defPal );
    m_clearSearchOnClick = false;
}

void
SearchExtension::searchFieldChanged()
{
    ui.playButton->setEnabled( !ui.searchEdit->text().isEmpty() );
}

void
SearchExtension::search()
{
    if ( m_type != ui.searchType->currentIndex() )
    {
        // search mode changed, wipe selections
        clearSelection();
        ui.resultBrowser->clear();
        m_type = ui.searchType->currentIndex();
    }

    ui.searchButton->setEnabled( false );
    ui.searchType->setEnabled( false );
    ui.resultBrowser->setItemsSelectable( false );
    ui.spinnerLabel->show();

    switch( m_type )
    {
        case 0:
            ui.statusLabel->setText( tr( "Generating similar artists..." ) );
            ui.resultBrowser->setItemType( ItemArtist );
            m_iWebService->similarArtists( ui.searchEdit->text() );
            break;

        case 1:
            ui.statusLabel->setText( tr( "Generating similar tags..." ) );
            ui.resultBrowser->setItemType( ItemTag );
            m_iWebService->similarTags( ui.searchEdit->text() );
            break;
    }

    m_lastSearch = ui.searchEdit->text();
    m_searching = true;
}


QString
SearchExtension::stationUrl()
{
    QString s, url;

    s = QString( QUrl::toPercentEncoding( ui.searchEdit->text() ) );

    switch ( m_type )
    {
        case 0:
            url = "lastfm://artist/" + s + "/similarartists";
            break;

        case 1:
            url = "lastfm://globaltags/" + s;
            break;
    }

    return url;
}


void
SearchExtension::searchResults( const QList<QPair<QString, int> >& items, const QString& searchToken )
{
    int total = k_maxItems;
    if ( m_topTags )
    {
        // We want more tags for the initial top tags cloud
        m_topTags = false;
        total = k_maxItemsTopTags;
    }
    else
    {
        m_currentCloudType = m_type;
    }

    QList<QPair<QString, int> > sortedAlpha = items.mid( 0, total );

    // They are still sorted by weight when they get here so pos 0 will have the biggest tag
    int maxWeight = sortedAlpha.isEmpty() ? 0 : sortedAlpha.first().second;
    int minWeight = sortedAlpha.size() < 2 ? 0 : sortedAlpha.last().second;

    qSort( sortedAlpha.begin(), sortedAlpha.end() );

    ui.resultBrowser->clear();
    for ( int i = 0; i < sortedAlpha.count(); i++ )
    {
        QString name = sortedAlpha.at( i ).first;
        ui.resultBrowser->append( name );

        int weight = sortedAlpha.at( i ).second;
        
        float normalisedWeight =
            (float)(weight - minWeight) / qMax( 1, maxWeight - minWeight );
        
        int fontSize =
            normalisedWeight * ( k_fontMax - k_fontMin ) + k_fontMin + 0.5f;

        QFont f( "Arial" );
        f.setBold( true );
        f.setPixelSize( fontSize );

        ui.resultBrowser->setItemFont( i, f );
        ui.resultBrowser->setItemColor( i, k_tagFontColour );

        QHash<QString, QString> data;
        data.insert( "artist", name ); // eh?
        ui.resultBrowser->setItemData( i, data );
    }

    ui.searchButton->setEnabled( true );
    ui.searchType->setEnabled( true );
    ui.resultBrowser->setItemsSelectable( true );
    ui.statusLabel->clear();
    ui.spinnerLabel->hide();

    if ( items.count() )
    {
        if ( !searchToken.isEmpty() )
        {
            // This will happen for tags
            m_lastSearch = searchToken;
        }

        switch ( m_currentCloudType )
        {
            case 0:
                ui.resultBrowser->setJustified( false );
                break;

            case 1:
                ui.resultBrowser->setJustified( true );
                break;
        }
    }
    else
    {
        ui.playButton->setEnabled( false );

        ui.statusLabel->setText( tr( "Sorry, your search didn't return any results." ) );
    }

    m_searching = false;
}

void
SearchExtension::itemClicked( int index )
{
    if ( m_searching )
    {
        return;
    }

    QString item = ui.resultBrowser->items().at( index );
    
    if ( m_currentCloudType != ui.searchType->currentIndex() )
    {
        ui.searchType->setCurrentIndex( m_currentCloudType );
    }

    ui.searchEdit->setText( item );
    restoreSearchField();

    search();

    emit stationFocused( stationUrl() );
}

void
SearchExtension::clearSelection()
{
    ui.resultBrowser->clearSelections();
}

void
SearchExtension::play()
{
    m_type = ui.searchType->currentIndex();
    QString station = stationUrl();
    if ( !station.isEmpty() )
    {
        m_iInput->load( station );
    }
}
    


Q_EXPORT_PLUGIN2( extension_search, SearchExtension )
