/***************************************************************************
 *   Copyright (C) 2005 - 2006 by                                          *
 *      Christian Muehlhaeuser, Last.fm Ltd <chris@last.fm>                *
 *      Erik Jaelevik, Last.fm Ltd <erik@last.fm>                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Steet, Fifth Floor, Boston, MA  02111-1307, USA.          *
 ***************************************************************************/

#include "containerutils.h"
#include "settingsservice.h"
#include "Loqqer.h"

#ifdef Q_WS_MAC
#include <ApplicationServices/ApplicationServices.h>
#endif

#include <QApplication>

#include <list>

using namespace std;

/******************************************************************************
    UserSettingsService
******************************************************************************/
UserSettingsService::UserSettingsService( QString username )
{
    Q_ASSERT( username != "" );
    loadUsername( username );
}

void
UserSettingsService::loadUsername( QString username )
{
    m_username = username;
    if ( !m_username.isEmpty() )
    {
        #ifdef WIN32
            m_config = new QSettings(
                QString( "HKEY_CURRENT_USER\\Software\\Last.fm\\Client\\Users\\%1" ).arg( username ),
                QSettings::NativeFormat,
                this );
        #else
            m_config = new QSettings( QSettings::IniFormat, QSettings::UserScope, "Last.fm", "Users", this );
            m_config->beginGroup( QString( "%1" ).arg( username ) );
        #endif
    }
}


QString
UserSettingsService::password() const
{
    return m_config->value( "Password" ).toString();
}


void
UserSettingsService::setPassword( QString password )
{
    if ( !password.isEmpty() && password != "********" )
    {
        password = MD5Digest( password.toUtf8() );
        m_config->setValue( "Password", password );
        emit userChanged( username() );
    }
}


UserIconColour
UserSettingsService::icon() const
{
    // This will return eRed if there is no entry. Don't want that.
    if ( m_config->contains( "Icon" ) )
    {
        int icon = m_config->value( "Icon" ).toInt();
        return static_cast<UserIconColour>(icon);
    }
    else
    {
        return eNone; 
    }
}


void
UserSettingsService::setIcon( UserIconColour colour )
{
    m_config->setValue( "Icon", static_cast<int>(colour) );
    emit userChanged( username() );
}


// LogToProfile and RememberPass are written as ints for backwards
// compatibility with the MFC Audioscrobbler
bool
UserSettingsService::logToProfile() const
{
    return static_cast<bool>(m_config->value( "LogToProfile", 1 ).toInt());
}


void
UserSettingsService::setLogToProfile( bool state )
{
    m_config->setValue( "LogToProfile", static_cast<int>(state) );
    emit userChanged( username() );
}


bool
UserSettingsService::sidebarEnabled() const
{
    return m_config->value( "SidebarEnabled", false ).toBool();
}


void
UserSettingsService::setSidebarEnabled( bool state )
{
    m_config->setValue( "SidebarEnabled", state );
    emit userChanged( username() );
}


bool
UserSettingsService::rememberPass() const
{
    return static_cast<bool>(m_config->value( "RememberPass", true ).toInt());
}


void
UserSettingsService::setRememberPass( bool remember )
{
    m_config->setValue( "RememberPass", static_cast<int>(remember) );
    emit userChanged( username() );
}


QString
UserSettingsService::stationName( QString url ) const
{
    QString name = m_config->value( "StationNames/" + url ).toString();
    return ( name.isNull() || name.isEmpty() ) ? url : name;
}


void
UserSettingsService::setStationName( QString url, QString name )
{
    m_config->setValue( "StationNames/" + url, name );
    m_config->sync();

    emit userChanged( username() );
    emit historyChanged();
}


void
UserSettingsService::addRecentStation( QString url, QString name )
{
    QStringList l = recentStations();
    l.insert( 0, url );

    clearRecentStations( false );
    m_config->sync();

    m_config->beginGroup( "RecentStations" );
    for ( int i = 0; i < 10 && i < l.count(); i++ )
    {
        if ( l.at( i ) != url || i == 0 )
            m_config->setValue( QString::number( i ), l.at( i ) );
    }
    m_config->endGroup();

    m_config->setValue( "StationNames/" + url, name );
    m_config->sync();

    emit userChanged( username() );
}


void
UserSettingsService::clearRecentStations( bool emitting )
{
    m_config->remove( "RecentStations" );

    if ( emitting )
        emit historyChanged();
}


QStringList
UserSettingsService::recentStations() const
{
    m_config->beginGroup( "RecentStations" );

    QStringList items;
    QStringList keys = m_config->allKeys();
    for ( int i = 0; i < keys.count(); i++ )
    {
        QString s = m_config->value( keys.at( i ) ).toString();
        if ( !s.isEmpty() )
            items << s;
    }

    m_config->endGroup();

    return items;
}


void
UserSettingsService::save()
{
    m_config->sync();
}


void
UserSettingsService::setExcludedDirs( QStringList dirs )
{
    m_config->setValue( "ExclusionDirs", dirs );
    emit userChanged( username() );
}


QStringList
UserSettingsService::excludedDirs() const
{
    return m_config->value( "ExclusionDirs" ).toStringList();
}


void
UserSettingsService::setIncludedDirs( QStringList dirs )
{
    m_config->setValue( "InclusionDirs", dirs );
    emit userChanged( username() );
}


QStringList
UserSettingsService::includedDirs() const
{
    return m_config->value( "InclusionDirs" ).toStringList();
}

void
UserSettingsService::setMetadataEnabled( bool enabled )
{
    m_config->setValue( "DownloadMetadata", enabled );
    emit userChanged( username() );
}

bool
UserSettingsService::metadataEnabled()
{
    return m_config->value( "DownloadMetadata", true ).toBool();
}

void
UserSettingsService::setCrashReportingEnabled( bool enabled )
{
    m_config->setValue( "ReportCrashes", enabled );
    emit userChanged( username() );
}

bool
UserSettingsService::crashReportingEnabled()
{
    return m_config->value( "ReportCrashes", true ).toBool();
}


/******************************************************************************
    SettingsService
******************************************************************************/
SettingsService::SettingsService()
{
    LOG( 3, "Initialising Settings Service" );

    #ifdef WIN32
        m_config = new QSettings( "HKEY_CURRENT_USER\\Software\\Last.fm\\Client", QSettings::NativeFormat, this );
        m_configLM = new QSettings( "HKEY_LOCAL_MACHINE\\Software\\Last.fm\\Client", QSettings::NativeFormat, this );
        m_configUsers = new QSettings( "HKEY_CURRENT_USER\\Software\\Last.fm\\Client\\Users", QSettings::NativeFormat, this );
        m_configPlugins = new QSettings( "HKEY_LOCAL_MACHINE\\Software\\Last.fm\\Client\\Plugins", QSettings::NativeFormat, this );
        m_configMediaDevices = new QSettings( "HKEY_CURRENT_USER\\Software\\Last.fm\\Client\\MediaDevices", QSettings::NativeFormat, this );
    #else
        m_config = new QSettings( QSettings::IniFormat, QSettings::UserScope, "Last.fm", "Client", this );
        m_configLM = new QSettings( QSettings::IniFormat, QSettings::UserScope, "Last.fm", "Client", this );
        m_configUsers = new QSettings( QSettings::IniFormat, QSettings::UserScope, "Last.fm", "Users", this );
        m_configPlugins = new QSettings( QSettings::IniFormat, QSettings::UserScope, "Last.fm", "Plugins", this );
        m_configMediaDevices = new QSettings( QSettings::IniFormat, QSettings::UserScope, "Last.fm", "MediaDevices", this );
    #endif

    loadUsers();

    // Write out application path every time
    QString path = QApplication::applicationFilePath();
    m_config->setValue( "Path", path );
}


void
SettingsService::loadUsers()
{
    foreach( QString username, allUsers() )
    {
        UserSettingsService* user = new UserSettingsService( username );
        m_userList.insert( username, user );
        connect( m_userList[username], SIGNAL( userChanged( QString ) ),
                 this,                 SLOT  ( userChanged( QString ) ) );
    }
}


SettingsService::~SettingsService()
{
    QHash<QString, UserSettingsService*>::const_iterator i = m_userList.constBegin();
    while ( i != m_userList.constEnd() )
    {
        UserSettingsService* user = i.value();
        delete user;
        ++i;
    }
}


UserSettingsServiceInterface*
SettingsService::user( QString username )
{
    if ( existingUser( username ) )
    {
        return m_userList[username];
    }
    else
    {
        UserSettingsService* newUser = new UserSettingsService( username );
        m_userList.insert( username, newUser );
        connect( m_userList[username], SIGNAL( userChanged( QString ) ),
                 this,                 SLOT  ( userChanged( QString ) ) );

        return newUser;
    }
}


bool
SettingsService::deleteUser( QString username )
{
    if ( existingUser( username ) )
    {
        UserSettingsServiceInterface* user = m_userList.value( username );
        m_userList.remove( username );
        delete user;

        m_configUsers->remove( username );
        return true;
    }
    else
    {
        return false;
    }
}


int
SettingsService::externalSoundSystem()
{
    int externalSystem = -1;
    #ifdef WIN32
    externalSystem = 1;
    #endif
    #ifdef Q_WS_X11
    externalSystem = 2;
    #endif
    #ifdef Q_WS_MAC
    externalSystem = 1;
    #endif

    return externalSystem;
}


QString
SettingsService::browser()
{
    return m_config->value( "Browser" ).toString();
}


void
SettingsService::setBrowser( QString browser )
{
    m_config->setValue( "Browser", browser );
}


void
SettingsService::setDontAsk( const QString op, bool value )
{
    QString key = QString( "%1DontAsk" ).arg( op );
    m_config->setValue( key, value );
}

bool
SettingsService::dontAsk( const QString op )
{
    QString key = QString( "%1DontAsk" ).arg( op );
    return m_config->value( key ).toBool();
}

void
SettingsService::setAppLanguage( QString langCode )
{
    m_config->setValue( "AppLanguage", langCode );
}

QString
SettingsService::appLanguage()
{
    return m_config->value( "AppLanguage" ).toString();
}


void
SettingsService::save( bool restartConnection, bool restartAudio )
{
    m_config->sync();

    if ( restartConnection )
        emit reconnect();

    if ( restartAudio )
        emit resetAudio();
}


QStringList
SettingsService::allPlugins()
{
    QStringList list = m_configPlugins->childGroups();
    for (int i = 0; i < list.size(); ++i)
    {
        m_configPlugins->beginGroup(list.at(i));
        QString name = m_configPlugins->value("Name").toString();
        QString ver = m_configPlugins->value("Version").toString();
        list[i] = name + " " + tr("plugin, version") + " " + ver;
        m_configPlugins->endGroup();
    }

    return list;
}


QString
SettingsService::pluginVersion(
    QString pluginId )
{
    Q_ASSERT(pluginId.size() != 0);
    QString key(QString("%1/Version").arg(pluginId));
    QString version = m_configPlugins->value(key, "").toString();
    return version;
}


QString
SettingsService::pluginPlayerPath(
    QString pluginId )
{
    Q_ASSERT(pluginId.size() != 0);
    QString key(QString("%1/PlayerPath").arg(pluginId));
    QString path = m_configPlugins->value(key, "").toString();
    return path;
}


void
SettingsService::setPluginPlayerPath(
    QString pluginId,
    QString path )
{
    Q_ASSERT(pluginId.size() != 0);
    if (pluginId.size() == 0) { return; }

    QString key = QString("%1/PlayerPath").arg(pluginId);
    m_configPlugins->setValue(key, path);
}


QStringList
SettingsService::allMediaDevices()
{
    QStringList list = m_configMediaDevices->childGroups();
    return list;
}


QString
SettingsService::MediaDeviceUser( const QString& uid )
{
    m_configMediaDevices->beginGroup( m_configMediaDevices->childGroups().at( 0 ) );
    QString user = m_configMediaDevices->value( "user", "" ).toString();
    m_configMediaDevices->endGroup();

    return user;
}


void
SettingsService::addMediaDevice( const QString& uid, const QString& username )
{
    m_configMediaDevices->beginGroup( uid );
    m_configMediaDevices->setValue( "user", username );
    m_configMediaDevices->endGroup();

    m_configMediaDevices->sync();
}


UserIconColour
SettingsService::getFreeColour()
{
    list<int> listUnUsed;

    // Fill it with all colours
    for (int i = 0; i < 5; ++i)
    {
        listUnUsed.push_back(i);
    }

    QStringList users = m_configUsers->childGroups();

    // Remove the ones in use
    for (int i = 0; i < users.size(); ++i)
    {
        UserSettingsService user( users.at(i) );
        UserIconColour col = user.icon();
        if ( col != eNone )
        {
            listUnUsed.remove( static_cast<int>( col ) );
        }
        if (listUnUsed.size() == 0)
        {
            LOG(2, "We ran out of colours, returning random\n");
            return static_cast<UserIconColour>(rand() % 5);
        }
    }

    return static_cast<UserIconColour>(listUnUsed.front());
}


void
SettingsService::userChanged( QString username )
{
    if ( username == currentUsername() )
    {
        emit currentUserChanged( currentUser() );
    }
}

Q_EXPORT_PLUGIN2( service_settings, SettingsService )
