/***************************************************************************
 *   Copyright (C) 2005 - 2006 by                                          *
 *      Christian Muehlhaeuser, Last.fm Ltd <chris@last.fm>                *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Steet, Fifth Floor, Boston, MA  02111-1307, USA.          *
 ***************************************************************************/

#include <QtGui>

#include "sidebarextension.h"
#include "sidebartree.h"
#include "browserthread.h"
#include <containerutils.h>

#include "Loqqer.h"

SidebarExtension::SidebarExtension()
    : m_widget( 0 )
{
    qDebug() << "Initialising Sidebar Extension";

    gLogger.Init( savePath( "sidebar.log" ), false );
    gLogger.SetLevel( 4 );

    m_sidebarModel = new QStandardItemModel();
    m_delegate = new SideBarDelegate();

    m_iWebService = webService();
    m_iSettings = settingsService();
    m_iInput = inputService();

    VERIFY( connect( m_iWebService, SIGNAL( handshakeResult( int ) ),
                     this,            SLOT( updateAll() ) ) );
    VERIFY( connect( m_iWebService, SIGNAL( friendsResult( QString, QStringList ) ),
                     this,            SLOT( updateFriends( QString, QStringList ) ) ) );
    VERIFY( connect( m_iWebService, SIGNAL( neighboursResult( QString, const QList<QPair<QString, int> >& ) ),
                     this,            SLOT( updateNeighbours( QString, const QList<QPair<QString, int> >& ) ) ) );
    VERIFY( connect( m_iWebService, SIGNAL( userTagsResult( QString, const QList<QPair<QString, int> >& ) ),
                     this,            SLOT( updateUserTags( QString, const QList<QPair<QString, int> >& ) ) ) );
    VERIFY( connect( m_iWebService, SIGNAL( recentTracksResult( QString, QList< QPair<QString, QString> > ) ),
                     this,            SLOT( updateRecentTracks( QString, QList< QPair<QString, QString> > ) ) ) );
    VERIFY( connect( m_iWebService, SIGNAL( recentLovedTracksResult( QString, QList< QPair<QString, QString> > ) ),
                     this,            SLOT( updateRecentLovedTracks( QString, QList< QPair<QString, QString> > ) ) ) );
    VERIFY( connect( m_iWebService, SIGNAL( recentBannedTracksResult( QString, QList< QPair<QString, QString> > ) ),
                     this,            SLOT( updateRecentBannedTracks( QString, QList< QPair<QString, QString> > ) ) ) );
    VERIFY( connect( m_iWebService, SIGNAL( stationChanged( QString, QString ) ),
                     this,            SLOT( stationChanged( QString, QString ) ) ) );
    VERIFY( connect( m_iWebService, SIGNAL( stationNameChanged( QString, QString ) ),
                     this,            SLOT( stationName( QString, QString ) ) ) );
    VERIFY( connect( m_iWebService, SIGNAL( loveDone( QString, QString ) ),
                     this,            SLOT( addLovedTrack( QString, QString) ) ) );
    VERIFY( connect( m_iWebService, SIGNAL( banDone( QString, QString ) ),
                     this,            SLOT( addBannedTrack( QString, QString ) ) ) );
    VERIFY( connect( m_iWebService, SIGNAL( unlistenDone( QString, QString ) ),
                     this,            SLOT( removeListenedTrack( QString, QString) ) ) );
    VERIFY( connect( m_iWebService, SIGNAL( unloveDone( QString, QString ) ),
                     this,            SLOT( removeLovedTrack( QString, QString) ) ) );
    VERIFY( connect( m_iWebService, SIGNAL( unbanDone( QString, QString ) ),
                     this,            SLOT( removeBannedTrack( QString, QString ) ) ) );
    VERIFY( connect( m_iWebService, SIGNAL( deleteFriendDone( QString ) ),
                     this,            SLOT( removeFriend( QString ) ) ) );
}


SidebarExtension::~SidebarExtension()
{
    delete m_widget;
}


QString
SidebarExtension::name() const
{
    return QString( "Sidebar Extension" );
}


QString
SidebarExtension::version() const
{
    return QString( "0.0.1" );
}


void
SidebarExtension::initGui()
{
    qDebug() << "Initialising Sidebar GUI";

    m_widget = new QWidget( m_parent );
    ui.setupUi( m_widget );

    m_sidebarModel->clear();
    m_sidebarModel->insertRows( 0, 14 );
    m_sidebarModel->insertColumns( 0, 1 );

#ifdef Q_WS_MAC
    QPalette p = ui.treeWidget->palette();
    p.setColor( QPalette::Base, QColor( 0xe5, 0xed, 0xf7 ) );
    p.setColor( QPalette::Highlight, QColor( 0xa4, 0xb4, 0xcb ) );
    ui.treeWidget->setPalette( p );

    ui.treeWidget->setFrameStyle( QFrame::NoFrame );
    ui.treeWidget->setIconSize( QSize( 24, 24 ) );
#endif

    ui.treeWidget->header()->hide();

    ui.treeWidget->setRootIsDecorated( true );
    ui.treeWidget->setDragEnabled( false );
    ui.treeWidget->setAcceptDrops( true );
    ui.treeWidget->setAutoScroll( true );
    ui.treeWidget->setSelectionMode( QAbstractItemView::SingleSelection );
    ui.treeWidget->setSelectionBehavior( QAbstractItemView::SelectRows );

    QByteArray encUser = QUrl::toPercentEncoding( settingsService()->currentUsername() );

    for ( int i = 0; i < 14; i++ )
    {
        QString token;
        QString type;
        QString station;
        QIcon icon;
        int itemType;

        switch ( i )
        {
            case 0:
            {
                token = settingsService()->currentUsername();
                itemType = SidebarExtension::UserInfo;
                type = "userinfo";
            }
            break;

            case 1:
            {
                itemType = -1;
                type = "spacer";
            }
            break;

            case 2:
            {
                token = tr( "My Recommendations" );
                itemType = SidebarExtension::RecommendedRadio;
                type = "station";
                station = "lastfm://user/" + encUser + "/recommended";
                icon = QIcon( dataPath( "icons/recommended_radio.png" ) );
            }
            break;

            case 3:
            {
                token = tr( "Personal Radio" );
                itemType = SidebarExtension::PersonalRadio;
                type = "station";
                station="lastfm://user/" + encUser + "/personal";
                icon = QIcon( dataPath( "icons/personal_radio.png" ) );
            }
            break;

            case 4:
            {
                token = tr( "Loved Tracks Radio" );
                itemType = SidebarExtension::LovedRadio;
                type = "station";
                station = "lastfm://user/" + encUser + "/loved";
                icon = QIcon( dataPath( "icons/loved_radio.png" ) );
            }
            break;

            case 5:
            {
                token = tr( "Neighbourhood Radio" );
                itemType = SidebarExtension::NeighbourRadio;
                type = "station";
                station = "lastfm://user/" + encUser + "/neighbours";
                icon = QIcon( dataPath( "icons/neighbour_radio.png" ) );
            }
            break;

            case 6:
            {
                itemType = -1;
                type = "spacer";
            }
            break;

            case 7:
            {
                token = tr( "Recently Played" );
                itemType = SidebarExtension::RecentlyPlayed;
                icon = QIcon( dataPath( "icons/recent_tracks.png" ) );
            }
            break;

            case 8:
            {
                token = tr( "Recently Loved" );
                itemType = SidebarExtension::RecentlyLoved;
                icon = QIcon( dataPath( "icons/recently_loved.png" ) );
            }
            break;

            case 9:
            {
                token = tr( "Recently Banned" );
                itemType = SidebarExtension::RecentlyBanned;
                icon = QIcon( dataPath( "icons/recently_banned.png" ) );
            }
            break;

            case 10:
            {
                token = tr( "My Tags" );
                itemType = SidebarExtension::Tags;
                icon = QIcon( dataPath( "icons/my_tags.png" ) );
            }
            break;

            case 11:
            {
                token = tr( "Friends" );
                itemType = SidebarExtension::Friends;
                icon = QIcon( dataPath( "icons/my_friends.png" ) );
            }
            break;

            case 12:
            {
                token = tr( "Neighbours" );
                itemType = SidebarExtension::Neighbours;
                icon = QIcon( dataPath( "icons/my_neighbours.png" ) );
            }
            break;

            case 13:
            {
                token = tr( "History" );
                itemType = SidebarExtension::History;
                #ifdef Q_WS_MAC
                    icon = QIcon( dataPath( "icons/history32.png" ) );
                #else
                    icon = QIcon( dataPath( "icons/history16.png" ) );
                #endif
            }
            break;
        }

        m_sidebarModel->setData( m_sidebarModel->index( i, 0, QModelIndex() ), QVariant( token ), Qt::DisplayRole );
        m_sidebarModel->setData( m_sidebarModel->index( i, 0, QModelIndex() ), QVariant( type ), kTypeRole );
        m_sidebarModel->setData( m_sidebarModel->index( i, 0, QModelIndex() ), QVariant( icon ), Qt::DecorationRole );
        m_sidebarModel->setData( m_sidebarModel->index( i, 0, QModelIndex() ), QVariant( station ), kStationRole );
        m_sidebarModel->setData( m_sidebarModel->index( i, 0, QModelIndex() ), QVariant( itemType ), kItemTypeRole );
    }

    ui.treeWidget->setModel( m_sidebarModel );
    ui.treeWidget->setItemDelegate( m_delegate );

    #ifdef HIDE_RADIO
    ui.treeWidget->setRowHidden( 2, QModelIndex(), true );
    ui.treeWidget->setRowHidden( 3, QModelIndex(), true );
    ui.treeWidget->setRowHidden( 4, QModelIndex(), true );
    ui.treeWidget->setRowHidden( 5, QModelIndex(), true );
    ui.treeWidget->setRowHidden( 6, QModelIndex(), true );

    ui.treeWidget->setRowHidden( 9, QModelIndex(), true );
    ui.treeWidget->setRowHidden( 13, QModelIndex(), true );
    #endif

    connect( ui.treeWidget, SIGNAL( currentChanged( QModelIndex ) ),
             this,            SLOT( currentChanged( QModelIndex ) ) );
    connect( ui.treeWidget, SIGNAL( doubleClicked( QModelIndex ) ),
             this,            SLOT( doubleClicked( QModelIndex ) ) );
    connect( ui.treeWidget, SIGNAL( stationRequested( QString ) ),
             m_iInput,        SLOT( load( QString ) ) );
    connect( ui.treeWidget, SIGNAL( statusMessage( QString ) ),
             this,          SIGNAL( statusMessage( QString ) ) );
}


QWidget*
SidebarExtension::gui()
{
    if ( !m_widget )
        initGui();

    return m_widget;
}


void
SidebarExtension::updateAll()
{
    if ( m_iSettings->currentUser() )
        connect( m_iSettings->currentUser(), SIGNAL( historyChanged() ),
                 this,                         SLOT( updateHistory() ) );

    updateStations();
    m_iWebService->friends();
    m_iWebService->neighbours();
    m_iWebService->userTags();
    m_iWebService->recentTracks();
    m_iWebService->recentLovedTracks();

#ifndef HIDE_RADIO
    m_iWebService->recentBannedTracks();
#endif // HIDE_RADIO

    ui.treeWidget->setUser( settingsService()->currentUser() );

    updateHistory();
}


void
SidebarExtension::addListenedTrack( const QString& artist, const QString& track )
{
}


void
SidebarExtension::removeListenedTrack( const QString& artist, const QString& track )
{
    QModelIndex idx = m_sidebarModel->index( 7, 0, QModelIndex() );

    for ( int i = m_sidebarModel->rowCount( idx ) - 1; i >= 0; i-- )
    {
        if ( idx.child( i, 0 ).data( kArtistRole ).toString() == artist &&
             idx.child( i, 0 ).data( kTrackRole ).toString() == track )
        {
            m_sidebarModel->removeRow( i, idx );
        }
    }
}


void
SidebarExtension::addLovedTrack( const QString& artist, const QString& track )
{
    QModelIndex idx = m_sidebarModel->index( 8, 0, QModelIndex() );
    m_sidebarModel->insertRows( 0, 1, idx );

    m_sidebarModel->setData( m_sidebarModel->index( 0, 0, idx ), QVariant( QString( "%1 - %2" ).arg( artist ).arg( track ) ), Qt::DisplayRole );
    m_sidebarModel->setData( m_sidebarModel->index( 0, 0, idx ), QVariant( QString( "smallitem" ) ), kTypeRole );
    m_sidebarModel->setData( m_sidebarModel->index( 0, 0, idx ), QVariant( artist ), kArtistRole );
    m_sidebarModel->setData( m_sidebarModel->index( 0, 0, idx ), QVariant( track ), kTrackRole );
    m_sidebarModel->setData( m_sidebarModel->index( 0, 0, idx ), QVariant( SidebarExtension::Track ), kItemTypeRole );
    m_sidebarModel->setData( m_sidebarModel->index( 0, 0, idx ), QVariant( QIcon( dataPath( "icons/icon_track.png" ) ) ), Qt::DecorationRole );
}


void
SidebarExtension::removeLovedTrack( const QString& artist, const QString& track )
{
    QModelIndex idx = m_sidebarModel->index( 8, 0, QModelIndex() );

    for ( int i = m_sidebarModel->rowCount( idx ) - 1; i >= 0; i-- )
    {
        if ( idx.child( i, 0 ).data( kArtistRole ).toString() == artist &&
             idx.child( i, 0 ).data( kTrackRole ).toString() == track )
        {
            m_sidebarModel->removeRow( i, idx );
        }
    }
}


void
SidebarExtension::addBannedTrack( const QString& artist, const QString& track )
{
#ifndef HIDE_RADIO

    QModelIndex idx = m_sidebarModel->index( 9, 0, QModelIndex() );
    m_sidebarModel->insertRows( 0, 1, idx );

    m_sidebarModel->setData( m_sidebarModel->index( 0, 0, idx ), QVariant( QString( "%1 - %2" ).arg( artist ).arg( track ) ), Qt::DisplayRole );
    m_sidebarModel->setData( m_sidebarModel->index( 0, 0, idx ), QVariant( QString( "smallitem" ) ), kTypeRole );
    m_sidebarModel->setData( m_sidebarModel->index( 0, 0, idx ), QVariant( artist ), kArtistRole );
    m_sidebarModel->setData( m_sidebarModel->index( 0, 0, idx ), QVariant( track ), kTrackRole );
    m_sidebarModel->setData( m_sidebarModel->index( 0, 0, idx ), QVariant( SidebarExtension::Track ), kItemTypeRole );
    m_sidebarModel->setData( m_sidebarModel->index( 0, 0, idx ), QVariant( QIcon( dataPath( "icons/icon_track.png" ) ) ), Qt::DecorationRole );

#endif // HIDE_RADIO
}


void
SidebarExtension::removeBannedTrack( const QString& artist, const QString& track )
{
#ifndef HIDE_RADIO

    QModelIndex idx = m_sidebarModel->index( 9, 0, QModelIndex() );

    for ( int i = m_sidebarModel->rowCount( idx ) - 1; i >= 0; i-- )
    {
        if ( idx.child( i, 0 ).data( kArtistRole ).toString() == artist &&
             idx.child( i, 0 ).data( kTrackRole ).toString() == track )
        {
            m_sidebarModel->removeRow( i, idx );
        }
    }

#endif // HIDE_RADIO
}


void
SidebarExtension::removeFriend( const QString& username )
{
    QModelIndex idx = m_sidebarModel->index( 11, 0, QModelIndex() );

    for ( int i = m_sidebarModel->rowCount( idx ) - 1; i >= 0; i-- )
    {
        if ( idx.child( i, 0 ).data().toString() == username )
            m_sidebarModel->removeRow( i, idx );
    }
}


void
SidebarExtension::currentChanged( const QModelIndex& index )
{
    #ifndef HIDE_RADIO
    QString station = index.data( kStationRole ).toString();

    if ( !station.isEmpty() )
        emit stationFocused( station );
    #endif // HIDE_RADIO
}


void
SidebarExtension::doubleClicked( const QModelIndex& index )
{
    QString type = index.data( kTypeRole ).toString();

    if ( type == "userinfo" )
    {
        // Goto profile via evil code duplication from Container that's
        // really inexcusable. I want a REFACTORING MONTH!
        QByteArray user = QUrl::toPercentEncoding( settingsService()->currentUsername() );
        #ifdef HIDE_RADIO
            QString dashUrl = "http://www.lastfm.jp/user/" + user;
        #else
            QString dashUrl = "http://www.last.fm/user/" + user;
        #endif

        new BrowserThread( dashUrl );
    }
    else
    {
        // REFACTOR. The way this has been designed, there's basically no way of
        // finding out whether an index is a station or not, hence this
        // blanket approach.

        #ifndef HIDE_RADIO
        QString station = index.data( kStationRole ).toString();

        if ( !station.isEmpty() )
        {
            qDebug() << "Starting radio:" << station;
            m_iInput->load( station );
        }
        #endif // HIDE_RADIO
    }
}


void
SidebarExtension::stationChanged( QString url, QString name )
{
#ifndef HIDE_RADIO
    m_iSettings->currentUser()->setResumeStation( url );
    m_iSettings->currentUser()->addRecentStation( url, name );

//     updateHistory();
#endif // HIDE_RADIO
}


void
SidebarExtension::stationName( QString url, QString name )
{
#ifndef HIDE_RADIO
    m_iSettings->currentUser()->setStationName( url, name );

    updateHistory();
#endif // HIDE_RADIO
}


void
SidebarExtension::updateStations()
{
    QByteArray user = QUrl::toPercentEncoding( settingsService()->currentUsername() );

    QString station = "lastfm://user/" + user + "/personal";
    m_sidebarModel->setData( m_sidebarModel->index( 3, 0, QModelIndex() ), QVariant( station ), kStationRole );

    station = "lastfm://user/" + user + "/neighbours";
    m_sidebarModel->setData( m_sidebarModel->index( 5, 0, QModelIndex() ), QVariant( station ), kStationRole );

    station = "lastfm://user/" + user + "/loved";
    m_sidebarModel->setData( m_sidebarModel->index( 4, 0, QModelIndex() ), QVariant( station ), kStationRole );

    station = "lastfm://user/" + user + "/recommended";
    m_sidebarModel->setData( m_sidebarModel->index( 2, 0, QModelIndex() ), QVariant( station ), kStationRole );
}


void
SidebarExtension::updateFriends( const QString& username, const QStringList& friends )
{
    if ( username.toLower() != m_iWebService->currentUsername().toLower() )
        return;

    // This cumbersome bit of code here is how the Qt docs suggests you sort
    // a string list case-insensitively
    QMap<QString, QString> alphaSorted;
    foreach( QString s, friends )
    {
        alphaSorted.insert( s.toLower(), s );
    }
    QStringList l;
    QMapIterator<QString, QString> iter( alphaSorted );
    while ( iter.hasNext() )
    {
        iter.next();
        l << iter.value();
    }

    QModelIndex idx = m_sidebarModel->index( 11, 0, QModelIndex() );
    m_sidebarModel->removeRows( 0, m_sidebarModel->rowCount( idx ), idx );
    m_sidebarModel->removeColumns( 0, m_sidebarModel->columnCount( idx ), idx );
    m_sidebarModel->insertRows( 0, l.count(), idx );
    m_sidebarModel->insertColumns( 0, 1, idx );

    for ( int i = 0; i < l.count(); i++ )
    {
        m_sidebarModel->setData( m_sidebarModel->index( i, 0, idx ), QVariant( l.at( i ) ), Qt::DisplayRole );
        m_sidebarModel->setData( m_sidebarModel->index( i, 0, idx ), QVariant( QString( "smallitem" ) ), kTypeRole );
        m_sidebarModel->setData( m_sidebarModel->index( i, 0, idx ),
            QVariant( QString( "lastfm://user/%1/personal" ).arg( l.at( i ) ) ),
            kStationRole );
        m_sidebarModel->setData( m_sidebarModel->index( i, 0, idx ), QVariant( SidebarExtension::User ), kItemTypeRole );
        m_sidebarModel->setData( m_sidebarModel->index( i, 0, idx ), QVariant( QIcon( dataPath( "icons/user_blue.png" ) ) ), Qt::DecorationRole );
    }
}


void
SidebarExtension::updateNeighbours( const QString& username, const QList<QPair<QString, int> >& neighbours )
{
    if ( username.toLower() != m_iWebService->currentUsername().toLower() )
        return;

    QModelIndex idx = m_sidebarModel->index( 12, 0, QModelIndex() );
    m_sidebarModel->removeRows( 0, m_sidebarModel->rowCount( idx ), idx );
    m_sidebarModel->removeColumns( 0, m_sidebarModel->columnCount( idx ), idx );
    m_sidebarModel->insertRows( 0, neighbours.count(), idx );
    m_sidebarModel->insertColumns( 0, 1, idx );

    for ( int i = 0; i < neighbours.count(); i++ )
    {
        m_sidebarModel->setData( m_sidebarModel->index( i, 0, idx ),
            QVariant( neighbours.at( i ).first ), Qt::DisplayRole );
        m_sidebarModel->setData( m_sidebarModel->index( i, 0, idx ),
            QVariant( neighbours.at( i ).second ), kWeightRole );
        m_sidebarModel->setData( m_sidebarModel->index( i, 0, idx ),
            QVariant( QString( "smallitem" ) ), kTypeRole );
        m_sidebarModel->setData( m_sidebarModel->index( i, 0, idx ),
            QVariant( QString( "lastfm://user/%1/personal" ).arg( neighbours.at( i ).first ) ),
            kStationRole );
        m_sidebarModel->setData( m_sidebarModel->index( i, 0, idx ),
            QVariant( SidebarExtension::User ), kItemTypeRole );
        m_sidebarModel->setData( m_sidebarModel->index( i, 0, idx ),
            QVariant( QIcon( dataPath( "icons/user_purple.png" ) ) ), Qt::DecorationRole );
    }
    
}


void
SidebarExtension::updateUserTags( const QString& username, const QList<QPair<QString, int> >& tags )
{
    if ( username.toLower() != m_iWebService->currentUsername().toLower() )
        return;

    QModelIndex idx = m_sidebarModel->index( 10, 0, QModelIndex() );
    m_sidebarModel->removeRows( 0, m_sidebarModel->rowCount( idx ), idx );
    m_sidebarModel->removeColumns( 0, m_sidebarModel->columnCount( idx ), idx );
    m_sidebarModel->insertRows( 0, tags.count(), idx );
    m_sidebarModel->insertColumns( 0, 1, idx );

    for ( int i = 0; i < tags.count(); i++ )
    {
        m_sidebarModel->setData( m_sidebarModel->index( i, 0, idx ), QVariant( tags.at( i ).first ), Qt::DisplayRole );
        m_sidebarModel->setData( m_sidebarModel->index( i, 0, idx ), QVariant( tags.at( i ).second ), kWeightRole );
        m_sidebarModel->setData( m_sidebarModel->index( i, 0, idx ), QVariant( QString( "smallitem" ) ), kTypeRole );
        m_sidebarModel->setData( m_sidebarModel->index( i, 0, idx ),
            QVariant( QString( "lastfm://usertags/%1/%2" ).arg( m_iWebService->currentUsername() ).arg( tags.at( i ).first ) ), kStationRole );
        m_sidebarModel->setData( m_sidebarModel->index( i, 0, idx ), QVariant( SidebarExtension::Tag ), kItemTypeRole );
        m_sidebarModel->setData( m_sidebarModel->index( i, 0, idx ), QVariant( QIcon( dataPath( "icons/icon_tag.png" ) ) ), Qt::DecorationRole );
    }

    // Interrupted attempt at sorting on weight
    //m_sidebarModel->setSortRole( kWeightRole );
    //QStandardItem* item = m_sidebarModel->itemFromIndex( idx );
    //item->sortChildren( 0, Qt::AscendingOrder ); // column = 0
}


void
SidebarExtension::updateRecentTracks( const QString& username, QList< QPair<QString, QString> > songs )
{
    if ( username.toLower() != m_iWebService->currentUsername().toLower() )
        return;

    QModelIndex idx = m_sidebarModel->index( 7, 0, QModelIndex() );
    m_sidebarModel->removeRows( 0, m_sidebarModel->rowCount( idx ), idx );
    m_sidebarModel->removeColumns( 0, m_sidebarModel->columnCount( idx ), idx );
    m_sidebarModel->insertRows( 0, songs.count(), idx );
    m_sidebarModel->insertColumns( 0, 1, idx );

    for ( int i = 0; i < songs.count(); i++ )
    {
        m_sidebarModel->setData( m_sidebarModel->index( i, 0, idx ), QVariant( QString( "%1 - %2" ).arg( songs.at( i ).first ).arg( songs.at( i ).second ) ), Qt::DisplayRole );
        m_sidebarModel->setData( m_sidebarModel->index( i, 0, idx ), QVariant( QString( "smallitem" ) ), kTypeRole );
        m_sidebarModel->setData( m_sidebarModel->index( i, 0, idx ), QVariant( songs.at( i ).first ), kArtistRole );
        m_sidebarModel->setData( m_sidebarModel->index( i, 0, idx ), QVariant( songs.at( i ).second ), kTrackRole );
        m_sidebarModel->setData( m_sidebarModel->index( i, 0, idx ), QVariant( SidebarExtension::Track ), kItemTypeRole );
        m_sidebarModel->setData( m_sidebarModel->index( i, 0, idx ), QVariant( QIcon( dataPath( "icons/icon_track.png" ) ) ), Qt::DecorationRole );
    }
}


void
SidebarExtension::updateRecentLovedTracks( const QString& username, QList< QPair<QString, QString> > songs )
{
    if ( username.toLower() != m_iWebService->currentUsername().toLower() )
        return;

    QModelIndex idx = m_sidebarModel->index( 8, 0, QModelIndex() );
    m_sidebarModel->removeRows( 0, m_sidebarModel->rowCount( idx ), idx );
    m_sidebarModel->removeColumns( 0, m_sidebarModel->columnCount( idx ), idx );
    m_sidebarModel->insertRows( 0, songs.count(), idx );
    m_sidebarModel->insertColumns( 0, 1, idx );

    for ( int i = 0; i < songs.count(); i++ )
    {
        m_sidebarModel->setData( m_sidebarModel->index( i, 0, idx ), QVariant( QString( "%1 - %2" ).arg( songs.at( i ).first ).arg( songs.at( i ).second ) ), Qt::DisplayRole );
        m_sidebarModel->setData( m_sidebarModel->index( i, 0, idx ), QVariant( QString( "smallitem" ) ), kTypeRole );
        m_sidebarModel->setData( m_sidebarModel->index( i, 0, idx ), QVariant( songs.at( i ).first ), kArtistRole );
        m_sidebarModel->setData( m_sidebarModel->index( i, 0, idx ), QVariant( songs.at( i ).second ), kTrackRole );
        m_sidebarModel->setData( m_sidebarModel->index( i, 0, idx ), QVariant( SidebarExtension::Track ), kItemTypeRole );
        m_sidebarModel->setData( m_sidebarModel->index( i, 0, idx ), QVariant( QIcon( dataPath( "icons/icon_track.png" ) ) ), Qt::DecorationRole );
    }
}


void
SidebarExtension::updateRecentBannedTracks( const QString& username, QList< QPair<QString, QString> > songs )
{
#ifndef HIDE_RADIO

    if ( username.toLower() != m_iWebService->currentUsername().toLower() )
        return;

    QModelIndex idx = m_sidebarModel->index( 9, 0, QModelIndex() );
    m_sidebarModel->removeRows( 0, m_sidebarModel->rowCount( idx ), idx );
    m_sidebarModel->removeColumns( 0, m_sidebarModel->columnCount( idx ), idx );
    m_sidebarModel->insertRows( 0, songs.count(), idx );
    m_sidebarModel->insertColumns( 0, 1, idx );

    for ( int i = 0; i < songs.count(); i++ )
    {
        m_sidebarModel->setData( m_sidebarModel->index( i, 0, idx ), QVariant( QString( "%1 - %2" ).arg( songs.at( i ).first ).arg( songs.at( i ).second ) ), Qt::DisplayRole );
        m_sidebarModel->setData( m_sidebarModel->index( i, 0, idx ), QVariant( QString( "smallitem" ) ), kTypeRole );
        m_sidebarModel->setData( m_sidebarModel->index( i, 0, idx ), QVariant( songs.at( i ).first ), kArtistRole );
        m_sidebarModel->setData( m_sidebarModel->index( i, 0, idx ), QVariant( songs.at( i ).second ), kTrackRole );
        m_sidebarModel->setData( m_sidebarModel->index( i, 0, idx ), QVariant( SidebarExtension::Track ), kItemTypeRole );
        m_sidebarModel->setData( m_sidebarModel->index( i, 0, idx ), QVariant( QIcon( dataPath( "icons/icon_track.png" ) ) ), Qt::DecorationRole );
    }

#endif // HIDE_RADIO
}


void
SidebarExtension::updateHistory()
{
#ifndef HIDE_RADIO

    QStringList l = m_iSettings->currentUser()->recentStations();
    QModelIndex idx = m_sidebarModel->index( 13, 0, QModelIndex() );
    m_sidebarModel->removeRows( 0, m_sidebarModel->rowCount( idx ), idx );
    m_sidebarModel->removeColumns( 0, m_sidebarModel->columnCount( idx ), idx );
    m_sidebarModel->insertRows( 0, l.count(), idx );
    m_sidebarModel->insertColumns( 0, 1, idx );

    for ( int i = 0; i < l.count(); i++ )
    {
        m_sidebarModel->setData( m_sidebarModel->index( i, 0, idx ), QVariant( m_iSettings->currentUser()->stationName( l.at( i ) ) ), Qt::DisplayRole );
        m_sidebarModel->setData( m_sidebarModel->index( i, 0, idx ), QVariant( QString( "smallitem" ) ), kTypeRole );
        m_sidebarModel->setData( m_sidebarModel->index( i, 0, idx ), QVariant( l.at( i ) ), kStationRole );
        m_sidebarModel->setData( m_sidebarModel->index( i, 0, idx ), QVariant( SidebarExtension::Station ), kItemTypeRole );
        m_sidebarModel->setData( m_sidebarModel->index( i, 0, idx ), QVariant( QIcon( dataPath( "icons/icon_radio.png" ) ) ), Qt::DecorationRole );
    }

#endif // HIDE_RADIO
}


Q_EXPORT_PLUGIN2( extension_history, SidebarExtension )
