/***************************************************************************
 *   Copyright (C) 2005 - 2006 by                                          *
 *      Christian Muehlhaeuser, Last.fm Ltd <chris@last.fm>                *
 *      Erik Jaelevik, Last.fm Ltd <erik@last.fm>                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Steet, Fifth Floor, Boston, MA  02111-1307, USA.          *
 ***************************************************************************/

#include "systray.h"
#include "Loqqer.h"
#include "iconshack.h"
#include "container.h"
#include "containerutils.h"


/******************************************************************************
    SysTray
******************************************************************************/
SysTray::SysTray( Container* parent ) :
    m_trayIcon( parent ),
    m_parent( parent ),
    m_tipFormat( "%1 - %2 (%3)" ),
    m_badStatus( false )
{
    // Build popup menu
    QAction* openAct = m_trayMenu.addAction( tr( "Open" ), parent, SLOT( restoreWindow() ) );
    m_trayMenu.setDefaultAction( openAct );

    m_trayMenu.addSeparator();
    m_trayMenu.addAction( parent->ui.actionDashboard );
    m_trayMenu.addAction( parent->ui.actionToggleScrobbling );
    m_userMenu = m_trayMenu.addMenu( tr( "Change user" ) );

    m_trayMenu.addSeparator();
    m_trayMenu.addAction( parent->ui.actionLove );
    m_trayMenu.addAction( parent->ui.actionBan );

#ifndef HIDE_RADIO
    m_trayMenu.addSeparator();
#endif // HIDE_RADIO

    m_trayMenu.addAction( parent->ui.actionSkip );
    m_trayMenu.addAction( parent->ui.actionPlay );
    m_trayMenu.addAction( parent->ui.actionStop );

    m_trayMenu.addSeparator();
    m_trayMenu.addAction( tr( "Exit" ), parent, SLOT( actualClose() ) );

    #if QT_VERSION >= 0x040200
    m_trayIcon.setContextMenu( &m_trayMenu );
    #else
    m_trayIcon.setPopup( &m_trayMenu );
    #endif

    #ifdef Q_WS_MAC
        QPixmap defaultIcon( dataPath( "icons/systray_mac.png" ) );
    #else
        QPixmap defaultIcon = IconShack::instance().GetGoodUserIcon( eRed );
    #endif
    m_trayIcon.setIcon( defaultIcon );

    #if QT_VERSION >= 0x040200
    connect( &m_trayIcon, SIGNAL( activated( QSystemTrayIcon::ActivationReason ) ),
             this,        SLOT  ( trayAction( QSystemTrayIcon::ActivationReason ) ) );
    #else
    connect( &m_trayIcon, SIGNAL( doubleClicked( QPoint ) ),
             m_parent,    SLOT  ( restoreWindow() ) );
    #endif

    #ifdef Q_WS_MAC
    // On the mac there is no aboutToShow() signal for the tray-menu.
    // Load the content of the user-menu on startup, till we find a proper fix.
    aboutToShowMenu();
    #else
    connect( &m_trayMenu, SIGNAL( aboutToShow() ),
             this,        SLOT  ( aboutToShowMenu() ) );
    #endif

    connect( m_userMenu,  SIGNAL( triggered( QAction* ) ),
             m_parent,    SLOT( userSelected( QAction* ) ) );
}


/******************************************************************************
    setUser
******************************************************************************/
void
SysTray::setUser( UserSettingsServiceInterface* currentUser )
{
    UserIconColour iconColour = currentUser->icon();
    QPixmap icon;
    if ( currentUser->logToProfile() )
    {
        #ifdef Q_WS_MAC
            icon = QPixmap( dataPath( "icons/systray_mac.png" ) );
        #else
            icon = IconShack::instance().GetGoodUserIcon( iconColour );
        #endif
    }
    else
    {
        #ifdef Q_WS_MAC
            QIcon temp( dataPath( "icons/systray_mac.png" ) );
            icon = temp.pixmap( 19, 12, QIcon::Disabled );
        #else
            icon = IconShack::instance().GetDisabledUserIcon( iconColour );
        #endif
    }

    m_trayIcon.setIcon( icon );

    m_user = currentUser->username();
    refreshToolTip();
}

/******************************************************************************
    setTrack
******************************************************************************/
void
SysTray::setTrack( MetaData track )
{
    m_artist = track.artist();
    m_track = track.track();
    refreshToolTip();
}

/******************************************************************************
    setStatus
******************************************************************************/
void
SysTray::setStatus( bool statusOK )
{
    m_badStatus = !statusOK;
}


/******************************************************************************
    aboutToShowMenu
******************************************************************************/
void
SysTray::aboutToShowMenu()
{
    m_userMenu->clear();

    // For each user, add them and put a check next to the current
    QString currentUser = m_parent->m_iSettings->currentUsername();
    QStringList users = m_parent->m_iSettings->allUsers();

    for ( int i = 0; i < users.size(); ++i )
    {
        QString username = users.at( i );
        QAction* action = new QAction( m_userMenu );
        action->setData( "user" );
        action->setText( username );
        action->setCheckable( true );

        if ( username == currentUser )
        {
            action->setChecked( true );
        }

        m_userMenu->addAction( action );
    }
}

/******************************************************************************
    refreshToolTip
******************************************************************************/
void
SysTray::refreshToolTip()
{
    QString tip;
    if ( m_artist != "" && m_track != "" )
    {
        tip = m_tipFormat.arg( m_artist, m_track, m_user );
    }
    else
    {
        tip = tr( "%1 is logged in" ).arg( m_user );
    }

    #ifdef WIN32
        // Good old crappy Win32
        tip.replace( "&", "&&&" );
    #endif

    m_trayIcon.setToolTip( tip );
}


#if QT_VERSION >= 0x040200
void
SysTray::trayAction( QSystemTrayIcon::ActivationReason reason )
{
    if ( reason == QSystemTrayIcon::DoubleClick )
        m_parent->restoreWindow();
}
#endif
