/***************************************************************************
 *   Copyright (C) 2005 - 2006 by                                          *
 *      Christian Muehlhaeuser, Last.fm Ltd <chris@last.fm>                *
 *      Erik Jaelevik, Last.fm Ltd <erik@last.fm>                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Steet, Fifth Floor, Boston, MA  02111-1307, USA.          *
 ***************************************************************************/

#ifdef WIN32
#include <windows.h>
#endif

#include <QtGui>
#include <QTimer>

#include <containerutils.h>
#include "tagdialog.h"
#include "webserviceinterface.h"

#include "Loqqer.h"

TagDialog::TagDialog( int defaultTagType, QWidget *parent )
    : QDialog( parent, Qt::Dialog )
{
    m_webService = webService();
    Q_ASSERT( m_webService != NULL );

    ui.setupUi( this );

    QRegExp rx( "[a-zA-Z0-9\\-:, ]{1,255}" );
    QValidator *validator = new QRegExpValidator( rx, this );
    ui.tagEdit->setValidator( validator );

    if ( defaultTagType == -1 )
    {
        ui.tagTypeBox->setCurrentIndex( settingsService()->currentUser()->lastTagType( 1 ) );
    }
    else
    {
        ui.tagTypeBox->setCurrentIndex( defaultTagType );
    }

    VERIFY( connect( m_webService, SIGNAL( artistTagsResult( QString, QStringList ) ),
                     this,         SLOT( tagsForArtistResult( QString, QStringList ) ) ) );
    VERIFY( connect( m_webService, SIGNAL( userTagsResult( QString, const QList<QPair<QString, int> >& ) ),
                     this,         SLOT( tagsForUserResult( QString, const QList<QPair<QString, int> >& ) ) ) );
    VERIFY( connect( m_webService, SIGNAL( userArtistTagsResult( QString, QStringList ) ),
                     this,         SLOT( tagsForUserArtistResult( QString, QStringList ) ) ) );
    VERIFY( connect( m_webService, SIGNAL( userTrackTagsResult( QString, QStringList ) ),
                     this,         SLOT( tagsForUserTrackResult( QString, QStringList ) ) ) );
    VERIFY( connect( m_webService, SIGNAL( userAlbumTagsResult( QString, QStringList ) ),
                     this,         SLOT( tagsForUserAlbumResult( QString, QStringList ) ) ) );

    VERIFY( connect( ui.tagTypeBox, SIGNAL( activated( int ) ),
                     this,          SLOT( tagTypeChanged( int ) ) ) );
    VERIFY( connect( ui.okButton,   SIGNAL( clicked() ),
                     this,          SLOT( okPressed() ) ) );
    VERIFY( connect( ui.personalTagsList, SIGNAL( itemDoubleClicked( QListWidgetItem* ) ),
                     this,                SLOT( addTag( QListWidgetItem* ) ) ) );
    VERIFY( connect( ui.publicTagsList, SIGNAL( itemDoubleClicked( QListWidgetItem* ) ),
                     this,              SLOT( addTag( QListWidgetItem* ) ) ) );

    m_webService->userTags();
}


void
TagDialog::setSong( MetaData metaData )
{
    m_metaData = metaData;
    m_webService->artistTags( m_metaData.artist() );
    tagTypeChanged( ui.tagTypeBox->currentIndex() );

    ui.tagTypeBox->setItemText( 0, QString( tr( "Artist: %1" ) ).arg( metaData.artist() ) );
    ui.tagTypeBox->setItemText( 1, QString( tr( "Track: %1 - %2" ) ).arg( metaData.artist() ).arg( metaData.track() ) );
    ui.tagTypeBox->setItemText( 2, QString( tr( "Album: %1 - %2" ) ).arg( metaData.artist() ).arg( metaData.album() ) );
}


void
TagDialog::okPressed()
{
    switch( ui.tagTypeBox->currentIndex() )
    {
        case 0: // artist
            m_webService->tagArtist( ui.tagEdit->text(), m_metaData.artist() );
            break;

        case 1: // track
            m_webService->tagTrack( ui.tagEdit->text(), m_metaData.artist(), m_metaData.track() );
            break;

        case 2: // album
            m_webService->tagAlbum( ui.tagEdit->text(), m_metaData.artist(), m_metaData.album() );
            break;
    }
}


void
TagDialog::tagTypeChanged( int type )
{
    ui.tagTypeBox->setEnabled( false );
    ui.tagEdit->setEnabled( false );
    ui.tagEdit->setText( tr( "Loading..." ) );

    switch ( type )
    {
        case 0:
            m_webService->userArtistTags( m_metaData.artist() );
            break;

        case 1:
            m_webService->userTrackTags( m_metaData.artist(), m_metaData.track() );
            break;

        case 2:
            m_webService->userAlbumTags( m_metaData.artist(), m_metaData.album() );
            break;

    }

    settingsService()->currentUser()->setLastTagType( ui.tagTypeBox->currentIndex() );
}


void
TagDialog::tagsForArtistResult( const QString& artist, const QStringList &result )
{
    ui.publicTagsList->clear();

    for ( int i = 0; i < result.count(); i++ )
    {
        ui.publicTagsList->addItem( result[i] );
    }
}


void
TagDialog::tagsForUserResult( const QString& user, const QList<QPair<QString, int> >& result )
{
    ui.personalTagsList->clear();

    for ( int i = 0; i < result.count(); i++ )
    {
        ui.personalTagsList->addItem( result[i].first );
    }
}


void
TagDialog::tagsForUserArtistResult( const QString& user, const QStringList &result )
{
    QString tags = "";

    for ( int i = 0; i < result.count(); i++ )
    {
        tags += ( i > 0 ? ", " : "" ) + result[i];
    }

    ui.tagTypeBox->setEnabled( true );
    ui.tagEdit->setEnabled( true );
    ui.tagEdit->setText( tags );
    ui.tagEdit->setFocus();
}


void
TagDialog::tagsForUserTrackResult( const QString& user, const QStringList &result )
{
    QString tags = "";

    for ( int i = 0; i < result.count(); i++ )
    {
        tags += ( i > 0 ? ", " : "" ) + result[i];
    }

    ui.tagTypeBox->setEnabled( true );
    ui.tagEdit->setEnabled( true );
    ui.tagEdit->setText( tags );
    ui.tagEdit->setFocus();
}


void
TagDialog::tagsForUserAlbumResult( const QString& user, const QStringList &result )
{
    QString tags = "";

    for ( int i = 0; i < result.count(); i++ )
    {
        tags += ( i > 0 ? ", " : "" ) + result[i];
    }

    ui.tagTypeBox->setEnabled( true );
    ui.tagEdit->setEnabled( true );
    ui.tagEdit->setText( tags );
    ui.tagEdit->setFocus();
}


void
TagDialog::addTag( QListWidgetItem *item )
{
    if ( !ui.tagEdit->text().trimmed().isEmpty() )
        ui.tagEdit->setText( ui.tagEdit->text() + ", " + item->text() );
    else
        ui.tagEdit->setText( item->text() );
}

