/***************************************************************************
 *   Copyright (C) 2005 - 2006 by                                          *
 *      Christian Muehlhaeuser, Last.fm Ltd <chris@last.fm>                *
 *      Erik Jaelevik, Last.fm Ltd <erik@last.fm>                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Steet, Fifth Floor, Boston, MA  02111-1307, USA.          *
 ***************************************************************************/

#ifndef WEBSERVICE_H
#define WEBSERVICE_H

#include <QObject>
#include <QHash>

#include <containerutils.h>
#include <http.h>
#include <metadata.h>
#include "webserviceinterface.h"

class QHttpResponseHeader;
class QUrl;
class QTimer;

class WebService : public WebServiceInterface
{
    Q_OBJECT
    Q_INTERFACES( WebServiceInterface )

    public:

        enum RadioErrorCode
        {
            RADIO_ERROR_NOT_ENOUGH_CONTENT = 1, // There is not enough content to play this station.
            RADIO_ERROR_FEWGROUPMEMBERS,        // This group does not have enough members for radio.
            RADIO_ERROR_FEWFANS,                // This artist does not have enough fans for radio.
            RADIO_ERROR_UNAVAILABLE,            // This item is not available for streaming.
            RADIO_ERROR_SUBSCRIBE,              // This feature is only available to subscribers.
            RADIO_ERROR_FEWNEIGHBOURS,          // There are not enough neighbours for this radio.
            RADIO_ERROR_OFFLINE                 // The streaming system is offline for maintenance, please try again later
        };

        WebService();

        void handshake( const QString& username, const QString& password, const QString& version );
        void changeStation( QString url );

        QString currentUsername() { return m_username; }
        QString currentPassword() { return m_password; }
        QString currentStation() { return m_station; }
        QString session() { return m_session; }
        QUrl streamUrl() { return m_streamUrl; }

        void userTags( QString username = QString() );
        void userArtistTags( QString artist, QString username = QString() );
        void userAlbumTags( QString artist, QString album, QString username = QString() );
        void userTrackTags( QString artist, QString track, QString username = QString() );
        void userPics( QStringList names );

        void artistTags( QString artist );
        void topTags();

        void similarArtists( QString artist );
        void searchTag( QString tag );
        void similarTags( QString tag );

        void recommend( int type, QString username, QString message, QString artist, QString token = QString() );
        void recommendArtist( QString username, QString message, QString artist ) { recommend( ItemArtist, username, message, artist ); }
        void recommendAlbum( QString username, QString message, QString artist, QString album ) { recommend( ItemAlbum, username, message, artist, album ); }
        void recommendTrack( QString username, QString message, QString artist, QString track ) { recommend( ItemTrack, username, message, artist, track ); }

        void deleteFriend( QString user );

        void tagItem( int type, int mode, QString tag, QString artist, QString token = QString() );
        void tagArtist( QString tag, QString artist, int mode = TAG_OVERWRITE ) { tagItem( ItemArtist, mode, tag, artist ); }
        void tagAlbum( QString tag, QString artist, QString album, int mode = TAG_OVERWRITE ) { tagItem( ItemAlbum, mode, tag, artist, album ); }
        void tagTrack( QString tag, QString artist, QString track, int mode = TAG_OVERWRITE ) { tagItem( ItemTrack, mode, tag, artist, track ); }

        /*********************************************************************/ /**
            Verify with server that a supplied user/pass combo is valid. Password
            should be MD5 hashed.
        **************************************************************************/
        void verifyUser( const QString& user, const QString& pass, const QString& passLower );

        QString baseHost() { return m_baseHost; }

    public slots:
        void requestMetaDataRadio();

        // These two functions will not buffer up requests. If you call it
        // again before the previous request has come back, the previous
        // request will be cancelled.
        void requestMetaDataArtist( const MetaData& artist );
        void requestMetaDataTrack( const MetaData& track );

        // H.H.H.H.A.A.A.A.C.C.C.C.K.K.K.K.K.K.K.K.K.K.K.K.K.K.K.K.K.K.K.K.K.K.K
        void cancelLastMetaDataArtistRequest();
        void cancelLastMetaDataTrackRequest();

        void enableScrobbling( bool enabled, bool forceWebReq = false );
        void enableDiscoveryMode( bool enabled );

        void friends( QString username );
        void neighbours( QString username );
        void recentTracks( QString username );
        void recentLovedTracks( QString username );
        void recentBannedTracks( QString username );

        void unlisten( QString artist, QString track );
        void love( QString artist, QString track );
        void unlove( QString artist, QString track );
        void ban( QString artist, QString track );
        void unban( QString artist, QString track );
        void skip();

    signals:
        void actionStarted();
        void actionFinished();

        void handshakeResult( int result );
        void streamingUrl( const QUrl& url );
        void stationChanged( QString url, QString name );
        void stationChangeFailed( QString error );
        void stationNameChanged( QString url, QString name );
        void songQueued();
        void enableScrobblingDone();
        void enableDiscoveryModeDone();

        void metaDataResultRadio( const MetaData &metaData );
        void metaDataResultArtist( const MetaData &metaData, bool error = false );
        void metaDataResultTrack( const MetaData &metaData, bool error = false );
        void metaDataResultStation( const QString &name, const QString &feed, bool newStation );
        void trackStatusChanged( QString msg );
        void radioStatusChanged( QString msg );

        void unlistenDone( const QString& artist, const QString& track, bool error = false );
        void loveDone( const QString& artist, const QString& track, bool error = false );
        void unloveDone( const QString& artist, const QString& track, bool error = false );
        void banDone( const QString& artist, const QString& track, bool error = false );
        void unbanDone( const QString& artist, const QString& track, bool error = false );
        void skipDone();
        void recommendDone( const QString& artist, const QString& album, const QString& track, bool error = false );
        void tagDone( const QString& artist, const QString& album, const QString& track, bool error = false );
        void deleteFriendDone( const QString& username, bool error = false );

        void friendsResult( const QString& username, const QStringList& friends );
        void friendsResult( const QString& username, const QList<UserMetaData>& friends );
        void neighboursResult( const QString& username, const QList<QPair<QString, int> >& neighbours );
        void recentTracksResult( const QString& username, QList< QPair<QString, QString> > songs );
        void recentLovedTracksResult( const QString& username, QList< QPair<QString, QString> > songs );
        void recentBannedTracksResult( const QString& username, QList< QPair<QString, QString> > songs );

        void userTagsResult( const QString& username, const QList<QPair<QString, int> >& tags );
        void userArtistTagsResult( const QString& username, const QStringList& tags );
        void userAlbumTagsResult( const QString& username, const QStringList& tags );
        void userTrackTagsResult( const QString& username, const QStringList& tags );
        void userPicsResult( const QMap<QString, QString>& picUrls, bool error = false );

        void artistTagsResult( const QString& artist, const QStringList& tags );
        void topTagsResult( const QList<QPair<QString, int> >& tags );

        void similarArtistsResult( const QList<QPair<QString, int> >& result, const QString& artist = QString() );
        void searchTagResult( const QList<QPair<QString, int> >& result );
        void similarTagsResult( const QList<QPair<QString, int> >& result );

        void verifyResult( UserAuthCode result, BootstrapCode bootstrap );

    private:
        QHash<int, Http *> m_httpStack;

        MetaData m_radioSong;
        int m_radioMetaDataTimeout;
        QTimer m_radioMetaDataRetryTimer;

        QString m_username;
        QString m_password;
        QString m_station;
        QString m_session;
        QString m_baseHost;
        QString m_basePath;

        QUrl m_streamUrl;
        QString m_stationUrl;
        QString m_stationName;

        bool m_connected;
        bool m_subscriber;
        bool m_scrobbling;

        QString m_actionArtist;
        QString m_actionAlbum;
        QString m_actionTrack;
        QString m_actionUser;

        // Bit hacky this as it breaks the common policy of the web service but
        // we keep instances of these objects around so we can cancel a request
        // if a new one comes in before the last one has finished.
        Http* m_artistMetadataHttp;
        Http* m_trackMetadataHttp;

        void clearActionCache() { m_actionArtist = ""; m_actionAlbum = ""; m_actionTrack = ""; }

        QString parameter( QString keyName, QString data );
        QStringList parameterArray( QString keyName, QString data );
        QStringList parameterKeys( QString keyName, QString data );

        void stackAppend( Http *http, int id );
        Http* stackGet( int id );
        void stackRemove( int id, bool keepAnimation = false );

        QByteArray formatXmlRpc( QString method, QList<QVariant> params );
        bool parseXmlRpc( QByteArray xmlResponse, QList<QVariant>& returnValues, QString& error );
        QVariant parseXmlRpcValue( QDomElement elem );
        void makeXmlRpcCall( QByteArray xmlData, const char* returnSlot, bool useCache = false, Http* http = NULL );

        QString xmlEncode( QString xml );
        QString xmlDecode( QString xml );

        // Returns the time string used as a challenge for authorising web calls
        QString challengeString();

        void logHttpError( QString msg, Http* http );

    private slots:
        void handshakeHeaderReceived( const QHttpResponseHeader &resp );
        void handshakeError( int error, const QString& errorString );
        void handshakeFinished( const QByteArray& data );

        void changeStationFinished( int id, bool error );
        void enableScrobblingFinished( int id, bool error );
        void enableDiscoveryModeFinished( int id, bool error );

        void metaDataRadioFinished( const QByteArray& buffer );
        void metaDataArtistFinished( const QByteArray& buffer );
        void metaDataTrackFinished( const QByteArray& buffer );

        void unlistenFinished( int id, bool error );
        void loveFinished( int id, bool error );
        void unloveFinished( int id, bool error );
        void banFinished( int id, bool error );
        void unbanFinished( int id, bool error );
        void skipFinished( int id, bool error );

        void friendsFinished( int id, bool error );
        void neighboursFinished( int id, bool error );
        void recentTracksFinished( int id, bool error );
        void recentLovedTracksFinished( int id, bool error );
        void recentBannedTracksFinished( int id, bool error );

        void userTagsFinished( int id, bool error );
        void userArtistTagsFinished( int id, bool error );
        void userAlbumTagsFinished( int id, bool error );
        void userTrackTagsFinished( int id, bool error );
        void userPicsFinished( QByteArray buffer );

        void artistTagsFinished( int id, bool error );
        void topTagsFinished( int id, bool error );

        void similarArtistsFinished( int id, bool error );
        void searchTagFinished( int id, bool error );
        void similarTagsFinished( QByteArray buffer );

        void recommendFinished( int id, bool error );
        void deleteFriendFinished( int id, bool error );
        void setTagFinished( int id, bool error );
        void verifyFinished( int id, bool error );
};

#endif
