/***************************************************************************
 *   Copyright (C) 2005 - 2006 by                                          *
 *      Christian Muehlhaeuser, Last.fm Ltd <chris@last.fm>                *
 *      Erik Jaelevik, Last.fm Ltd <erik@last.fm>                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Steet, Fifth Floor, Boston, MA  02111-1307, USA.          *
 ***************************************************************************/

#ifndef WEBSERVICE_INTERFACE_H
#define WEBSERVICE_INTERFACE_H

#include <QObject>

#include "http.h"
#include "metadata.h"
#include "usermetadata.h"

enum UserAuthCode
{
    AUTH_OK = 0,
    AUTH_OK_LOWER,
    AUTH_BADUSER,
    AUTH_BADPASS,
    AUTH_ERROR
};

enum BootstrapCode
{
    BOOTSTRAP_DENIED = 0,
    BOOTSTRAP_ALLOWED
};

enum TagMode
{
    TAG_OVERWRITE = 0,                  // Overwrites existing tag associations
    TAG_APPEND                          // Appends a tag to the list of existing associations
};

class WebServiceInterface : public QObject
{
    public:
        virtual ~WebServiceInterface() {}

        virtual void handshake( const QString& username, const QString& password, const QString& version ) = 0;
        virtual void changeStation( QString url ) = 0;

        virtual QString currentUsername() = 0;
        virtual QString currentPassword() = 0;
        virtual QString currentStation() = 0;
        virtual QString session() = 0;
        virtual QUrl streamUrl() = 0;

        virtual void userTags( QString username = QString() ) = 0;
        virtual void userArtistTags( QString artist, QString username = QString() ) = 0;
        virtual void userAlbumTags( QString artist, QString album, QString username = QString() ) = 0;
        virtual void userTrackTags( QString artist, QString track, QString username = QString() ) = 0;
        virtual void userPics( QStringList names ) = 0;

        virtual void artistTags( QString artist ) = 0;
        virtual void topTags() = 0;

        virtual void similarArtists( QString artist ) = 0;
        virtual void searchTag( QString tag ) = 0;
        virtual void similarTags( QString tag ) = 0;

        virtual void recommend( int type, QString username, QString message, QString artist, QString token = QString() ) = 0;
        virtual void recommendArtist( QString username, QString message, QString artist ) = 0;
        virtual void recommendAlbum( QString username, QString message, QString artist, QString album ) = 0;
        virtual void recommendTrack( QString username, QString message, QString artist, QString track ) = 0;

        virtual void deleteFriend( QString user ) = 0;

        virtual void tagItem( int type, int mode, QString tag, QString artist, QString token = QString() ) = 0;
        virtual void tagArtist( QString tag, QString artist, int mode = TAG_OVERWRITE ) = 0;
        virtual void tagAlbum( QString tag, QString artist, QString album, int mode = TAG_OVERWRITE ) = 0;
        virtual void tagTrack( QString tag, QString artist, QString track, int mode = TAG_OVERWRITE ) = 0;

        virtual void verifyUser( const QString& user, const QString& pass, const QString& passLower ) = 0;

        virtual QString baseHost() = 0;
        
    public slots:
        virtual void requestMetaDataRadio() = 0;
        virtual void requestMetaDataArtist( const MetaData& artist ) = 0;
        virtual void requestMetaDataTrack( const MetaData& track ) = 0;
        virtual void cancelLastMetaDataArtistRequest() = 0;
        virtual void cancelLastMetaDataTrackRequest() = 0;

        virtual void enableScrobbling( bool enabled, bool forceWebReq = false ) = 0;
        virtual void enableDiscoveryMode( bool enabled ) = 0;

        virtual void friends( QString username = QString() ) = 0;
        virtual void neighbours( QString username = QString() ) = 0;
        virtual void recentTracks( QString username = QString() ) = 0;
        virtual void recentLovedTracks( QString username = QString() ) = 0;
        virtual void recentBannedTracks( QString username = QString() ) = 0;

        virtual void unlisten( QString artist, QString track ) = 0;
        virtual void love( QString artist, QString track ) = 0;
        virtual void unlove( QString artist, QString track ) = 0;
        virtual void ban( QString artist, QString track ) = 0;
        virtual void unban( QString artist, QString track ) = 0;
        virtual void skip() = 0;

    signals:
        virtual void actionStarted() = 0;
        virtual void actionFinished() = 0;

        virtual void handshakeResult( int result ) = 0;
        virtual void streamingUrl( const QUrl& url ) = 0;
        virtual void stationChanged( QString url, QString name ) = 0;
        virtual void stationChangeFailed( QString error ) = 0;
        virtual void stationNameChanged( QString url, QString name ) = 0;
        virtual void songQueued() = 0;
        virtual void enableScrobblingDone() = 0;
        virtual void enableDiscoveryModeDone() = 0;

        virtual void metaDataResultRadio( const MetaData &metaData ) = 0;
        virtual void metaDataResultArtist( const MetaData &metaData, bool error = false ) = 0;
        virtual void metaDataResultTrack( const MetaData &metaData, bool error = false ) = 0;
        virtual void metaDataResultStation( const QString &name, const QString &feed, bool newStation ) = 0;
        virtual void radioStatusChanged( QString msg ) = 0;

        virtual void unlistenDone( const QString& artist, const QString& track, bool error = false ) = 0;
        virtual void loveDone( const QString& artist, const QString& track, bool error = false ) = 0;
        virtual void unloveDone( const QString& artist, const QString& track, bool error = false ) = 0;
        virtual void banDone( const QString& artist, const QString& track, bool error = false ) = 0;
        virtual void unbanDone( const QString& artist, const QString& track, bool error = false ) = 0;
        virtual void skipDone() = 0;
        virtual void recommendDone( const QString& artist, const QString& album, const QString& track, bool error = false ) = 0;
        virtual void tagDone( const QString& artist, const QString& album, const QString& track, bool error = false ) = 0;
        virtual void deleteFriendDone( const QString& username, bool error = false ) = 0;

        virtual void friendsResult( const QString& username, const QList<UserMetaData>& friends ) = 0;
        virtual void friendsResult( const QString& username, const QStringList& friends ) = 0;
        virtual void neighboursResult( const QString& username, const QList<QPair<QString, int> >& neighbours ) = 0;
        virtual void recentTracksResult( const QString& username, QList< QPair<QString, QString> > songs ) = 0;
        virtual void recentLovedTracksResult( const QString& username, QList< QPair<QString, QString> > songs ) = 0;
        virtual void recentBannedTracksResult( const QString& username, QList< QPair<QString, QString> > songs ) = 0;

        virtual void userTagsResult( const QString& username, const QList<QPair<QString, int> >& tags ) = 0;
        virtual void userArtistTagsResult( const QString& username, const QStringList& tags ) = 0;
        virtual void userAlbumTagsResult( const QString& username, const QStringList& tags ) = 0;
        virtual void userTrackTagsResult( const QString& username, const QStringList& tags ) = 0;
        virtual void userPicsResult( const QMap<QString, QString>& picUrls, bool error = false ) = 0;

        virtual void artistTagsResult( const QString& artist, const QStringList& tags ) = 0;
        virtual void topTagsResult( const QList<QPair<QString, int> >& tags ) = 0;

        virtual void similarArtistsResult( const QList<QPair<QString, int> >& result, const QString& artist = QString() ) = 0;
        virtual void searchTagResult( const QList<QPair<QString, int> >& result ) = 0;

        virtual void verifyResult( UserAuthCode result, BootstrapCode bootstrap ) = 0;
};

Q_DECLARE_INTERFACE( WebServiceInterface, "fm.last.Service.Web/1.0" )

#endif
